/**CFile***********************************************************************

  FileName    [baigCmd.c]

  PackageName [baig]

  Synopsis    [Functions to initialize and shut down the bAig manager.]

  Author      [Mohammad Awedh]

  Copyright [ This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]

******************************************************************************/

#include "maigInt.h"
#include "ntk.h"
#include "cmd.h"
#include "baigInt.h"

static char rcsid[] UNUSED = "$Id: baigCmd.c,v 1.18 2002/09/12 22:11:44 awedh Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Stucture declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int CommandPrintPartitionAig(Hrc_Manager_t ** hmgr, int argc, char ** argv);
static int CommandPrintAigStats(Hrc_Manager_t ** hmgr, int argc, char ** argv);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Initialize baig package]

  SideEffects [none]

  SeeAlso     [bAig_End]

******************************************************************************/
void
bAig_Init(void)
{
  Cmd_CommandAdd("print_partition_aig_dot", CommandPrintPartitionAig, 0);
  Cmd_CommandAdd("print_aig_stats", CommandPrintAigStats, 0);
}


/**Function********************************************************************

  Synopsis [End baig package]

  SideEffects []

  SeeAlso     [bAig_Init]

******************************************************************************/
void
bAig_End(void)
{
}

/**Function********************************************************************

  Synopsis    [Creates a new bAig manager.]

  Description [Creates a new bAig manager, initializes the NodeArray
  and noOfOutputFunctions. Returns a pointer to the manager if successful; 
  NULL otherwise.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
bAig_Manager_t *
bAig_initAig(
  int maxSize
)
{
  int i;

    bAig_Manager_t *manager = ALLOC(bAig_Manager_t,1);
 

    if (manager == NIL( bAig_Manager_t)){
      return manager;
    }
    if (maxSize >  bAigArrayMaxSize ) {
      maxSize =  bAigArrayMaxSize;
    }
    manager->NodesArray = ALLOC(bAigEdge_t, maxSize);
    manager->maxNodesArraySize = maxSize;
    manager->nodesArraySize = bAigFirstNodeIndex;

    manager->SymbolTable = st_init_table(strcmp,st_strhash);
    manager->nameList = ALLOC(char * , manager->maxNodesArraySize/bAigNodeSize);

    manager->HashTable =  ALLOC(bAigEdge_t, bAig_HashTableSize);
    for (i=0; i<bAig_HashTableSize; i++)
      manager->HashTable[i]= bAig_NULL; 

    manager->mVarList = array_alloc(mAigMvar_t, 0);
    manager->bVarList = array_alloc(mAigBvar_t, 0);
    return(manager);

} /* end of bAig_Init */

/**Function********************************************************************

  Synopsis [Quit bAig manager]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
bAig_quit(
  bAig_Manager_t *manager)
{
  char          *name;
  st_generator  *stGen;
  bAigEdge_t    varIndex;

  if (manager->mVarList != NIL(array_t)){
    array_free(manager->mVarList);
  }
  if (manager->bVarList != NIL(array_t)){
    array_free(manager->bVarList);
  }
  FREE(manager->HashTable);
  st_foreach_item_int(manager->SymbolTable, stGen, &name, &varIndex) {
    FREE(name);
  }
  st_free_table(manager->SymbolTable);
  /* i is too small to represent 80000
  for (i=0; i< manager->maxNodesArraySize/bAigNodeSize ; i++){
    FREE(manager->nameList[i]);
  }
  */
  FREE(manager->nameList);
  FREE(manager->NodesArray);
  FREE(manager);
}

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
void 
bAig_NodePrint(
   bAig_Manager_t  *manager,
   bAigEdge_t node)
{

  if (node == bAig_Zero){
    printf("0");
	return;
  }
  if (node == bAig_One){
    printf("1");
	return;
  }  

 if (bAig_IsInverted(node)){
	 printf(" NOT(");
 }


 if ( rightChild(node) == bAig_NULL){
   printf("Var Node");
     if (bAig_IsInverted(node)){
     	 printf(")");
	 }
	 return;
 }
 printf("(");
 bAig_NodePrint(manager, leftChild(node));
 printf("AND");
 bAig_NodePrint(manager, rightChild(node));
 printf(")");
 if (bAig_IsInverted(node)){
	 printf(")");
 }

}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Implements the print_partition_aig_dot command.]
  
  CommandName [print_partition_aig_dot]

  CommandSynopsis [print a dot description of the AND/INVERTER graph that was built
  for the flattened network]

  CommandArguments [\[-h\] &lt;file_name&gt;]

  CommandDescription [Write a file in the format taken by the tool <tt>dot</tt>
  depicting the topology of the AND/INVERTER graph. 
  Dot is a tool that given a description of a graph in a certain format it 
  produces a postscript print of the graph. 
  For more information about <tt>dot</tt> look in <a
  href="http://www.research.att.com/orgs/ssr/book/reuse">
  http://www.research.att.com/orgs/ssr/book/reuse</a>. Once a dot file is
  produced with this command, the shell command <tt>dot -Tps <filename>
  &gt;<file>.ps</tt> will produce the postscript file depicting the network.<p>

  If no argument is specified on the command line, the description is written
  to the standard output.<p>

  Command options:<p>
  <dl><dt> -h
  <dd> Print the command usage.
  </dl>]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static int
CommandPrintPartitionAig(
  Hrc_Manager_t ** hmgr,
  int  argc,
  char ** argv)
{
  FILE *fp;
  int c;
  Ntk_Network_t *network = Ntk_HrcManagerReadCurrentNetwork(*hmgr);
  mAig_Manager_t *manager;

  util_getopt_reset();
  while ((c = util_getopt(argc,argv,"h")) != EOF){
    switch(c){
      case 'h':
        goto usage;
      default:
        goto usage;
    }
  }
  /* Check if the network has been read in */
  if (network == NIL(Ntk_Network_t)) {
    return 1;
  }

  if (argc == 1) {
    fp = stdout;
  }
  else if (argc == 2) {
    fp = Cmd_FileOpen(*(++argv), "w", NIL(char *), /* silent */ 1);
    if (fp == NIL(FILE)) {
      (void) fprintf(vis_stderr, "Cannot write to %s\n", *argv);
      return 1;
    }
  }
  else {
    goto usage;
  }

  manager = Ntk_NetworkReadMAigManager(network);

  if (manager == NIL(mAig_Manager_t)) {
    return 1;
  }
  bAig_PrintDot(fp, manager);
  return 0;
usage:
  (void) fprintf(vis_stderr, "usage: print_partition_aig_dot [-h] [file]\n");
  (void) fprintf(vis_stderr, "    -h\t\tprint the command usage\n");
  return 1;
}

/**Function********************************************************************

  Synopsis    [Implements the print_aig_stats command.]
  CommandName [print_aig_stats]

  CommandSynopsis [print statistics about the AND/INVERTER graph.]

  CommandArguments [\[-h\] \[-n\]]

  CommandDescription [Print the following statistics about the AND/INVERTER 
  graph:

 <dl>
  <li> <b>Maximum number of nodes:</b> the maximum number of nodes in the AND/
  INVERTER graph.

  <li> <b>Current number of nodes:</b> the number of nodes in the AND/INVERTER
  graph that are used.

 </dl>

  Note that <tt>build_partition_maigs</tt> must be called before this command.<p>

  Command options:<p>

  <dl>
  <dt> -n
  <dd> Print the name of the nodes.<p>
  <dt> -h
  <dd> Print the command usage.<p>
  </dl>
  ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static int
CommandPrintAigStats(
  Hrc_Manager_t ** hmgr,
  int  argc,
  char ** argv)
{
  Ntk_Network_t *network = Ntk_HrcManagerReadCurrentNetwork(*hmgr);
  mAig_Manager_t *manager;
  int i;
  int printVar = 0;
  int c;

  /*
   * Parse command line options.
   */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "hn")) != EOF) {
    switch(c) {
      case 'h':
        goto usage;
     case 'n':
       printVar = 1;
        break;
      default:
        goto usage;
    }
  }
  if (network == NIL(Ntk_Network_t)) {
    return 1;
  }

  manager = Ntk_NetworkReadMAigManager(network);

  if (manager == NIL(mAig_Manager_t)) {
    return 1;
  }
  (void) fprintf(vis_stdout,"And/Inverter graph Static Report\n");
  (void) fprintf(vis_stdout,"Maximum number of nodes: %d\n", manager->maxNodesArraySize);
  (void) fprintf(vis_stdout,"Current number of nodes: %d\n\n", manager->nodesArraySize);
  if (printVar == 1){
     (void) fprintf(vis_stdout,"Nodes name\n");
     for(i=1; i< manager->nodesArraySize/4; i++){
        if (manager->nameList[i][0] != 'N'){
           (void) fprintf(vis_stdout,"%s\n", manager->nameList[i]);
        }
     }
  }
  return 0;
usage:
  (void) fprintf(vis_stderr, "usage: print_aig_stats [-h] [-n] \n");
  (void) fprintf(vis_stderr, "   -h print the command usage\n");
  (void) fprintf(vis_stderr, "   -n list of variables\n");
  return 1;		/* error exit */
}
