/**CFile***********************************************************************

  FileName    [baigNode.c]

  PackageName [baig]

  Synopsis    [Routines to access node data structure of the And/Inverter
               graph.]

  Author      [Mohammad Awedh]

  Copyright [ This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]


******************************************************************************/

#include "baigInt.h"

static char rcsid[] UNUSED = "$Id: baigNode.c,v 1.13 2002/09/21 03:01:14 awedh Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void incRefCount(bAig_Manager_t *manager, bAigEdge_t nodeIndex);
static void decRefCount(bAig_Manager_t *manager, bAigEdge_t nodeIndex);
static bAigEdge_t HashTableLookup(bAig_Manager_t *manager, bAigEdge_t node1, bAigEdge_t node2);
static int HashTableAdd(bAig_Manager_t *manager, bAigEdge_t nodeIndexParent, bAigEdge_t nodeIndex1, bAigEdge_t nodeIndex2);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Read Node's name.]

  Description [Read the name of a node given its index.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
nameType_t *
bAig_NodeReadName(
   bAig_Manager_t *manager,
   bAigEdge_t     node)
{
  return manager->nameList[bAigNodeID(node)];
}

/**Function********************************************************************

  Synopsis    [Set Node's name.]

  Description [Set the name of node in Symbol table and name List]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void 
bAig_NodeSetName(
   bAig_Manager_t *manager,
   bAigEdge_t     node,
   nameType_t     *name)
{
  st_insert(manager->SymbolTable, name, (char*) (long) node);
  manager->nameList[bAigNodeID(node)] = name;
}

/**Function********************************************************************

  Synopsis    [Returns the index of the right node.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
int
bAig_NodeReadIndexOfRightChild(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex)
{
  return rightChild(nodeIndex);
}

/**Function********************************************************************

  Synopsis    [Returns the index of the left node.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_NodeReadIndexOfLeftChild(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex)
{
  return leftChild(nodeIndex);
}

/**Function********************************************************************

  Synopsis    [Performs the Logical AND of two nodes.]

  Description [This function performs the Logical AND of two nodes.  The inputs
               are the indices of the two nodes.  This function returns the index 
               of the result node.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_And(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex1,
   bAigEdge_t nodeIndex2)
{

  bAigEdge_t newNodeIndex;

  newNodeIndex = nodeIndex1;     /* The left node has the smallest index */
  if (nodeIndex1 > nodeIndex2){
    nodeIndex1 = nodeIndex2;
    nodeIndex2 = newNodeIndex;
  }

  if ( nodeIndex2 == bAig_Zero ) {
    return bAig_Zero;
  }
  if ( nodeIndex1 == bAig_Zero ) {
    return bAig_Zero;
  }
  if ( nodeIndex2 == bAig_One ) {
    return nodeIndex1;
  }
  if ( nodeIndex1 == bAig_One ) {
    return nodeIndex2;
  }
  if ( nodeIndex1 == nodeIndex2 ) {
    return nodeIndex1;
  }
  if ( nodeIndex1 == bAig_Not(nodeIndex2) ) {
    return bAig_Zero;
  }

  /* Look for the new node in the Hash table */
  newNodeIndex = HashTableLookup(manager, nodeIndex1, nodeIndex2);

  if (newNodeIndex == bAig_NULL){ 
     newNodeIndex = bAigCreateAndNode(manager, nodeIndex1, nodeIndex2) ;
     HashTableAdd(manager, newNodeIndex, nodeIndex1, nodeIndex2);
  }
    
  incRefCount(manager, nodeIndex1); /*Increment no of fanouts of this node */   
  incRefCount(manager, nodeIndex2); 
  
  return newNodeIndex;

}

/**Function********************************************************************

  Synopsis    [Performs the Logical OR of two nodes.]

  Description [This function performs the Logical OR of two nodes.  The inputs
               are the indices of the two nodes.  This function returns the index 
               of the result node.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_Or(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex1,
   bAigEdge_t nodeIndex2)
{
  return ( bAig_Not(bAig_And(manager,  bAig_Not(nodeIndex1),  bAig_Not(nodeIndex2))));
}

/**Function********************************************************************

  Synopsis    [Performs the Logical XOR of two nodes.]

  Description [This function performs the Logical XOR of two nodes.  The inputs
               are the indices of the two nodes.  This function returns the index 
               of the result node.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_Xor(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex1,
   bAigEdge_t nodeIndex2)
{
 return bAig_Or(manager,
	    bAig_And(manager, nodeIndex1, bAig_Not(nodeIndex2)),
	    bAig_And(manager, bAig_Not(nodeIndex1),nodeIndex2)
	);
}

/**Function********************************************************************

  Synopsis    [Performs the Logical Equal ( <--> XNOR) of two nodes.]

  Description [This function performs the Logical XNOR of two nodes.  The inputs
               are the indices of the two nodes.  This function returns the index 
               of the result node.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_Eq(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex1,
   bAigEdge_t nodeIndex2)
{
 return bAig_Not(
	    bAig_Xor(manager, nodeIndex1, nodeIndex2)
	    );
}

/**Function********************************************************************

  Synopsis    [Performs the Logical Then (--> Implies) of two nodes.]

  Description [This function performs the Logical Implies of two nodes.  The inputs
               are the indices of the two nodes.  This function returns the index 
               of the result node.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_Then(
   bAig_Manager_t *manager,
   bAigEdge_t nodeIndex1,
   bAigEdge_t nodeIndex2)
{
 return bAig_Or(manager,
	        bAig_Not(nodeIndex1),
                nodeIndex2);
}

/**Function********************************************************************

  Synopsis    [create new node] 

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_CreateNode(
   bAig_Manager_t  *manager,
   bAigEdge_t nodeIndex1,
   bAigEdge_t nodeIndex2)
{
  
  bAigEdge_t newNode = manager->nodesArraySize;

  if (manager->nodesArraySize >= manager->maxNodesArraySize ){
    manager->maxNodesArraySize = 2* manager->maxNodesArraySize;
    manager->NodesArray = REALLOC(bAigEdge_t, manager->NodesArray, manager->maxNodesArraySize);
    manager->nameList   = REALLOC(char *,     manager->nameList  , manager->maxNodesArraySize/bAigNodeSize);
  }
  manager->NodesArray[bAig_NonInvertedEdge(newNode)+bAigRight] = nodeIndex2;
  manager->NodesArray[bAig_NonInvertedEdge(newNode)+bAigLeft]  = nodeIndex1;
 
  next(newNode)     = bAig_NULL;
  refCount(newNode) = 0;        

  manager->nodesArraySize +=bAigNodeSize;
  return(newNode);
}

/**Function********************************************************************

  Synopsis    [create var node ] 

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAig_CreateVarNode(
   bAig_Manager_t  *manager,
   nameType_t	   *name)
{
   
  bAigEdge_t varIndex;


  if (!st_lookup(manager->SymbolTable, name, (char**) &varIndex)){
    varIndex = bAig_CreateNode(manager, bAig_NULL, bAig_NULL);
    if (varIndex == bAig_NULL){
      return (varIndex);
    } 
    /* Insert the varaible in the Symbol Table */
    st_insert(manager->SymbolTable, name, (char*) (long) varIndex);
    manager->nameList[bAigNodeID(varIndex)] = name; 
    return(varIndex);
  }
  else {
    return (varIndex);
  }
}

/**Function********************************************************************

  Synopsis [Return True if this node is Variable node]

  SideEffects []

******************************************************************************/
int 
bAig_isVarNode(
   bAig_Manager_t *manager,
   bAigEdge_t      node)
{
  if((rightChild(node) == bAig_NULL) && (leftChild(node) == bAig_NULL)) {
    return 1;
  }
  return 0;
}

/**Function********************************************************************

  Synopsis [Build the binary Aig for a given bdd]

  SideEffects [Build the binary Aig for a given bdd.
               We assume that the return bdd nodes from the foreach_bdd_node are
	       in the order from childeren to parent. i.e all the childeren
	       nodes are returned before the parent node.]

  SideEffects []

  SeeAlso     []
  
******************************************************************************/
bAigEdge_t
bAig_bddtobAig(
   bAig_Manager_t *bAigManager,
   bdd_t          *fn)
{
  bdd_gen     *gen;
  bdd_node    *node, *thenNode, *elseNode, *funcNode;
  bdd_manager *bddManager = bdd_get_manager(fn);
  bdd_node    *one        = bdd_read_one(bddManager);
  bdd_node    *zero       = bdd_read_logic_zero(bddManager);
  int         is_complemented;
  
  bAigEdge_t  var, left, right, result;
  
  char     *name;
  st_table *bddTObaigTable;
  
  bddTObaigTable = st_init_table(st_numcmp, st_numhash);

  if (fn == NULL){
    return bAig_NULL;
  }
  
  st_insert(bddTObaigTable, (char *) (long) bdd_node_read_index(one),  (char *) bAig_One);
  st_insert(bddTObaigTable, (char *) (long) bdd_node_read_index(zero), (char *) bAig_Zero);

  funcNode = bdd_get_node(fn, &is_complemented);
  
  if (bdd_is_constant(funcNode)){
    st_lookup(bddTObaigTable, (char *) (long) bdd_node_read_index(funcNode),
	      (char **)&result);
    return result;
  }
  
  foreach_bdd_node(fn, gen, node){
    name = strcat("BDD", util_inttostr(bdd_node_read_index(node)));
    var  = bAig_CreateVarNode(bAigManager, name);

    thenNode  = bdd_bdd_T(node);
    st_lookup(bddTObaigTable, (char *) (long) bdd_node_read_index(thenNode),
	      (char **)&left);
      
    elseNode  = bdd_bdd_E(node);
    st_lookup(bddTObaigTable, (char *) (long) bdd_node_read_index(elseNode),
	      (char **) &right);

    result =  bAig_Or(bAigManager, bAig_And(bAigManager, var, left),
		                   bAig_And(bAigManager, bAig_Not(var), right));
    st_insert(bddTObaigTable, (char *) (long) bdd_node_read_index(node),
	      (char *) (long) result);
  }
  st_lookup(bddTObaigTable, (char *) (long) bdd_node_read_index(funcNode),
	    (char **) &result);  
  st_free_table(bddTObaigTable);
  if (is_complemented){
    return bAig_Not(result);
  } else {
    return result;
  }
} /* end of bAig_bddtobAig() */

/**Function********************************************************************

  Synopsis [.]

  SideEffects []

******************************************************************************/
int
bAig_PrintDot(
  FILE *fp,
  bAig_Manager_t *manager)
{
  int i;

  /*
   * Write out the header for the output file.
   */
  (void) fprintf(fp, "digraph \"AndInv\" {\n  rotate=90;\n");
  (void) fprintf(fp, "  margin=0.5;\n  label=\"AndInv\";\n");
  (void) fprintf(fp, "  size=\"10,7.5\";\n  ratio=\"fill\";\n");
 
  
  for (i=bAigFirstNodeIndex ; i< manager->nodesArraySize ; i+=bAigNodeSize){
    (void) fprintf(fp,"Node%d  [label=\"%s \"];\n",i,bAig_NodeReadName(manager, i));
  }
  for (i=bAigFirstNodeIndex ; i< manager->nodesArraySize ; i+=bAigNodeSize){
    if (rightChild(i) != bAig_NULL){
      if (bAig_IsInverted(rightChild(i))){
        (void) fprintf(fp,"Node%d -> Node%d [color = red];\n",bAig_NonInvertedEdge(rightChild(i)), i);
      }
      else{
        (void) fprintf(fp,"Node%d -> Node%d;\n",bAig_NonInvertedEdge(rightChild(i)), i);
      }
      if (bAig_IsInverted(leftChild(i))){
        (void) fprintf(fp,"Node%d -> Node%d [color = red];\n",bAig_NonInvertedEdge(leftChild(i)), i);
      }
      else{
        (void) fprintf(fp,"Node%d -> Node%d;\n",bAig_NonInvertedEdge(leftChild(i)), i);
      }
    }/* if */ 
  }/*for */

  (void) fprintf(fp, "}\n");

  return 1;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Create AND node and assign name to it ] 

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
bAigEdge_t
bAigCreateAndNode(
   bAig_Manager_t  *manager,
   bAigEdge_t node1,
   bAigEdge_t node2)
{
   
  bAigEdge_t varIndex;
  char       *name = NIL(char);
  char       *node1Str = util_inttostr(node1);
  char       *node2Str = util_inttostr(node2);

  name = util_strcat4("Nd", node1Str,"_", node2Str);
  while (st_lookup_int(manager->SymbolTable, name, &varIndex)){
    printf("Find redundant node at %d %d\n", node1, node2);
    name = util_strcat3(name, node1Str, node2Str);
  }
  FREE(node1Str);
  FREE(node2Str);
  varIndex = bAig_CreateNode(manager, node1, node2);
  if (varIndex == bAig_NULL){
    FREE(name);
    return (varIndex);
  } 
  /* Insert the varaible in the Symbol Table */
  st_insert(manager->SymbolTable, name, (char*) (long) varIndex);
  manager->nameList[bAigNodeID(varIndex)] = name; 
  return(varIndex);

}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Increment the reference count of a bAig node.]

  Description [.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static void 
incRefCount(
  bAig_Manager_t *manager,  
  bAigEdge_t  nodeIndex)
{
  refCount(bAig_NonInvertedEdge(nodeIndex))++;
}

/**Function********************************************************************

  Synopsis    [Decrement the reference count of a bAig node.]

  Description [.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static void
decRefCount(
  bAig_Manager_t *manager,  
  bAigEdge_t  nodeIndex)
{
    refCount(bAig_NonInvertedEdge(nodeIndex))--;
}

/**Function********************************************************************

  Synopsis    [Look for the node in the Hash Table.]

  Description [.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static bAigEdge_t
HashTableLookup(
  bAig_Manager_t *manager,  
  bAigEdge_t  node1,
  bAigEdge_t  node2)
{
  bAigEdge_t key = HashTableFunction(node1, node2);
  bAigEdge_t node;
  
  node = manager->HashTable[key];
  if  (node == bAig_NULL) {
    return bAig_NULL;
  }
  else{
    while ( (rightChild(bAig_NonInvertedEdge(node)) != node2) ||
	    (leftChild(bAig_NonInvertedEdge(node))  != node1)) {
  
      if (next(bAig_NonInvertedEdge(node)) == bAig_NULL){
	return(bAig_NULL); 
      }
      node = next(bAig_NonInvertedEdge(node)); /* Get the next Node */
    } /* While loop */ 
    return(node);

  } /* If Then Else */
    
} /* End of HashTableLookup() */

/**Function********************************************************************

  Synopsis    [Add a node in the Hash Table.]

  Description []

  SideEffects []

  SeeAlso     []

******************************************************************************/
static int
HashTableAdd(
  bAig_Manager_t   *manager,  
  bAigEdge_t  nodeIndexParent,
  bAigEdge_t  nodeIndex1,
  bAigEdge_t  nodeIndex2)
{
  bAigEdge_t key = HashTableFunction(nodeIndex1, nodeIndex2);
  bAigEdge_t nodeIndex;
  bAigEdge_t node;
                                                
  nodeIndex = manager->HashTable[key];
  if  (nodeIndex == bAig_NULL) {
    manager->HashTable[key] = nodeIndexParent;
    return TRUE;
  }
  else{
    node = nodeIndex;
    nodeIndex = next(bAig_NonInvertedEdge(nodeIndex));  /* Get the Node */
    while (nodeIndex != bAig_NULL) {
      node = nodeIndex;
      nodeIndex = next(bAig_NonInvertedEdge(nodeIndex));
    }
    next(bAig_NonInvertedEdge(node)) = nodeIndexParent;
    return TRUE;
  }
    
} /* End of HashTableAdd() */

















