/**CFile***********************************************************************

  FileName    [bmcAbstract.c]

  PackageName [bmc]

  Synopsis    [Routines to deal with Abstract model]

  Author      [Mohammad Awedh]

  Copyright   [This file was created at the University of Colorado at Boulder.
  The University of Colorado at Boulder makes no warranty about the suitability
  of this software for any purpose.  It is presented on an AS IS basis.]

******************************************************************************/

#include "bmcInt.h"
#include "mdd.h"

static char rcsid[] UNUSED = "$Id: bmcAbstract.c,v 1.21 2002/09/21 07:48:45 fabio Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Check whether the abstract path holds in the concrete model]

  Description [Check whether the abstract path holds in the concrete model.
  This function generates a concrete path of length equal to the length of the
  abstract path restricting the variables that appear in the abstract path to
  their values in the abstract path.  The generated path is then passed to SAT
  solver.  If the SAT solve find a satisfying assignmnet, then the path is real.
  otherwise, the path is spurious.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
Bmc_AbstractCheckAbstractPath(
  Ntk_Network_t *network,
  McPath_t      *abstractPath,
  array_t	*supportList)
{
  mAig_Manager_t  *maigManager       = Ntk_NetworkReadMAigManager(network);
  st_table        *nodeToMvfAigTable = NIL(st_table);  /* mapes each node with its mvfAig */
  lsGen           gen;
  Ntk_Node_t        *latch, *latchData;
  bAigEdge_t        *latchBAig, *dataBAig;
  MvfAig_Function_t *dataMvfAig, *latchMvfAig;
  int               pathLength = array_n(abstractPath->stemArray)  - 1 +
                               array_n(abstractPath->cycleArray) - 2;
  BmcCnfClauses_t   *cnfClauses;
  FILE              *cnfFile;
  BmcOption_t       *options = BmcOptionAlloc();
  
  int        i, loopState, mvfSize;
  int        stemArraySize = array_n(abstractPath->stemArray);
  array_t    *result;
  mdd_t      *lastState = array_fetch_last(mdd_t *, abstractPath->stemArray);
  st_table   *COI= st_init_table(st_ptrcmp, st_ptrhash);
  
  (void) fprintf(vis_stdout, "Enter check Abstract path \n");
  options->verbosityLevel = BmcVerbositySome_c;
  
  /* nodeToMvfAigTable Maps each node with its Multi-function And/Inv graph */
  nodeToMvfAigTable = (st_table *) Ntk_NetworkReadApplInfo(network, MVFAIG_NETWORK_APPL_KEY);
  if (nodeToMvfAigTable == NIL(st_table)){
    (void) fprintf(vis_stderr, "** bmc error: please run buid_partiton_maigs first\n");
    BmcOptionFree(options);
    return;
  }
  /*
  options->satInFile = "abstract.cnf";
  */
  /* create SAT Solver input file */
  if (options->cnfFileName == NIL(char)) {
    options->satInFile = BmcCreateTmpFile(); 
    if (options->satInFile == NIL(char)){
      BmcOptionFree(options);
      return;
    }
  } else {
    options->satInFile = options->cnfFileName;
  }
  /* create SAT Solver output file */
  options->satOutFile = BmcCreateTmpFile();
  if (options->satOutFile == NIL(char)){
    BmcOptionFree(options);
    return;
  }
  cnfFile = Cmd_FileOpen(options->satInFile, "w", NIL(char *), 0);  
  if (cnfFile == NIL(FILE)) {
    (void) fprintf(vis_stderr,
		   "** bmc error: Cannot create CNF output file %s\n",
		   options->satInFile);
    return;
  }
  cnfClauses = BmcCnfClausesAlloc();
  (void) fprintf(vis_stdout, "# BMC: Generate CNF for concrete counterexample of length %d\n",
		 pathLength);
  BmcGenerateCnfForPath(network, 0, pathLength, cnfClauses, nodeToMvfAigTable, COI);
  st_free_table(COI);
  /*
    Get state number at which cycle loop back to in the stem array
   */
  for( loopState = 0; loopState < array_n(abstractPath->stemArray) ; loopState++ ) {
    mdd_t *tmpState = array_fetch( mdd_t *, abstractPath->stemArray, loopState);
    if ( mdd_equal(lastState, tmpState) ) {
      break;
    }
  }
  if (loopState > array_n(abstractPath->stemArray)){
    fprintf(vis_stderr,"The last state in the cycle path is NOT in the stem path\n");
    BmcOptionFree(options);
    BmcCnfClausesFree(cnfClauses);
    return;
  }
  /* Generate the CNF for the loop back from the last state of the cycle path
     to state i in the stem path
  */
  Ntk_NetworkForEachLatch(network, gen, latch) {
   latchData  = Ntk_LatchReadDataInput(latch);
  
   dataMvfAig = BmcReadMvfAig(latchData, nodeToMvfAigTable);
   if (dataMvfAig ==  NIL(MvfAig_Function_t)){
     (void) fprintf(vis_stdout, "No multi-valued function for this node %s \n",
		    Ntk_NodeReadName(latchData));
     break;
   }
   latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);

   mvfSize   = array_n(dataMvfAig);
   dataBAig  = ALLOC(bAigEdge_t, mvfSize);
   latchBAig = ALLOC(bAigEdge_t, mvfSize);   
   
   for(i=0; i< mvfSize; i++){
     dataBAig[i]  = MvfAig_FunctionReadComponent(dataMvfAig,  i);
     latchBAig[i] = MvfAig_FunctionReadComponent(latchMvfAig, i);
   }
   BmcGenerateCnfFromStateTostate(maigManager, dataBAig, latchBAig, mvfSize,
				  pathLength, loopState,
				  cnfClauses, 0);
   FREE(dataBAig);
   FREE(latchBAig);
  } /* ForEachLatch loop*/
  
  (void) fprintf(vis_stdout, "No of latches = %d \n",Ntk_NetworkReadNumLatches(network));
  { /* generate clauses for states that their vlaues are known */
   bAigEdge_t  bAigState;
   int         lastStateNo;
   int         j;
   mdd_manager *mddManager = Ntk_NetworkReadMddManager(network);
   array_t     *valueArray;
   Ntk_Node_t  *latch;
   int         mddId, value;
   mdd_t       *state;
   array_t     *clause = array_alloc(int, 1);
   
   for( i = 0; i < stemArraySize; i++ ) {
     state      = array_fetch( mdd_t *, abstractPath->stemArray, i );
     valueArray = McConvertMintermToValueArray(state, supportList, mddManager);
     bAigState = bAig_One;
     arrayForEachItem(int, supportList, j, mddId ) {
        latch       = Ntk_NetworkFindNodeByMddId(network, mddId );
        latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);
        value       = array_fetch( int, valueArray, j );
	(void) fprintf(vis_stdout, "Abstract node in stem %s = %d\n",
		       Ntk_NodeReadName(latch), value);
	bAigState = bAig_And(maigManager, bAigState,
			     MvfAig_FunctionReadComponent(latchMvfAig, value));

     }
     array_insert(int, clause, 0, BmcGenerateCnfFormulaForAigFunction(
                                  maigManager, bAigState, i, cnfClauses));
     BmcCnfInsertClause(cnfClauses, clause);
   } /* for each item in abstract stem array */

   lastStateNo = stemArraySize-1;
   for( i = 1 ; i < array_n(abstractPath->cycleArray)-1; i++ ) {
     state      = array_fetch( mdd_t *, abstractPath->cycleArray, i );
     valueArray = McConvertMintermToValueArray(state, supportList, mddManager);
     bAigState = bAig_One;
     arrayForEachItem(int , supportList, j, mddId ) {
        latch       = Ntk_NetworkFindNodeByMddId(network, mddId );
        latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);
        value       = array_fetch( int, valueArray, j );
	(void) fprintf(vis_stdout, "Abstract node in cycle %s = %d\n",
		       Ntk_NodeReadName(latch), value);
        bAigState   = bAig_And(maigManager, bAigState,
			       MvfAig_FunctionReadComponent(latchMvfAig, value));
     }
     array_insert(int, clause, 0, BmcGenerateCnfFormulaForAigFunction(
                                  maigManager, bAigState, lastStateNo + i, cnfClauses));
     BmcCnfInsertClause(cnfClauses, clause);
   } /* for each item in abstract cycle array */
   array_free(clause);
  } /* generate clauses for states that their vlaues are known */

  BmcWriteClauses(maigManager, cnfFile, cnfClauses); 
  (void) fclose(cnfFile);
  result = BmcCheckSAT(options);
  (void) fprintf(vis_stdout,"Maximum Counterexample length = %d \n", pathLength);

  if(result != NIL(array_t)){
    (void) fprintf(vis_stdout,"Found Counterexample of length = %d \n", pathLength);
    /*
    BmcPrintCounterExample(network, nodeToMvfAigTable, cnfClauses, result, pathLength, options);
    */
  }  
  /*
  if(result != NIL(array_t)){
    for (i = 0; i < array_n(result); i++) {
       (void) fprintf(vis_stdout, "%d\n",array_fetch(int, result, i));
    }
  }
  */
  /* free all used memories */
  BmcCnfClausesFree(cnfClauses);
  array_free(result);
  BmcOptionFree(options);
  (void) fprintf(vis_stdout, "End Of Abstract \n");
  return;
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
