/**CFile***********************************************************************

  FileName    [bmcBmc.c]

  PackageName [bmc]

  Synopsis    [BMC ltl model checker.]

  Author      [Mohammad Awedh]

  Copyright   [This file was created at the University of Colorado at Boulder.
  The University of Colorado at Boulder makes no warranty about the suitability
  of this software for any purpose.  It is presented on an AS IS basis.]
  
******************************************************************************/

#include "bmcInt.h"

static char rcsid[] UNUSED = "$Id: bmcBmc.c,v 1.42 2002/09/21 07:49:20 fabio Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static int checkIndex(int index, BmcCnfClauses_t *cnfClauses);
static int doAnd(int left, int right);
static int doOr(int left, int right);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Generates a counterexample of length controlled by fields in
  options structure, and then calls SAT solver to check for the existance of
  this counterexample.]

  Description [This function implements the Bounded Model Checking that was
  introdunced in the paper: Symbolic Model Checking without BDDs by A Biere,
  E. M. Clarke, R. Raimi, and Y. Zhu.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void 
Bmc_CheckLTLFormula(
   Ntk_Network_t   *network,
   Ctlsp_Formula_t *ltlFormula,
   BmcOption_t     *options,
   st_table        *CoiTable)
{
  mAig_Manager_t    *manager = Ntk_NetworkReadMAigManager(network);
  st_table          *nodeToMvfAigTable = NIL(st_table);  /* node to mvfAig */
  Ctlsp_Formula_t   *negFormula;
  BmcCnfClauses_t   *cnfClauses = NIL(BmcCnfClauses_t);
  FILE              *cnfFile;

  Ntk_Node_t        *latch, *latchData;
  MvfAig_Function_t *dataMvfAig, *latchMvfAig;
  bAigEdge_t        *latchBAig, *dataBAig;
  lsGen             gen;

  array_t           *orClause, *loopClause, *tmpclause;
  int               i, l, mvfSize, loopIndex, andIndex, loop;
  int               noLoopIndex, allLoopIndex;
  array_t           *result = NIL(array_t);
  Ctlsp_Formula_t   *ltlFormula1; 
  int               leftValue  = 0;
  int               rightValue = 0;
  long              startTime, endTime;
  int               k;
  int               minK = options->minK;
  int               maxK = options->maxK;
  boolean           boundedFormula;
  int               depth;

  /* ************** */
  /* Initialization */
  /* ************** */
  startTime = util_cpu_ctime();
  /* nodeToMvfAigTable maps each node to its multi-function And/Inv graph */
  nodeToMvfAigTable =
    (st_table *) Ntk_NetworkReadApplInfo(network, MVFAIG_NETWORK_APPL_KEY);
  if (nodeToMvfAigTable == NIL(st_table)){
    (void) fprintf(vis_stderr,
		   "** bmc error: please run buid_partiton_maigs first\n");
    return;
  }

  negFormula  = Ctlsp_FormulaCreate(Ctlsp_NOT_c, ltlFormula, NIL(Ctlsp_Formula_t));
  ltlFormula1 = Ctlsp_LtlFormulaExpandAbbreviation(negFormula);
  ltlFormula  = Ctlsp_LtlFormulaNegationNormalForm(ltlFormula1);
  if(ltlFormula != ltlFormula1) Ctlsp_FormulaFree(ltlFormula1);
  FREE(negFormula);

  negFormula = ltlFormula;
  ltlFormula = Ctlsp_LtlFormulaSimpleRewriting(ltlFormula);
  Ctlsp_FormulaFree(negFormula);

  Ctlsp_LtlSetClass(ltlFormula);
  /*
  fprintf(vis_stdout, "\nNeg Formula: ");
  Ctlsp_FormulaPrint(vis_stdout, ltlFormula);
  fprintf(vis_stdout, "\n");
  */

  /* For bounded formulae use a tighter upper bound if possible. */
  boundedFormula = Ctlsp_LtlFormulaTestIsBounded(ltlFormula, &depth);
  if (boundedFormula && depth < maxK && depth >= minK) {
    maxK = depth;
  }

  if (options->verbosityLevel != BmcVerbosityNone_c){
    (void) fprintf(vis_stdout, "General LTL BMC\n");
    if (boundedFormula){
      (void) fprintf(vis_stdout, "Bounded formula of depth %d\n", depth);
    }
    (void) fprintf(vis_stdout, "Apply BMC on counterexample of length >= %d and <= %d (+%d) \n",
		  minK, maxK, options->step);
  }

  k= minK;
  while( (result == NIL(array_t)) && (k <= maxK)){
    if (options->verbosityLevel == BmcVerbosityMax_c) {
      (void) fprintf(vis_stdout, "\nGenerate counterexample of length %d\n", k);
    }
    cnfClauses = BmcCnfClausesAlloc();

    cnfFile = Cmd_FileOpen(options->satInFile, "w", NIL(char *), 0);  
    if (cnfFile == NIL(FILE)) {
      (void)fprintf(vis_stderr,
		    "** bmc error: Cannot create CNF output file %s\n",
		    options->satInFile);
      return;
    }
    BmcGenerateCnfForPath(network, 0, k, cnfClauses, nodeToMvfAigTable, CoiTable);

    noLoopIndex = BmcGenerateCnfForLtl(network, ltlFormula, 0, k, -1, cnfClauses);
    leftValue   = checkIndex(noLoopIndex, cnfClauses);

    orClause    = array_alloc(int, 0);   

    if (leftValue != 0){
      loopClause   = array_alloc(int, k+2);
      allLoopIndex = cnfClauses->cnfGlobalIndex++;

      array_insert(int, loopClause, k+1, -allLoopIndex);
      if (leftValue == -1){
	andIndex   = cnfClauses->cnfGlobalIndex++;
	/* !Lk * NoLoop(f) */
	tmpclause  = array_alloc(int, 2);
	array_insert(int, tmpclause, 0, noLoopIndex);
	array_insert(int, tmpclause, 1, -andIndex);
	BmcCnfInsertClause(cnfClauses, tmpclause);

	array_insert(int, tmpclause, 0, -allLoopIndex);
	array_insert(int, tmpclause, 1, -andIndex);
	BmcCnfInsertClause(cnfClauses, tmpclause);
	array_free(tmpclause);

	array_insert_last(int, orClause, andIndex);
      } else {
	array_insert_last(int, orClause, -allLoopIndex);
      }
    } else {
      loopClause = array_alloc(int, k+1);
    }
    for(l=0; l<=k; l++){
      loopIndex = BmcGenerateCnfForLtl(network, ltlFormula, 0, k, l, cnfClauses);
      rightValue = checkIndex(loopIndex, cnfClauses);
      if (rightValue == 0){
	break;
      }
      if (rightValue !=0){
	loop = cnfClauses->cnfGlobalIndex++;
	Ntk_NetworkForEachLatch(network, gen, latch) {
	  int tmp;
	  /* skip the latch if it is not in COI of property*/
	  if (!st_lookup_int(CoiTable, (char *) latch, &tmp)){
	    /*printf("Not in COI %s\n",  Ntk_NodeReadName(latch));*/
	    continue;
	  }
	  latchData  = Ntk_LatchReadDataInput(latch);

	  dataMvfAig = BmcReadMvfAig(latchData, nodeToMvfAigTable);
	  if (dataMvfAig ==  NIL(MvfAig_Function_t)){
	    (void) fprintf(vis_stdout,
			   "No multi-valued function for this node %s \n",
			   Ntk_NodeReadName(latchData));
	    return;
	  }
	  latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);
	  if (latchMvfAig ==  NIL(MvfAig_Function_t)){
	    latchMvfAig = BmcNodeBuildMVF(network, latch);
	  }
	  mvfSize   = array_n(dataMvfAig);
	  dataBAig  = ALLOC(bAigEdge_t, mvfSize);
	  latchBAig = ALLOC(bAigEdge_t, mvfSize);
	  for(i=0; i< mvfSize; i++){
	    dataBAig[i]  = MvfAig_FunctionReadComponent(dataMvfAig,  i);
	    latchBAig[i] = MvfAig_FunctionReadComponent(latchMvfAig, i);
	  }
	  BmcGenerateCnfFromStateTostate(manager, dataBAig, latchBAig, mvfSize,
					 k, l, cnfClauses, loop);
	  FREE(dataBAig);
	  FREE(latchBAig);
	} /* ForEachLatch loop*/  
	array_insert(int, loopClause, l, loop);
	if(rightValue == -1){
	  andIndex   = cnfClauses->cnfGlobalIndex++;
	  tmpclause  = array_alloc(int, 2);
	  array_insert(int, tmpclause, 0, loop);
	  array_insert(int, tmpclause, 1, -andIndex);
	  BmcCnfInsertClause(cnfClauses, tmpclause);

	  array_insert(int, tmpclause, 0, loopIndex);
	  array_insert(int, tmpclause, 1, -andIndex);
	  BmcCnfInsertClause(cnfClauses, tmpclause);
	  array_free(tmpclause);

	  array_insert_last(int, orClause, andIndex);
	} else {
	  array_insert_last(int, orClause, loop);
	}
      }
    } /* for l loop */
    if((leftValue == 1) && (rightValue == 1)){
      assert(k==1);
      if (options->verbosityLevel != BmcVerbosityNone_c){
	(void) fprintf(vis_stdout,"# BMC: formula failed\n");
	(void) fprintf(vis_stdout,"# BMC: Empty counterexample because the property is always FALSE\n");
      }
      break;
    } else if((leftValue == 0) && (rightValue == 0)){
      if (options->verbosityLevel != BmcVerbosityNone_c){
	(void) fprintf(vis_stdout,"# BMC: the formula is trivially true");
	(void) fprintf(vis_stdout," for counterexamples of length %d\n", k);
      }
    } else {
      BmcCnfInsertClause(cnfClauses, loopClause);
      BmcCnfInsertClause(cnfClauses, orClause);
      array_free(loopClause);
      array_free(orClause);
      if (options->verbosityLevel == BmcVerbosityMax_c) {      
	fprintf(vis_stdout,
		"Number of Variables = %d Number of Clauses = %d\n",
		cnfClauses->cnfGlobalIndex-1,  cnfClauses->noOfClauses);
      }
      BmcWriteClauses(manager, cnfFile, cnfClauses);
      (void) fclose(cnfFile);
      
      result = BmcCheckSAT(options);
      if(result != NIL(array_t)) {
	(void) fprintf(vis_stdout, "# BMC: formula failed\n");
	if (options->dbgLevel == 1) {
	  (void) fprintf(vis_stdout,
			 "# BMC: Found a counterexample of length = %d \n", k);
	  BmcPrintCounterExample(network, nodeToMvfAigTable, cnfClauses,
				 result, k, CoiTable, options);
	}
      }
    }
    /* free all used memories */
    BmcCnfClausesFree(cnfClauses);
    k += options->step;
  } /* while result and k*/

  if ((result == NIL(array_t)) && (k > maxK)){
    if (boundedFormula && depth <= options->maxK){
      (void) fprintf(vis_stdout,"# BMC: formula passed\n");
    } else {
      (void) fprintf(vis_stdout,"# BMC: formula undecided\n");
    }
    if (options->verbosityLevel != BmcVerbosityNone_c){
     (void) fprintf(vis_stdout,
		    "# BMC: no counterexample found of length up to %d \n",
		    maxK);
    }
  }
  /* free all used memories */
  if (k == 0){
     BmcCnfClausesFree(cnfClauses);
  }
  if(result != NIL(array_t)){
    array_free(result);
  }
  /*
  if (k!=0){
    printf("\nMaximum Counterexample length = %d \n", k-options->step);
  }
  */
  if (options->verbosityLevel != BmcVerbosityNone_c) {
    endTime = util_cpu_ctime();
    fprintf(vis_stdout, "-- bmc time = %10g\n",
	    (double)(endTime - startTime) / 1000.0);
  }
  Ctlsp_FormulaFree(ltlFormula);
  fflush(vis_stdout);
  return;
} /* Bmc_CheckLTLFormula() */

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Generate CNF for an LTL formula]

  Description [Generate array of clauses for a trnasition from state i to
              state k. If l is non-zero, it will also generate the clauses
	      from state l to state i.
	      
              if the clause array is empty {}, the propety is always TRUE
	      if the the clause array has one empty clause {{}}, the property
	      is always FALSE.
	      return the clause index number.
  ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
int
BmcGenerateCnfForLtl(
   Ntk_Network_t   *network,
   Ctlsp_Formula_t *ltlFormula,
   int             i /* from state i */,
   int             k /* to state k   */,
   int             l /* loop         */,
   BmcCnfClauses_t *cnfClauses)
{
  int left, right, cnfIndex;
  int andIndex, orIndex;
  int j, n;
  int leftValue, rightValue, andValue, orValue;
  
  mAig_Manager_t  *manager = Ntk_NetworkReadMAigManager(network);
  array_t         *clause, *tmpclause, *orClause, *rightClause, *leftClause;
  BmcCnfStates_t  *state;
  
  assert(ltlFormula != NIL(Ctlsp_Formula_t));
  right = 0;
  rightValue = 0;
  
  if(Ctlsp_isPropositionalFormula(ltlFormula)){
    bAigEdge_t property = BmcCreateMaigOfPropFormula(network, manager, ltlFormula);
       
    if (property == bAig_Zero){ /* FALSE */
      /* add an empty clause to indicate FALSE property */
      BmcAddEmptyClause(cnfClauses);
      return 0;
    }
    if (property == bAig_One){ /* TRUE */
      /* Don't generate any clause to indicate TRUE property.*/
      return 0;
    }
    /* Generate the clause of the propositional formula */
    /* The state of the input variables is the same as the state of the state variables. */
    cnfIndex = BmcGenerateCnfFormulaForAigFunction(manager, property, i, cnfClauses);
    return cnfIndex;
  }
  /*
   * Formula is NOT propositional formula 
   */
  switch(ltlFormula->type) {
    case Ctlsp_NOT_c:
      /* reach here if formula-left is always not propositional*/
      /* NOT must only appears infront of a propositional formula.*/
      if (!Ctlsp_isPropositionalFormula(ltlFormula->left)){
	fprintf(vis_stderr,"BMC error: the LTL formula is not in NNF\n");
	return 0;
      }
      /*
      return -1*BmcGenerateCnfForLtlNoLoop(network, ltlFormula->left, i, k,
					   cnfIndexTable, clauseArray);
      */
      break;
    case Ctlsp_AND_c:
      /*
	  c = a * b  -->  (!a + !b + c) * (a + !c) * (b + !c).
	  Because a and b must be one, so we need only the last two clauses.
       */
      
      state = BmcCnfClausesFreeze(cnfClauses);
      rightValue = 1;
      
      left      = BmcGenerateCnfForLtl(network, ltlFormula->left,  i, k, l, cnfClauses);
      leftValue = checkIndex(left, cnfClauses);

      if (leftValue !=0){
	right      = BmcGenerateCnfForLtl(network, ltlFormula->right, i, k, l, cnfClauses);
	rightValue = checkIndex(right, cnfClauses);
      }
      if ((leftValue == 0) || (rightValue == 0)){
	/* restore the information */
	BmcCnfClausesRestore(cnfClauses, state);
	/* add an empty clause*/
	BmcAddEmptyClause(cnfClauses);
	FREE(state);
	return 0;  /* FALSE */
      }
      if ((leftValue == 1) && (rightValue == 1)){
	/* restore the information */
	BmcCnfClausesRestore(cnfClauses, state);
	FREE(state);
	return 0; /* TRUE */
      }
      BmcCnfClausesUnFreeze(cnfClauses, state);
      FREE(state);
      /*
      tmpclause  = array_alloc(int, 3);
      array_insert(int, tmpclause, 0, -left);
      array_insert(int, tmpclause, 1, -right);
      array_insert(int, tmpclause, 2,  cnfIndex);
      array_insert_last(array_t *, clauseArray, tmpclause);
      */
      if (leftValue == 1){
	return right;
      }
      if (rightValue == 1){
	return left;
      }
      cnfIndex = cnfClauses->cnfGlobalIndex++;      
      tmpclause  = array_alloc(int, 2);
      
      array_insert(int, tmpclause, 0, left);
      array_insert(int, tmpclause, 1, -cnfIndex);
      BmcCnfInsertClause(cnfClauses, tmpclause);
      
      array_insert(int, tmpclause, 0, right);
      array_insert(int, tmpclause, 1, -cnfIndex);
      BmcCnfInsertClause(cnfClauses, tmpclause);
      array_free(tmpclause);

      return cnfIndex;
    case Ctlsp_OR_c:
      /*
	  c = a + b  -->  (a + b + !c) * (!a + c) * (!b + c).
	  Because a and b must be one, so we need only the first clause.
       */
      state = BmcCnfClausesFreeze(cnfClauses);
            
      left      = BmcGenerateCnfForLtl(network, ltlFormula->left,  i, k, l, cnfClauses);
      leftValue = checkIndex(left, cnfClauses);

      if (leftValue !=1){
	right      = BmcGenerateCnfForLtl(network, ltlFormula->right, i, k, l, cnfClauses);
	rightValue = checkIndex(right, cnfClauses);
      }
      if ((leftValue == 1) || (rightValue == 1)){
	/* restore the information */
	BmcCnfClausesRestore(cnfClauses, state);
	FREE(state);
	return 0;  /* TRUE */
      } 
      if ((leftValue == 0) && (rightValue == 0)){
	/* restore the information */
	BmcCnfClausesRestore(cnfClauses, state);
	/* add an empty clause*/
	BmcAddEmptyClause(cnfClauses);
	FREE(state);
	return 0;  /* FALSE */
      }
      BmcCnfClausesUnFreeze(cnfClauses, state);
      FREE(state);
      /* Either leftValue or rightValue = 0 and the other != 1
         At least one clause will be added
       */
      if (leftValue == 0){
	return right;
      }
      if (rightValue == 0){
	return left;
      }      
      cnfIndex = cnfClauses->cnfGlobalIndex++;

      tmpclause  = array_alloc(int, 0);
      array_insert_last(int, tmpclause, -cnfIndex);
      
      array_insert_last(int, tmpclause, left);
	/*
	tmpclause  = array_alloc(int, 2);
	array_insert(int, tmpclause, 0, -left);
	array_insert(int, tmpclause, 1, cnfIndex);
	array_insert_last(array_t *, clauseArray, tmpclause);
	*/
      array_insert_last(int, tmpclause, right);
	/*
	tmpclause  = array_alloc(int, 2);
	array_insert(int, tmpclause, 0, -right);
	array_insert(int, tmpclause, 1, cnfIndex);
	array_insert_last(array_t *, clauseArray, tmpclause);
	*/
      BmcCnfInsertClause(cnfClauses, tmpclause);
      array_free(tmpclause);
      
      return cnfIndex;
    case Ctlsp_F_c:
      if (l >= 0){ /* loop */
	i = (l < i)? l: i;
      }
      cnfIndex = cnfClauses->cnfGlobalIndex++;
      clause   = array_alloc(int, 0);
      for (j=i; j<= k; j++){
	left      = BmcGenerateCnfForLtl(network, ltlFormula->left, j, k, l, cnfClauses);
	leftValue = checkIndex(left, cnfClauses);
        if (leftValue != -1){
	  array_free(clause);
	  return 0;
	}
	array_insert_last(int, clause, left);
	  /*
	  tmpclause  = array_alloc(int, 2);
	  array_insert(int, tmpclause, 1, cnfIndex);
	  array_insert(int, tmpclause, 0, -left);
	  array_insert_last(array_t *, clauseArray, tmpclause);
	  */
      }
      array_insert_last(int, clause, -cnfIndex);
      BmcCnfInsertClause(cnfClauses, clause);
      array_free(clause);
 
      return cnfIndex;
    case Ctlsp_G_c:
      if (l < 0){ /* FALSE */
	/* add an empty clause */
	BmcAddEmptyClause(cnfClauses);
	return 0;
      } else {
       i = (l < i)? l: i;
       andIndex = cnfClauses->cnfGlobalIndex++;
       for (j=i; j<= k; j++){
	 left      = BmcGenerateCnfForLtl(network, ltlFormula->left, j, k, l, cnfClauses);
	 leftValue = checkIndex(left, cnfClauses);
         if (leftValue != -1){
	   return 0;
	 }
	 tmpclause  = array_alloc(int, 2);
	 array_insert(int, tmpclause, 0, left);
	 array_insert(int, tmpclause, 1, -andIndex);
	 BmcCnfInsertClause(cnfClauses, tmpclause);
	 array_free(tmpclause);
       }/* for j loop*/
      } /* else */
      return andIndex;
    case Ctlsp_X_c:
      /* X left */
      if((l >= 0) && (i == k) ){
	i = l;
      } else if (i < k){
	i = i + 1;
      } else { /* FALSE */
	/* add an empty clause */
	BmcAddEmptyClause(cnfClauses);
	return 0;
      }
      return BmcGenerateCnfForLtl(network, ltlFormula->left, i, k, l, cnfClauses);
    case Ctlsp_U_c: /* (left U right) */
      state = BmcCnfClausesFreeze(cnfClauses);
      
      leftValue  = 1; /* left is TRUE*/
      rightValue = 1; /* right is TRUE*/
      orIndex    = cnfClauses->cnfGlobalIndex++;
      orClause   = array_alloc(int, 0);
      array_insert_last(int, orClause, -orIndex);

      orValue = 0;
      for (j=i; j<= k; j++){
	right = BmcGenerateCnfForLtl(network, ltlFormula->right, j, k, l, cnfClauses);
	rightValue = checkIndex(right, cnfClauses);
	andIndex   = cnfClauses->cnfGlobalIndex++;
	if (rightValue == -1){
	  rightClause  = array_alloc(int, 2);
	  array_insert(int, rightClause, 0, right);
	  array_insert(int, rightClause, 1, -andIndex);
	  BmcCnfInsertClause(cnfClauses, rightClause);
	  array_free(rightClause);
	}
	andValue = rightValue;
	for (n=i; (n <= j-1) && (andValue != 0); n++){
	  left = BmcGenerateCnfForLtl(network, ltlFormula->left, n, k, l, cnfClauses);
	  leftValue = checkIndex(left, cnfClauses);
	  andValue  = doAnd(andValue, leftValue);
	  if (leftValue == -1){
	    leftClause  = array_alloc(int, 2);
	    array_insert(int, leftClause, 0, left);
	    array_insert(int, leftClause, 1, -andIndex);
	    BmcCnfInsertClause(cnfClauses, leftClause);
	    array_free(leftClause);
	  }
	} /* for n loop */
	orValue = doOr(orValue, andValue);
        if (orValue == 1){ /* TRUE */
	  break;
	}
	if (andValue != 0){
	  array_insert_last(int, orClause, andIndex);
	}
      } /* for j loop*/
      if ( (l >=0) && (orValue !=1)){ /* loop */
	for (j=l; j<= i-1; j++){
	  right = BmcGenerateCnfForLtl(network, ltlFormula->right, j, k, l, cnfClauses);
	  rightValue = checkIndex(right, cnfClauses);
	  andIndex   = cnfClauses->cnfGlobalIndex++;
	  if (rightValue == -1){
	    rightClause  = array_alloc(int, 2);
	    array_insert(int, rightClause, 0, right);
	    array_insert(int, rightClause, 1, -andIndex);
	    BmcCnfInsertClause(cnfClauses, rightClause);
	    array_free(rightClause);
	  }
	  andValue = rightValue;  
	  for (n=i; (n<= k) && (andValue != 0); n++){
	    left = BmcGenerateCnfForLtl(network, ltlFormula->left, n, k, l, cnfClauses);
	    leftValue = checkIndex(left, cnfClauses);
	    andValue  = doAnd(andValue, leftValue);
	    if (leftValue == -1){
	      leftClause  = array_alloc(int, 2);
	      array_insert(int, leftClause, 0, left);
	      array_insert(int, leftClause, 1, -andIndex);
	      BmcCnfInsertClause(cnfClauses, leftClause);
	      array_free(leftClause);
	    }
	  } /* for n loop */
	  for (n=l; (n<= j-1)  && (andValue != 0); n++){
	    left = BmcGenerateCnfForLtl(network, ltlFormula->left, n, k, l, cnfClauses);
	    leftValue = checkIndex(left, cnfClauses);
	    andValue  = doAnd(andValue, leftValue);
	    if (leftValue == -1){
	      leftClause  = array_alloc(int, 2);
	      array_insert(int, leftClause, 0, left);
	      array_insert(int, leftClause, 1, -andIndex);
	      BmcCnfInsertClause(cnfClauses, leftClause);
	      array_free(leftClause);
	    }
	  } /* for n loop */
	  orValue = doOr(orValue, andValue);
	  if (orValue == 1){ /* TRUE */
	    break;
	  }
	  if (andValue != 0){
	    array_insert_last(int, orClause, andIndex);
	  }
	}/* j loop*/
      } /* if (l>=0) */
      if ((orValue == 1) || (orValue == 0)){
	/*restore the infomration back*/
	BmcCnfClausesRestore(cnfClauses, state);
	FREE(state);
	array_free(orClause);
	if (orValue == 0){
	  /* add an empty clause*/
	  BmcAddEmptyClause(cnfClauses);
	}
	return 0;
      } else {
	BmcCnfClausesUnFreeze(cnfClauses, state);
	FREE(state);
	BmcCnfInsertClause(cnfClauses, orClause);
	array_free(orClause);
	return orIndex;
      }
    case Ctlsp_R_c:
      /* (left R right) */
      state = BmcCnfClausesFreeze(cnfClauses);
      
      orIndex  = cnfClauses->cnfGlobalIndex++;
      orClause = array_alloc(int, 0);
      array_insert_last(int, orClause, -orIndex);
      
      orValue = 0;
      if (l >= 0){ /* loop */
	andIndex = cnfClauses->cnfGlobalIndex++;
	andValue = 1;
	for (j=(i<l?i:l); (j<= k) && (andValue != 0); j++){
	  right = BmcGenerateCnfForLtl(network, ltlFormula->right, j, k, l, cnfClauses);
	  rightValue = checkIndex(right, cnfClauses);
	  andValue  = doAnd(andValue, rightValue);
	  if (rightValue == -1){
	    rightClause  = array_alloc(int, 2);
	    array_insert(int, rightClause, 0, right);
	    array_insert(int, rightClause, 1, -andIndex);
	    BmcCnfInsertClause(cnfClauses, rightClause);
	    array_free(rightClause);
	  }
	} /* for j loop*/
	if (andValue == -1){
	  array_insert_last(int, orClause, andIndex);
	}
	orValue = doOr(orValue, andValue);
      } /* loop */
      if(orValue != 1){
	for (j=i; j<= k; j++){
	  left = BmcGenerateCnfForLtl(network, ltlFormula->left,
				      j, k, l, cnfClauses);
	  leftValue = checkIndex(left, cnfClauses);
	  andIndex = cnfClauses->cnfGlobalIndex++;
	  if (leftValue == -1){
	    leftClause  = array_alloc(int, 2);
	    array_insert(int, leftClause, 0, left);
	    array_insert(int, leftClause, 1, -andIndex);
	    BmcCnfInsertClause(cnfClauses, leftClause);
	    array_free(leftClause);
	  }
	  andValue = leftValue;
	  for (n=i; (n<= j) && (andValue != 0); n++){
	    right = BmcGenerateCnfForLtl(network, ltlFormula->right,
					 n, k, l, cnfClauses);
	    rightValue = checkIndex(right, cnfClauses);
	    andValue   = doAnd(andValue, rightValue);
	    if (rightValue == -1){
	      rightClause  = array_alloc(int, 2);
	      array_insert(int, rightClause, 0, right);
	      array_insert(int, rightClause, 1, -andIndex);
	      BmcCnfInsertClause(cnfClauses, rightClause);
	      array_free(rightClause);
	    }
	  } /* for n loop */
	  orValue = doOr(orValue, andValue);
	  if (orValue == 1){ /* TRUE */
	    break;
	  }
	  if (andValue != 0){
	    array_insert_last(int, orClause, andIndex);
	  }
	}/* for j loop*/
	if ((l >= 0)  && (orValue !=1)) { /* loop */
	  for (j=l; j<= i-1; j++){
	    left = BmcGenerateCnfForLtl(network, ltlFormula->left,
					j, k, l, cnfClauses);
	    leftValue = checkIndex(left, cnfClauses);
	    andIndex = cnfClauses->cnfGlobalIndex++;
	    if (leftValue == -1){
	      leftClause  = array_alloc(int, 2);
	      array_insert(int, leftClause, 0, left);
	      array_insert(int, leftClause, 1, -andIndex);
	      BmcCnfInsertClause(cnfClauses, leftClause);
	      array_free(leftClause);
	    }
	    andValue = leftValue;
	    for (n=i; (n<= k)  && (andValue != 0); n++){
	      right = BmcGenerateCnfForLtl(network, ltlFormula->right,
					   n, k, l, cnfClauses);
	      rightValue = checkIndex(right, cnfClauses);
	      andValue = doAnd(andValue, rightValue);
	      if (rightValue == -1){
		rightClause  = array_alloc(int, 2);
		array_insert(int, rightClause, 0, right);
		array_insert(int, rightClause, 1, -andIndex);
		BmcCnfInsertClause(cnfClauses, rightClause);
		array_free(rightClause);
	      }
	    } /* for n loop */
	    for (n=l; (n<= j) && (andValue != 0); n++){
	      right = BmcGenerateCnfForLtl(network, ltlFormula->right,
					   n, k, l, cnfClauses);
	      rightValue = checkIndex(right, cnfClauses);
	      andValue = doAnd(andValue, rightValue);
	      if (rightValue == -1){
		rightClause  = array_alloc(int, 2);
		array_insert(int, rightClause, 0, right);
		array_insert(int, rightClause, 1, -andIndex);
		BmcCnfInsertClause(cnfClauses, rightClause);
		array_free(rightClause);
	      }
	    } /* for n loop */
	    orValue = doOr(orValue, andValue);
	    if (orValue == 1){ /* TRUE */
	      break;
	    }
	    if (andValue != 0){
	      array_insert_last(int, orClause, andIndex);
	    }
	  } /* for j loop*/
	} /* if (l>=0) */
      }/* orValue !=1*/
      if ((orValue == 1) || (orValue == 0)){
	/*restore the infomration back*/
	BmcCnfClausesRestore(cnfClauses, state);
	FREE(state);
	array_free(orClause);
	if (orValue == 0){
	  /* add an empty clause*/
	  BmcAddEmptyClause(cnfClauses);
	}
	return 0;
      } else {
	BmcCnfClausesUnFreeze(cnfClauses, state);
	FREE(state);
	BmcCnfInsertClause(cnfClauses, orClause);
	array_free(orClause);
	return orIndex;
      }
    default:
      fail("Unexpected LTL formula type");
      break;
  }
  return -1; /* we should never get here */
}
/*  BmcLTLFormulaNoLoopTranslation() */


/**Function********************************************************************

  Synopsis    [Generate CNF formula for unfolded Transition function of length  
               bound for F(p) property]

  Description [
           For a multi-valued latch of 4 values. Two binary variables are used 
           to rpresent X, x1 and x0. For this latch, there exist three multi-valued 
           functions.
           One for the binary reoresentation of the variable. For example the
           second entry of the mvf = 1, iff ~x1 and x0. 
           The second mfv is for the data input of the latch. If the And/Inv graph 
           attached to an entry of this mvf equal to 1, X equal to the binary 
           representation corresponding to this entry. For example, if the And/
           INV graph attached to the first entry =1, then X = ~x1 & ~x0. 
           To generate the CNF to the transition relation, first generate CNF to
           next state varaible using mvf of the latch.  Then, generate CNF for 
           latch data input using current state variavles. Finaly, generate CNF
           for the AND of these two MVF.  This for every entry of the MVF. Then 
           OR the results.
           The third MVF is for the initial value of the latch.  It is treat the
           same as the latch data input except if the initial value is constant. 
        
           ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
BmcGenerateCnfForAFformula(
   Ntk_Network_t   *network,
   Ctlsp_Formula_t *ltlFormula,
   BmcOption_t     *options,
   st_table        *CoiTable)
{
  mAig_Manager_t  *manager   = Ntk_NetworkReadMAigManager(network);
  st_table        *nodeToMvfAigTable = NIL(st_table);  /* maps each node to its mvfAig */
  lsGen           gen;
  Ntk_Node_t      *latch, *latchData;
  BmcCnfClauses_t *cnfClauses = NIL(BmcCnfClauses_t);
  MvfAig_Function_t  *dataMvfAig, *latchMvfAig;

  bAigEdge_t *latchBAig, *dataBAig;
  
  FILE       *cnfFile;
  array_t    *clause, *outClause;

  int        outIndex;
  int        i, k, mvfSize;
  array_t    *result  = NIL(array_t);
  long       startTime, endTime;
  bAigEdge_t property;
  int        bound;
  
  startTime = util_cpu_ctime();
  if (options->verbosityLevel != BmcVerbosityNone_c){
    (void)fprintf(vis_stdout, "LTL formula is of type F(p)\n");
    (void)fprintf(vis_stdout, "Apply BMC on counterexample of length >= %d and <= %d (+%d) \n",
	                 	                  options->minK, options->maxK, options->step);
  }
  property = BmcCreateMaigOfPropFormula(network, manager, ltlFormula->left);
  if (property == mAig_NULL){
    return;
  }

  property = bAig_Not(property);
  if (property == bAig_One){
    (void) fprintf(vis_stdout, "# BMC: formula failed\n");
    (void) fprintf(vis_stdout, "       Empty counterexample because the property is always FALSE\n");
    return;
  } else if (property == bAig_Zero){
    (void) fprintf(vis_stdout,"# BMC: The property is always TRUE\n");
    (void) fprintf(vis_stdout,"# BMC: formula passed\n");
    return;
  }    
  /* nodeToMvfAigTable Maps each node to its Multi-function And/Inv graph */
  nodeToMvfAigTable = (st_table *) Ntk_NetworkReadApplInfo(network, MVFAIG_NETWORK_APPL_KEY);
  if (nodeToMvfAigTable == NIL(st_table)){
    (void) fprintf(vis_stderr,
		   "** bmc error: please run buid_partiton_maigs first");
    return;
  } 
  bound = options->minK;
  while( (result == NIL(array_t)) && (bound <= options->maxK)){
    if (options->verbosityLevel == BmcVerbosityMax_c) {
      (void) fprintf(vis_stdout,
		     "\nBMC: Generate counterexample of length %d\n",
		     bound);
    }
    cnfClauses = BmcCnfClausesAlloc();

    cnfFile = Cmd_FileOpen(options->satInFile, "w", NIL(char *), 0);    
    if (cnfFile == NIL(FILE)) {
      (void)fprintf(vis_stderr,
		    "** bmc error: Cannot create CNF output file %s\n",
		    options->satInFile);
      return;
    }
    BmcGenerateCnfForPath(network, 0, bound, cnfClauses, nodeToMvfAigTable, CoiTable);
     
    /* Generate CNF for loop back from the final state to any previouse state */
    /* (Sk -> S0) + (Sk -> S1) + ..... + (Sk -> Sk-1) */
    /* Each transition consisits of one or more latches.  We AND the transiton
       relation of these latches.  For multi-valued latches, we OR the
       equivalence of each value of the latch. To do the AND we use the CNF
       equivalent of the AND.
       a = b*c =>  (b + !a) * (c + !a)
    */
    outClause = array_alloc(int, bound);   
    
    for (k=0; k <= bound; k++){
      /*
	Create new var for the output of the AND node.
      */
      outIndex = cnfClauses->cnfGlobalIndex++;
      
      Ntk_NetworkForEachLatch(network, gen, latch) {
	int tmp;
	/* skip the latch if it is not in COI of property*/
	if (!st_lookup_int(CoiTable, (char *) latch, &tmp)){
	  /*printf("Not in COI %s\n",  Ntk_NodeReadName(latch));*/
	  continue;
	}
	latchData  = Ntk_LatchReadDataInput(latch);
	
	dataMvfAig = BmcReadMvfAig(latchData, nodeToMvfAigTable);
	if (dataMvfAig ==  NIL(MvfAig_Function_t)){
	  (void) fprintf(vis_stdout,
			 "No multi-valued function for this node %s \n",
			 Ntk_NodeReadName(latchData));
	  return;
	}
	latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);
	if (latchMvfAig ==  NIL(MvfAig_Function_t)){
	  latchMvfAig = BmcNodeBuildMVF(network, latch);
	}
	mvfSize   = array_n(dataMvfAig);
	dataBAig  = ALLOC(bAigEdge_t, mvfSize);
	latchBAig = ALLOC(bAigEdge_t, mvfSize); 
	for(i=0; i< mvfSize; i++){
	  dataBAig[i]  = MvfAig_FunctionReadComponent(dataMvfAig,  i);
	  latchBAig[i] = MvfAig_FunctionReadComponent(latchMvfAig, i);
	}
	BmcGenerateCnfFromStateTostate(manager, dataBAig, latchBAig, mvfSize, bound, k, cnfClauses, outIndex);
	FREE(dataBAig);
	FREE(latchBAig); 
      } /* ForEachLatch loop*/  
      array_insert(int, outClause, k, outIndex);
    }  /* for k state loop */
    BmcCnfInsertClause(cnfClauses, outClause);
    array_free(outClause);
 
    /* Generate CNF for the F(p) property */
    for(k=0; k <= bound; k++){
      clause  = array_alloc(int, 1);
      array_insert(int, clause, 0,
		   BmcGenerateCnfFormulaForAigFunction(manager, property,
						       k, cnfClauses));
      BmcCnfInsertClause(cnfClauses, clause);
      array_free(clause);
    }
    if (options->verbosityLevel == BmcVerbosityMax_c) {
      fprintf(vis_stdout,
	      "Number of Variables = %d Number of Clauses = %d\n",
	      cnfClauses->cnfGlobalIndex-1, cnfClauses->noOfClauses);
    }
    BmcWriteClauses(manager, cnfFile, cnfClauses);
    (void) fclose(cnfFile);
    result = BmcCheckSAT(options);
    /* free all used memories */
    BmcCnfClausesFree(cnfClauses);
    
    bound += options->step;
  } /* while result and bound */
  if(result == NIL(array_t)) {
    (void) fprintf(vis_stdout,"# BMC: formula undecided\n");
    if (options->verbosityLevel != BmcVerbosityNone_c){
      (void) fprintf(vis_stdout,
		     "# BMC: no counterexample found of length up to %d \n",
		     options->maxK);
    }
  } else {
    (void) fprintf(vis_stdout, "# BMC: formula failed\n");
    if(options->dbgLevel == 1){
      (void) fprintf(vis_stdout,
		     "# BMC: found a counterexample of length %d\n", bound);
      BmcPrintCounterExample(network, nodeToMvfAigTable, cnfClauses,
			     result, bound, CoiTable, options);
      array_free(result);
    }
  }
  if (options->verbosityLevel != BmcVerbosityNone_c) {
    endTime = util_cpu_ctime();
    fprintf(vis_stdout, "-- bmc time = %10g\n",
	    (double)(endTime - startTime) / 1000.0);
  }
  fflush(vis_stdout);  
  return;
}

/**Function********************************************************************

  Synopsis    [Generate CNF formula for unfolded Transition function of length
               bound for G(p) property]

  Description [
           For a multi-valued latch of 4 vlaues. Two binary variables are used 
           to rpresent X, x1 and x0. For this latch, there exist three
	   multi-valued functions.
           One for the binary reoresentation of the variable. For example the
           second entry of the mvf = 1, iff ~x1 and x0. 
           The second mfv is for the data input of the latch. If the And/Inv
	   graph attached to an entry of this mvf equal to 1, X equal to the
	   binary representation corresponding to this entry. For example, if
	   the And/INV graph attached to the first entry =1, then
	   X = ~x1 & ~x0. 
           To generate the CNF to the transition relation, first generate CNF
	   to next state varaible using mvf of the latch.  Then, generate CNF
	   for latch data input using current state variavles. Finaly,
	   generate CNF for the AND of these two MVF.  This for every entry of
	   the MVF. Then OR the results.
           The third MVF is for the initial value of the latch.  It is treated
	   the same as the latch data input except if the initial value is
	   constant. 
           ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
BmcGenerateCnfForAGformula(
   Ntk_Network_t   *network,
   Ctlsp_Formula_t *ltlFormula,
   BmcOption_t     *options,
   st_table        *CoiTable)
{
  mAig_Manager_t    *manager = Ntk_NetworkReadMAigManager(network);
  st_table          *nodeToMvfAigTable = NIL(st_table);  /* maps each node to its mvfAig */
  BmcCnfClauses_t   *cnfClauses = NIL(BmcCnfClauses_t);
  BmcCnfStates_t    *state;

  FILE              *cnfFile;
  array_t           *Pclause = array_alloc(int, 0);

  int               k, bound, initK, propK;
  array_t           *result = NIL(array_t);
  long              startTime, endTime;
  bAigEdge_t        property;
  int               minK = options->minK;
  int               maxK = options->maxK;

  assert((Ctlsp_LtlFormulaIsGp(ltlFormula)) ||
	 (Ctlsp_isPropositionalFormula(ltlFormula)));
  startTime = util_cpu_ctime();
  if (Ctlsp_isPropositionalFormula(ltlFormula)){
     minK = 0;
     maxK = 0;
     if (options->verbosityLevel != BmcVerbosityNone_c){
       (void) fprintf(vis_stdout, "LTL formula is propositional\n");
     }
     property = BmcCreateMaigOfPropFormula(network, manager, ltlFormula);
  } else {
    if (options->verbosityLevel != BmcVerbosityNone_c){
      (void)fprintf(vis_stdout, "LTL formula is of type G(p)\n");
      (void)fprintf(vis_stdout, "Apply BMC on counterexample of length >= %d and <= %d (+%d)\n",
		    minK, maxK, options->step);
    }
    property = BmcCreateMaigOfPropFormula(network, manager, ltlFormula->left);
  }
  if (property == mAig_NULL){
    return;
  }
  property = bAig_Not(property);
  if (property == bAig_One){

    (void) fprintf(vis_stdout,"# BMC: Empty counterexample because the property is always FALSE\n");
    (void) fprintf(vis_stdout,"# BMC: formula failed\n");
    return;
  } else if (property == bAig_Zero){
    (void) fprintf(vis_stdout,"# BMC: The property is always TRUE\n");
    (void) fprintf(vis_stdout,"# BMC: formula passed\n");
    return;
  }
  initK = 0;
  propK = 0;
  cnfClauses = BmcCnfClausesAlloc();

  /* nodeToMvfAigTable Maps each node to its Multi-function And/Inv graph */
  nodeToMvfAigTable =
    (st_table *) Ntk_NetworkReadApplInfo(network, MVFAIG_NETWORK_APPL_KEY);
  if (nodeToMvfAigTable == NIL(st_table)){
    (void) fprintf(vis_stderr,
		   "** bmc error: please run buid_partiton_maigs first");
    return;
  }

  for(bound = minK; bound <= maxK; bound += options->step){
    if (options->verbosityLevel == BmcVerbosityMax_c) {
      (void) fprintf(vis_stdout,
		     "\nBMC: Generate counterexample of length %d\n",
		     bound);
    }
    cnfFile = Cmd_FileOpen(options->satInFile, "w", NIL(char *), 0);    
    if (cnfFile == NIL(FILE)) {
      (void) fprintf(vis_stderr,
		     "** bmc error: Cannot create CNF output file %s\n",
		     options->satInFile);
      return;
    }
    BmcGenerateCnfForPath(network, initK, bound, cnfClauses, nodeToMvfAigTable, CoiTable);
    /* Generate CNF for the property */
    for(k=propK; k <= bound; k++){
      array_insert_last(
	int, Pclause, BmcGenerateCnfFormulaForAigFunction(manager, property,
							  k, cnfClauses));
    }
    if (options->verbosityLevel == BmcVerbosityMax_c) {
      (void) fprintf(vis_stdout,
		     "Number of Variables = %d Number of Clauses = %d\n",
		     cnfClauses->cnfGlobalIndex-1, cnfClauses->noOfClauses);
    }
    state = BmcCnfClausesFreeze(cnfClauses);
    BmcCnfInsertClause(cnfClauses, Pclause);
    BmcWriteClauses(manager, cnfFile, cnfClauses);
    (void) fclose(cnfFile);
    result = BmcCheckSAT(options);
    if (result != NIL(array_t)){
      bound++;
      break;  /* formula is satisfiable */
    }
    BmcCnfClausesRestore(cnfClauses, state);
    FREE(state);
    initK = bound;
    propK = bound+1;
  } /* for bound loop */
  if(result == NIL(array_t)) {
    if (options->verbosityLevel != BmcVerbosityNone_c){
      (void) fprintf(vis_stdout,
		     "# BMC: no counterexample found of length up to %d\n",
		     maxK);
    }
    if (Ctlsp_isPropositionalFormula(ltlFormula)){
      (void) fprintf(vis_stdout,"# BMC: formula passed\n");
    } else {
      (void) fprintf(vis_stdout,"# BMC: formula undecided\n");
    }
  } else {
    (void) fprintf(vis_stdout, "# BMC: formula failed\n");
    if(options->dbgLevel == 1){
      (void) fprintf(vis_stdout,
		     "# BMC: found a counterexample of length %d\n",
		     bound-1);
      BmcPrintCounterExample(network, nodeToMvfAigTable, cnfClauses,
			     result, bound-1, CoiTable, options);
      array_free(result);
    }
  }
  /* free all used memories */
  BmcCnfClausesFree(cnfClauses);
  array_free(Pclause);
  if (options->verbosityLevel != BmcVerbosityNone_c){
    endTime = util_cpu_ctime();
    fprintf(vis_stdout, "-- bmc time = %10g\n",
	    (double)(endTime - startTime) / 1000.0);
  }
  fflush(vis_stdout);
  return;
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/



/**Function********************************************************************

  Synopsis    [Check if the index of the varaible in CNF is TURE, FALSE, or
               none]

  Description []

  SideEffects []

******************************************************************************/
static int
checkIndex(
  int             index,
  BmcCnfClauses_t *cnfClauses)
{
  int     rtnValue = -1; /* it is not TRUE or FALSE*/

  if (index == 0){ /* TRUE or FALSE*/
    if (cnfClauses->noOfClauses == 0){ /* empty clauses*/
      return 1; /* TRUE */
    }
    if (cnfClauses->emptyClause){   /* last added clause was empty = FALSE*/
      rtnValue = 0; /* FALSE */
    } else {       
      rtnValue = 1;
    }
  }
  return rtnValue;
}

/**Function********************************************************************

  Synopsis    []

  Description [
                0  1  -1
	       ----------	
	     0  0  0   0
	     1  0  1  -1
	    -1  0  -1 -1 

  ]

  SideEffects []

******************************************************************************/
static int
doAnd(
  int left,
  int right)
{
  if ((left == -1) && (right == -1)){
    return -1;
  }
  return (left * right);
  
} /* doAnd */

/**Function********************************************************************

  Synopsis    []

  Description [
                 0   1  -1
		----------- 
	     0   0   1  -1
	     1   1   1  -1
	    -1  -1  -1  -1 

  ]

  SideEffects []

******************************************************************************/
static int
doOr(
  int left,
  int right)
{
  if ((left == -1) || (right == -1)){
    return -1;
  }
  if ((left == 1) || (right == 1)){
    return 1;
  }
  return 0;
  
} /* doOr */
