/**CFile***********************************************************************

  FileName    [bmcCmd.c]

  PackageName [bmc]

  Synopsis    [Command interface for bounded model checking (bmc)]

  Author      [Mohammad Awedh]

  Copyright   [This file was created at the University of Colorado at
  Boulder.  The University of Colorado at Boulder makes no warranty
  about the suitability of this software for any purpose.  It is
  presented on an AS IS basis.]
******************************************************************************/

#include "bmcInt.h"

static char rcsid[] UNUSED = "$Id: bmcCmd.c,v 1.35 2002/09/21 07:48:20 fabio Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
static jmp_buf timeOutEnv;
static int bmcTimeOut;		/* timeout value */
static long alarmLapTime;	/* starting CPU time for timeout */

/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static BmcOption_t * ParseBmcOptions(int argc, char **argv);
static int CommandBmc(Hrc_Manager_t ** hmgr, int argc, char ** argv);
static void TimeOutHandle(void);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Initializes the BMC package.]

  SideEffects []

  SeeAlso     [Bmc_End]

******************************************************************************/
void
Bmc_Init(void)
{
  Cmd_CommandAdd("bounded_model_check", CommandBmc, /* doesn't change network */  0); 
}

/**Function********************************************************************

  Synopsis    [Ends BMC package.]

  SideEffects []

  SeeAlso     [Bmc_Init]

******************************************************************************/
void
Bmc_End(void)
{
}

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis [Parse the user input for command "bmc".]

  Description []

  SideEffects []

******************************************************************************/
static BmcOption_t *
ParseBmcOptions(
  int argc,
  char **argv)
{
  BmcOption_t  *options = BmcOptionAlloc();
  char *ltlFileName     = NIL(char);
  unsigned int i;
  int          c;
  
  if (!options){
    return NIL(BmcOption_t);
  }
  /*
   * Parse command line options.
   */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "F:hiv:m:k:s:o:t:d:")) != EOF) {
    switch(c) {
      case 'F':
	options->fairFile = Cmd_FileOpen(util_optarg, "r", NIL(char *), 0);
	if (options->fairFile == NIL(FILE)) {
	  (void) fprintf(vis_stdout,
			 "** bmc error: Cannot open the file %s\n",
			 util_optarg);
	  BmcOptionFree(options);
	  return NIL(BmcOption_t);
	}
        break;
      case 'm':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
        options->minK = atoi(util_optarg);
        break;
      case 'k':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
        options->maxK = atoi(util_optarg);
        break;
      case 's':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
        options->step = atoi(util_optarg);
        break;
      case 'd':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
        options->dbgLevel = atoi(util_optarg);
        break;	
      case 't' :
	options->timeOutPeriod = atoi(util_optarg);
	break;
      case 'i':
        options->printInputs = TRUE;
	break;	
      case 'h':
	goto usage;
      case 'v':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
	options->verbosityLevel = (Bmc_VerbosityLevel) atoi(util_optarg);
	break;
      case 'o':
	options->cnfFileName = util_strsav(util_optarg);
	break;	
      default:
	goto usage;
    }
  }
  if (options->minK > options->maxK){
    (void) fprintf(vis_stderr, "** bmc error: value for -m option must not be greater than vlaue for -k option\n");    
    goto usage;
  }
  /*
   * Open the ltl file.
   */
  if (argc - util_optind == 0) {
    (void) fprintf(vis_stderr, "** bmc error: file containing ltl formulae not provided\n");
    goto usage;
  }
  else if (argc - util_optind > 1) {
    (void) fprintf(vis_stderr, "** bmc error: too many arguments\n");
    goto usage;
  }
  ltlFileName = util_strsav(argv[util_optind]);
  
  /* Read LTL Formulae */
  if (!ltlFileName) 
      goto usage;

  options->ltlFile = Cmd_FileOpen(ltlFileName, "r", NIL(char *), 0);
  if (options->ltlFile == NIL(FILE)) {
    (void) fprintf(vis_stdout,"** bmc error: Cannot open the file %s\n", ltlFileName);
    FREE(ltlFileName);
    BmcOptionFree(options);
    return NIL(BmcOption_t);
  }
  FREE(ltlFileName);
  /* create SAT Solver input file */
  if (options->cnfFileName == NIL(char)) {
    options->satInFile = BmcCreateTmpFile(); 
    if (options->satInFile == NIL(char)){
      BmcOptionFree(options);
      return NIL(BmcOption_t);
    }
  } else {
    options->satInFile = options->cnfFileName;
  }
  /* create SAT Solver output file */
  options->satOutFile = BmcCreateTmpFile();
  if (options->satOutFile == NIL(char)){
    BmcOptionFree(options);
    return NIL(BmcOption_t);
  }
  return options;
  
 usage:
  (void) fprintf(vis_stderr, "usage: bmc [-h][-m minimum_length][-k maximum_length][-s step][-F <fairness_file>][-d debug level][-v verbosity_level][-t period][-o <cnf_file>] <ltl_file>\n");
  (void) fprintf(vis_stderr, "   -h \tprint the command usage\n");
  (void) fprintf(vis_stderr, "   -m \tminimum length of counterexample to be checked (default is 0)\n");  
  (void) fprintf(vis_stderr, "   -k \tmaximum length of counterexample to be checked (default is 1)\n");
  (void) fprintf(vis_stderr, "   -s \tincrementing value of counterexample length (default is 1)\n");
  (void) fprintf(vis_stderr, "   -F <fairness_file>\n");
  (void) fprintf(vis_stderr, "   -d <debug_level>\n");
  (void) fprintf(vis_stderr, "       debug_level = 0 => No debugging performed (Default)\n");
  (void) fprintf(vis_stderr, "       debug_level = 1 => Debugging with minimal output: generate counter-example if the LTL formula fails and the counterexample length.\n");
  (void) fprintf(vis_stderr, "   -i \tPrint input values in debug traces.\n");
  (void) fprintf(vis_stderr, "   -v <verbosity_level>\n");
  (void) fprintf(vis_stderr, "       verbosity_level = 0 => no feedback (Default)\n");
  (void) fprintf(vis_stderr, "       verbosity_level = 1 => code status\n");
  (void) fprintf(vis_stderr, "       verbosity_level = 2 => code status and CPU usage profile\n");
  (void) fprintf(vis_stderr, "   -t <period> time out period\n");
  (void) fprintf(vis_stderr, "   -o <cnf_file> contains CNF of the counterexample\n");
  (void) fprintf(vis_stderr, "   <ltl_file> The input file containing LTL formulae to be checked.\n");

  BmcOptionFree(options);
  return NIL(BmcOption_t);
}

/**Function********************************************************************

  Synopsis [Alloc Memory for BmcOption_t.]

  SideEffects []

  SeeAlso []
******************************************************************************/
BmcOption_t *
BmcOptionAlloc(void)
{
  BmcOption_t *result = ALLOC(BmcOption_t, 1);
  
  if (!result){
    return result;
  }
  result->ltlFile = NIL(FILE);
  result->fairFile = NIL(FILE);
  result->cnfFileName = NIL(char);
  result->minK = 0;
  result->maxK = 1;
  result->step = 1;
  result->timeOutPeriod  = 0;
  result->startTime = 0;
  result->verbosityLevel = BmcVerbosityNone_c;
  result->dbgLevel       = 0;
  result->printInputs    = FALSE;
  result->satInFile  = NIL(char);
  result->satOutFile = NIL(char);
  return result;
}

/**Function********************************************************************

  Synopsis [Free BmcOption_t, and close the ltl formula file.]

  SideEffects []

  SeeAlso []
******************************************************************************/
void
BmcOptionFree(
  BmcOption_t *result)
{
  if (result->ltlFile != NIL(FILE)){
    fclose(result->ltlFile);
  }
  if (result->cnfFileName != NIL(char)){
    FREE(result->cnfFileName);
  } else {
    unlink(result->satInFile);
    FREE(result->satInFile);
  }
  unlink(result->satOutFile);
  FREE(result->satOutFile);
  FREE(result);
}

/**Function********************************************************************

  Synopsis    [Create tmp file]

  Description []

  SideEffects []

******************************************************************************/
char *
BmcCreateTmpFile(void)
{
#if HAVE_MKSTEMP && HAVE_CLOSE  
  char   cnfFileName[] = "/tmp/vis.XXXXXX";
  int    fd;
#else
  char   *cnfFileName;
  char   buffer[512];
#endif

#if HAVE_MKSTEMP && HAVE_CLOSE    
  fd = mkstemp(cnfFileName);
  if (fd == -1){
#else
  cnfFileName = util_strsav(tmpnam(buffer));
  if (cnfFileName == NIL(char)){
#endif       
    (void)fprintf(vis_stderr,"** bmc error: Can not create temporary file. ");
    (void)fprintf(vis_stderr,"Clean up /tmp and try again.\n");
    return NIL(char);
  }
#if HAVE_MKSTEMP && HAVE_CLOSE
  close(fd);
#endif   
  return util_strsav(cnfFileName);
} /* createTmpFile() */

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis    [Implements the BMC command.]

  Description []
  
  CommandName [bounded_model_check]

  CommandSynopsis [performs an LTL bounded model checking on a flattened network]

  CommandArguments [\[-h\] \[-v &lt;verbosity_level&gt;\]  \[-d &lt;dbg_level&gt;\] \[-i\] 
                    -m &lt;minimum_length&gt; -k &lt;maximum_length&gt; -s &lt;step_Value&gt;
		    -o &lt;cnf_file&gt; &lt;ltl_file&gt; ]

		    -F &lt;fairness_file&gt; -o &lt;cnf_file&gt; &lt;ltl_file&gt; ]

  CommandDescription [Performs an LTL bounded model checking on a flattened
  network. This command looks for a counterexample of length &ge; minimum_length
  and &le; maximum_length. It increments the length by step_Value. 
  
  Before calling this command, the user should have initialized the design
  by calling the command <A HREF="flatten_hierarchyCmd.html"> <code>
  flatten_hierarchy</code></A>, and
  then calling the command <A HREF="build_partition_maigsCmd.html"> <code>
  build_partition_maigs</code></A>.<p>
  The value of maximum length must be >= minimum length.<p>

  Command options:
  <p>

  <dl>

  <dt> -h
  <dd> Print the command usage.
  <p>

  <dt>-m <code>&lt;minimum_length&gt;</code>
  <dd>Minimum length of counterexample to be checked (default is 0). 

  <dt>-k <code>&lt;maximum_length&gt;</code>
  <dd>Maximum length of counterexample to be checked (default is 1).

  <dt>-s <code>&lt;step_Value&gt;</code>
  <dd>Incrementing value of counterexample length (default is 1)\n");

  <dt>-F <code>&lt;fairness_file&gt;</code>
  <dd> Read fairness constraints from <code>&lt;fairness_file&gt;</code>.  Each
  formula in the file is a condition that must hold infinitely often on a fair
  path.

  <dt>-o <code>&lt;cnf_file&gt;</code>
  Save the CNF formula in <code>&lt;cnf_file&gt;</code>
  
  <dt> -t  <code>&lt;timeOutPeriod&gt;</code>
  <dd> Specify the time out period (in seconds) after which the command aborts. By default this option is set to infinity.
  <p>
  <dt> -v  <code>&lt;verbosity_level&gt;</code>
  <dd> Specify verbosity level. This sets the amount of feedback  on CPU usage and code status.

  <br><code>verbosity_level</code>  must be one of the following:<p>
  
  <code>0</code>: No feedback provided. This is the default.<p>

  <code>1</code>: Feedback on code location.<p>

  <code>2</code>: Feedback on code location and CPU usage.<p>

  <dt> -d <code> &lt;dbg_level&gt; </code> <dd> Specify the amount of
  debugging performed when the BMC models check the LTL formula. 
  <p> <dd> <code> dbg_level</code> must be one
  of the following:

  <p><code>0</code>: No debugging performed.
  <code>dbg_level</code>=<code>0</code> is the default.

  <p><code>1</code>: Debugging with minimal output: generate counter-example
  if the LTL formula fails and the counterexample length.

  <p>
  
  <dt> -h
  <dd> Print the command usage.
  <p>

  <dt> -i
  <dd> Print input values causing transitions between states during debugging.
  <p>

  <dt> -o <code>&lt;cnf_file&gt</code>;
  <dd> File containing CNF of the counterexample if exist. If specifies, the
  user can exam this file looking for information about the generated path.
  This file will be the input to the SAT solver.  The contents of this file
  is in dimacs format.
  
  <dt> <code> &lt;ltl_file&gt; </code>
  <dd> File containing LTL formulas to be model checked.
  
  </dl>
  ]

  SideEffects []

  SeeAlso     [Ntk_NetworkAddApplInfo]

******************************************************************************/
static int
CommandBmc(
  Hrc_Manager_t ** hmgr,
  int              argc,
  char          ** argv)
{
  Ntk_Network_t     *network;
  BmcOption_t       *options;
  int               i;
  array_t           *formulaArray;
  array_t           *LTLformulaArray;
  bAig_Manager_t    *manager;
  
  /*
   * Parse command line options.
   */
  if ((options = ParseBmcOptions(argc, argv)) == NIL(BmcOption_t)) {
      return 1;
  }
  /* Read the network */
  network = Ntk_HrcManagerReadCurrentNetwork(*hmgr);
  if (network == NIL(Ntk_Network_t)) {
    (void) fprintf(vis_stdout, "** bmc error: No network\n");
    BmcOptionFree(options);
    return 1;
  }
  manager = Ntk_NetworkReadMAigManager(network);
  if (manager == NIL(mAig_Manager_t)) {
    (void) fprintf(vis_stdout, "** bmc error: run build_partition_maigs command first\n");
    BmcOptionFree(options);
    return 1;
  }
  formulaArray = Ctlsp_FileParseFormulaArray(options->ltlFile);
  if (formulaArray == NIL(array_t)) {
    (void) fprintf(vis_stderr,
		   "** bmc error: Error in parsing CTL* Fromula from file\n");
    BmcOptionFree(options);
    return 1;
  }
  if (array_n(formulaArray) == 0) {
    (void) fprintf(vis_stderr, "** bmc error: No formula in file\n");
    BmcOptionFree(options);
    array_free(formulaArray);
    return 1;
  }
  if (Ctlsp_FormulaArrayAddLtlFairnessConstraints(formulaArray,
						  options->fairFile) != 0) {
    if (options->fairFile != NIL(FILE)) fclose(options->fairFile);
    BmcOptionFree(options);
    Ctlsp_FormulaArrayFree(formulaArray);
    return 1;
  }
  if (options->fairFile != NIL(FILE)) fclose(options->fairFile);
  /* time out */
  if (options->timeOutPeriod > 0) {
    /* Set the static variables used by the signal handler. */
    bmcTimeOut = options->timeOutPeriod;
    alarmLapTime = options->startTime = util_cpu_ctime();
    (void) signal(SIGALRM, (void(*)(int))TimeOutHandle);
    (void) alarm(options->timeOutPeriod);
    if (setjmp(timeOutEnv) > 0) {
      BmcOptionFree(options);
      (void) fprintf(vis_stdout,
		     "\n# BMC: timeout occurred after %d seconds.\n",
		     options->timeOutPeriod);
      (void) fprintf(vis_stdout, "# BMC: data may be corrupted.\n");
      alarm(0);
      return 1;
    }
  }
  LTLformulaArray = Ctlsp_FormulaArrayConvertToLTL(formulaArray);
  if (LTLformulaArray !=  NIL(array_t)){
    for (i = 0; i < array_n(LTLformulaArray); i++) { 
      Ctlsp_Formula_t *ltlFormula = array_fetch(Ctlsp_Formula_t *, LTLformulaArray, i);
      /*
      Ctlsp_Formula_t *saveLtlFormula = ltlFormula;
      */
      st_table        *CoiTable       = st_init_table(st_ptrcmp, st_ptrhash);
      
      BmcGetCoiForLtlFormula(network, ltlFormula, CoiTable);
      /* print out the given LTL formula */
      fprintf(vis_stdout, "Formula: ");
      Ctlsp_FormulaPrint(vis_stdout, ltlFormula);
      fprintf(vis_stdout, "\n");
      /* rewrite the ltl formula*/
      /*
      ltlFormula = Ctlsp_LtlFormulaSimpleRewriting(ltlFormula);
      Ctlsp_FormulaFree(saveLtlFormula);
      Ctlsp_LtlSetClass(ltlFormula);
      */
      
      if((Ctlsp_LtlFormulaIsGp(ltlFormula)) ||
	 (Ctlsp_isPropositionalFormula(ltlFormula))){
	BmcGenerateCnfForAGformula(network, ltlFormula, options, CoiTable);
	/*
	Bmc_CheckLTLFormula(network, ltlFormula, options);
	*/
      } else if (Ctlsp_LtlFormulaIsFp(ltlFormula)){
	BmcGenerateCnfForAFformula(network, ltlFormula, options, CoiTable);
	/*
	Bmc_CheckLTLFormula(network, ltlFormula, options);
	*/
      } else {
	Bmc_CheckLTLFormula(network, ltlFormula, options, CoiTable);
      }
      st_free_table(CoiTable);
    }
  } else {
   (void) fprintf(vis_stdout, "** bmc error: Invalid LTL formula\n");
   Ctlsp_FormulaArrayFree(formulaArray);
   BmcOptionFree(options);
   alarm(0);
   return 1;
  }

  Ctlsp_FormulaArrayFree(LTLformulaArray);
  Ctlsp_FormulaArrayFree(formulaArray);
  fflush(vis_stdout);
  fflush(vis_stderr);
  BmcOptionFree(options);
  alarm(0);
  return 0; /* normal exit */
}/* CommandBmc() */

/**Function********************************************************************

  Synopsis    [Handle function for timeout.]

  Description [This function is called when the time out occurs.
  Since alarm is set with an elapsed time, this function checks if the
  CPU time corresponds to the timeout of the command.  If not, it
  reprograms the alarm to fire again later.]

  SideEffects []

******************************************************************************/
static void
TimeOutHandle(void)
{
  int seconds = (int) ((util_cpu_ctime() - alarmLapTime) / 1000);

  if (seconds < bmcTimeOut) {
    unsigned slack = (unsigned) (bmcTimeOut - seconds);
    (void) signal(SIGALRM, (void(*)(int)) TimeOutHandle);
    (void) alarm(slack);
  } else {
    longjmp(timeOutEnv, 1);
  }
} /* TimeOutHandle */
