/**CHeaderFile*****************************************************************

  FileName    [bmcInt.h]

  PackageName [bmc]

  Synopsis    [Internal header declarations.]

  Author      [Mohammad Awedh]

  Copyright   [This file was created at the University of Colorado at Boulder.
  The University of Colorado at Boulder makes no warranty about the suitability
  of this software for any purpose.  It is presented on an AS IS basis.]

  Revision    [$Id: bmcInt.h,v 1.19 2002/09/21 07:48:11 fabio Exp $]

******************************************************************************/

#ifndef _BMCINT
#define _BMCINT

/*---------------------------------------------------------------------------*/
/* Nested includes                                                           */
/*---------------------------------------------------------------------------*/

#include "cmd.h"
#include "bmc.h"
#include "baigInt.h"

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef enum {
  BmcVerbosityNone_c,
  BmcVerbositySome_c,
  BmcVerbosityMax_c
} Bmc_VerbosityLevel;

/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/

/**Struct**********************************************************************

  Synopsis    [A struct for keeping the command line options for bmc command.]

******************************************************************************/
struct BmcOption {
  FILE               *ltlFile;  /* contains the LTL formulae */
  FILE               *fairFile; /* contains the fairness constraints */
  char               *cnfFileName; /* File contains CNF clauses if -o option
				      is used */
  Bmc_VerbosityLevel verbosityLevel;
  int                minK;
  int                maxK;
  int                step;
  int                timeOutPeriod;
  long               startTime; /* initial CPU time */
  int                rewriting;
  int                dbgLevel;
  boolean            printInputs;
  /* The folowing two pointers are used to point to the two tmp files that will
     be created in /tmp/ before calling the SAT solver. */
  char               *satInFile; 
  char               *satOutFile;
};

/**Struct**********************************************************************

  Synopsis    [A struct to store CNF clauses]

******************************************************************************/
struct BmcCnfClauses {
  array_t   *clauseArray;    /* array to store CNF clauses. Each clause is
			        terminated by the value 0.                        */
  st_table  *cnfIndexTable;  /* store the index of the key (name, state) in
			        the clauseArray. It uses to return the index of
			        of already generated clauses.                     */
  array_t   *freezedKeys;    /* When CNF is freezed, any new added key to
				cnfIndexTable will also be added to this
				array. When CNF restored, these keys in this
				array will be deleted from cnfIndexTable          */
  int       nextIndex;      /* holds the index of the next entry in clauseArray.  */
  int       noOfClauses;    /* total number of clauses */
  int       cnfGlobalIndex; /* the index of next variable.                        */
  boolean   emptyClause;    /* equal TRUE if the last added clause was an empty
			       clause.                                           */
  boolean   freezed;        /* TRUE if CNF is freezed */
};

/**Struct**********************************************************************

  Synopsis    [A struct to store CNF state. When CNF is freezed, a new state
               is created. This structure keeps the value of the varaible in
	       the current state.]

******************************************************************************/
struct BmcCnfStates {
  int       nextIndex;      
  int       noOfClauses;   
  int       cnfGlobalIndex;
  int       freezedKeySize;
  boolean   emptyClause;
  boolean   freezed;
};


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

EXTERN int BmcGenerateCnfForLtl(Ntk_Network_t *network, Ctlsp_Formula_t *ltlFormula, int i, int k, int l, BmcCnfClauses_t *cnfClauses);
EXTERN void BmcGenerateCnfForAFformula(Ntk_Network_t *network, Ctlsp_Formula_t *ltlFormula, BmcOption_t *options, st_table *CoiTable);
EXTERN void BmcGenerateCnfForAGformula(Ntk_Network_t *network, Ctlsp_Formula_t *ltlFormula, BmcOption_t *options, st_table *CoiTable);
EXTERN BmcOption_t * BmcOptionAlloc(void);
EXTERN void BmcOptionFree(BmcOption_t *result);
EXTERN char * BmcCreateTmpFile(void);
EXTERN MvfAig_Function_t * BmcReadMvfAig(Ntk_Node_t * node, st_table * nodeToMvfAigTable);
EXTERN mAigEdge_t BmcCreateMaigOfPropFormula(Ntk_Network_t *network, mAig_Manager_t *manager, Ctlsp_Formula_t *formula);
EXTERN int BmcGenerateCnfFormulaForAigFunction(bAig_Manager_t *manager, bAigEdge_t node, int k, BmcCnfClauses_t *cnfClauses);
EXTERN void BmcGenerateCnfFromStateTostate(bAig_Manager_t *manager, bAigEdge_t *fromAigArray, bAigEdge_t *toAigArray, int mvfSize, int fromState, int toState, BmcCnfClauses_t *cnfClauses, int outIndex);
EXTERN void BmcWriteClauses(mAig_Manager_t *maigManager, FILE *cnfFile, BmcCnfClauses_t *cnfClauses);
EXTERN array_t * BmcCheckSAT(BmcOption_t *options);
EXTERN MvfAig_Function_t * BmcNodeBuildMVF(Ntk_Network_t *network, Ntk_Node_t *node);
EXTERN void BmcPrintCounterExample(Ntk_Network_t *network, st_table *nodeToMvfAigTable, BmcCnfClauses_t *cnfClauses, array_t *result, int length, st_table *CoiTable, BmcOption_t *options);
EXTERN void BmcGenerateCnfForPath(Ntk_Network_t *network, int from, int to, BmcCnfClauses_t *cnfClauses, st_table *nodeToMvfAigTable, st_table *CoiTable);
EXTERN BmcCnfClauses_t * BmcCnfClausesAlloc(void);
EXTERN void BmcCnfClausesFree(BmcCnfClauses_t *cnfClauses);
EXTERN BmcCnfStates_t * BmcCnfClausesFreeze(BmcCnfClauses_t * cnfClauses);
EXTERN void BmcCnfClausesUnFreeze(BmcCnfClauses_t *cnfClauses, BmcCnfStates_t *state);
EXTERN void BmcCnfClausesRestore(BmcCnfClauses_t *cnfClauses, BmcCnfStates_t *state);
EXTERN void BmcCnfInsertClause(BmcCnfClauses_t *cnfClauses, array_t *clause);
EXTERN void BmcAddEmptyClause(BmcCnfClauses_t *cnfClauses);
EXTERN int BmcCnfReadOrInsertNode(BmcCnfClauses_t *cnfClauses, nameType_t *nodeName, int state, int *nodeIndex);
EXTERN void BmcGetCoiForLtlFormula(Ntk_Network_t *network, Ctlsp_Formula_t *formula, st_table *ltlCoiTable);
EXTERN void BmcGetCoiForLtlFormulaRecursive(Ntk_Network_t *network, Ctlsp_Formula_t *formula, st_table *ltlCoiTable, st_table *visited);
EXTERN void BmcGetCoiForNtkNode(Ntk_Node_t *node, st_table *CoiTable, st_table *visited);

/**AutomaticEnd***************************************************************/

#endif /* _BMCINT */
