/**CFile***********************************************************************

  FileName    [bmcUtil.c]

  PackageName [bmc]

  Synopsis    [Utilities for BMC package]

  Author      [Mohammad Awedh]

  Copyright   [This file was created at the University of Colorado at Boulder.
  The University of Colorado at Boulder makes no warranty about the suitability
  of this software for any purpose.  It is presented on an AS IS basis.]
  
******************************************************************************/

#include "bmcInt.h"

static char rcsid[] UNUSED = "$Id: bmcUtil.c,v 1.42 2002/09/21 07:48:31 fabio Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

#define MAX_LENGTH 80 /* Max. length of a string while reading a file     */

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Structure declarations                                                    */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static void printTable(st_table *mappingLatchesTable);
static int StringCheckIsInteger(char *string, int *value);
static int searchArray(array_t *result, int value);
static int nameCompare(const void * name1, const void * name2);
static void printValue(mAig_Manager_t *manager, Ntk_Network_t *network, st_table *nodeToMvfAigTable, BmcCnfClauses_t *cnfClauses, array_t *varNames, array_t *result, int state, int *prevValue);

/**AutomaticEnd***************************************************************/

/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Returns MvfAig corresponding to a node; returns NIL if node not in
  table.]

  SideEffects [None]

******************************************************************************/
MvfAig_Function_t *
BmcReadMvfAig(
  Ntk_Node_t * node,
  st_table   * nodeToMvfAigTable)
{
  MvfAig_Function_t *result;
  if (st_lookup(nodeToMvfAigTable, (char *) node, (char **) &result)){
     return result;
  }
  return NIL(MvfAig_Function_t);
}

/**Function********************************************************************

  Synopsis [Build AND/INV graph for propsitional formula]

  Description [Build AND/INV graph for propsitional formula. Returns bAig Id
  of the bAig Function equivalent to the propositional fomula]

  SideEffects []

  SeeAlso     [Ctlsp_FormulaReadClass]

******************************************************************************/
mAigEdge_t
BmcCreateMaigOfPropFormula(
  Ntk_Network_t   *network,
  mAig_Manager_t  *manager,
  Ctlsp_Formula_t *formula)
{
  mAigEdge_t left, right, result;

  if (formula == NIL(Ctlsp_Formula_t)) {
    return mAig_NULL;
  }
  if (formula->type == Ctlsp_TRUE_c){
    return mAig_One;
  }
  if (formula->type == Ctlsp_FALSE_c){
    return mAig_Zero;
  }
  if (!Ctlsp_isPropositionalFormula(formula)) {
    (void) fprintf(vis_stderr, "bmc error: Only propositional formula can be converted to AND/INVERTER graph\n");
    fprintf(vis_stdout, "\nFormula: ");
    Ctlsp_FormulaPrint(vis_stdout, formula);
    fprintf(vis_stdout, "\n");
    return mAig_NULL;
  }
  if (formula->type == Ctlsp_ID_c){
      char *nodeNameString  = Ctlsp_FormulaReadVariableName(formula);
      char *nodeValueString = Ctlsp_FormulaReadValueName(formula);
      Ntk_Node_t *node      = Ntk_NetworkFindNodeByName(network, nodeNameString);

      Var_Variable_t *nodeVar;
      int             nodeValue;

      MvfAig_Function_t  *tmpMvfAig;
      st_table           *nodeToMvfAigTable; /* maps each node to its mvfAig */
      
      if (node == NIL(Ntk_Node_t)) {
          (void) fprintf(vis_stderr, "bmc error: Could not find node corresponding to the name\t %s\n", nodeNameString);
          return mAig_NULL;
      }

      nodeToMvfAigTable = (st_table *) Ntk_NetworkReadApplInfo(network, MVFAIG_NETWORK_APPL_KEY);
      if (nodeToMvfAigTable == NIL(st_table)){
         (void) fprintf(vis_stderr, "bmc error: please run build_partiton_maigs first");
         return mAig_NULL;
      }
      tmpMvfAig = BmcReadMvfAig(node, nodeToMvfAigTable);
      if (tmpMvfAig ==  NIL(MvfAig_Function_t)){
          tmpMvfAig = BmcNodeBuildMVF(network, node);
	  array_free(tmpMvfAig);
	  tmpMvfAig = BmcReadMvfAig(node, nodeToMvfAigTable);
      }
      
      nodeVar = Ntk_NodeReadVariable(node);
      if (Var_VariableTestIsSymbolic(nodeVar)) {
        nodeValue = Var_VariableReadIndexFromSymbolicValue(nodeVar, nodeValueString);
        if ( nodeValue == -1 ) {
          (void) fprintf(vis_stderr, "Value specified in RHS is not in domain of variable\n");
          (void) fprintf(vis_stderr,"%s = %s\n", nodeNameString, nodeValueString);
          return mAig_NULL;
        }
      }
      else { 
        int check;    
         check = StringCheckIsInteger(nodeValueString, &nodeValue);
         if( check == 0 ) {
          (void) fprintf(vis_stderr,"Illegal value in the RHS\n");
          (void) fprintf(vis_stderr,"%s = %s\n", nodeNameString, nodeValueString);
          return mAig_NULL;
	 }
         if( check == 1 ) {
          (void) fprintf(vis_stderr,"Value in the RHS is out of range of int\n");
          (void) fprintf(vis_stderr,"%s = %s", nodeNameString, nodeValueString);
          return mAig_NULL;
	 }
         if ( !(Var_VariableTestIsValueInRange(nodeVar, nodeValue))) {
          (void) fprintf(vis_stderr,"Value specified in RHS is not in domain of variable\n");
          (void) fprintf(vis_stderr,"%s = %s\n", nodeNameString, nodeValueString);
          return mAig_NULL;

         }
      }
      result = MvfAig_FunctionReadComponent(tmpMvfAig, nodeValue);
      return result;
  }

  /*
    right can be mAig_NULL for unery operators, but left can't be mAig_Null
   */
  
  left  = BmcCreateMaigOfPropFormula(network, manager, formula->left);
  if (left == mAig_NULL){
    return mAig_NULL;
  }  
  right = BmcCreateMaigOfPropFormula(network, manager, formula->right);

  switch(formula->type) {
    case Ctlsp_NOT_c:
      result = mAig_Not(left);
      break;
    case Ctlsp_OR_c:
      result = mAig_Or(manager, left, right);
      break;
    case Ctlsp_AND_c:
      result = mAig_And(manager, left, right);
      break;
    case Ctlsp_THEN_c:
      result = mAig_Then(manager, left, right); 
      break;
    case Ctlsp_EQ_c:
      result = mAig_Eq(manager, left, right);
      break;
    case Ctlsp_XOR_c:
      result = mAig_Xor(manager, left, right);
      break;
    default:
      fail("Unexpected type");
  }
  return result;
}

/**Function********************************************************************

  Synopsis    [Generate the CNF formula of a given function represented by
               AND/INVERTER graph]

  Description [Generate an array of clausese for a function represented in
  AND/INVERETER graph structure. 
  
  the CNF formula of node to the output file specifies by cnfFile.
               It stores CNF index for each node in the cnfIndexTable. The generated CNF
               is in dimacs format.
	       It is an error to call this function on a constand zero/one node.]

  SideEffects []

  SeeAlso     []

******************************************************************************/
int
BmcGenerateCnfFormulaForAigFunction(
   bAig_Manager_t *manager,
   bAigEdge_t      node,
   int             k,
   BmcCnfClauses_t *cnfClauses)
{
  int        leftIndex, rightIndex, nodeIndex;
  array_t    *clause;

  assert( (node != bAig_One) && (node != bAig_Zero));
  
  if(bAig_IsInverted(node)){
    /*
      The generated clauses are in dimacs formate that uses negateve number to indicate complement
     */
    return -1*BmcGenerateCnfFormulaForAigFunction(manager, bAig_NonInvertedEdge(node), k, cnfClauses);
  }
  if (BmcCnfReadOrInsertNode(cnfClauses, bAig_NodeReadName(manager, node), k, &nodeIndex)){
    return nodeIndex;
  }
  if (bAig_isVarNode(manager, node)){
    return nodeIndex;
  }
  leftIndex  = BmcGenerateCnfFormulaForAigFunction(manager,  leftChild(node),  k, cnfClauses);
  rightIndex = BmcGenerateCnfFormulaForAigFunction(manager,  rightChild(node), k, cnfClauses);
 
  clause  = array_alloc(int, 3);
  array_insert(int, clause, 0, -leftIndex );
  array_insert(int, clause, 1, -rightIndex);
  array_insert(int, clause, 2,  nodeIndex );
  BmcCnfInsertClause(cnfClauses, clause);
  array_free(clause);
  
  clause  = array_alloc(int, 2);
  array_insert(int, clause, 0,  leftIndex);
  array_insert(int, clause, 1, -nodeIndex);
  BmcCnfInsertClause(cnfClauses, clause);
  array_free(clause);
  
  clause  = array_alloc(int, 2);
  array_insert(int, clause, 0,  rightIndex);
  array_insert(int, clause, 1, -nodeIndex);
  BmcCnfInsertClause(cnfClauses, clause);  
  array_free(clause);

  return(nodeIndex);

}

/**Function********************************************************************

  Synopsis [
               c = a <-> b = (a=0)*(b=0) + (a=1)*(b=1) ....
           For a given term (a=i)*(b=i), if either is Zero,  don't generate
	                                 clauses for this term.
	                                 if both are One, don't generate clauses
					 for c.
  ]

  SideEffects []

******************************************************************************/
void
BmcGenerateCnfFromStateTostate(
   bAig_Manager_t  *manager,
   bAigEdge_t      *fromAigArray,
   bAigEdge_t      *toAigArray,
   int             mvfSize, 
   int             fromState,
   int             toState,
   BmcCnfClauses_t *cnfClauses,
   int             outIndex) 
{
  array_t    *clause, *tmpclause;
  int        toIndex, fromIndex, cnfIndex;
  int        i;

  /* used to turn off the warning messages: might be left uninitialized.
     We are sure that these two variables will not be used uninitialized.
  */
  toIndex =0;
  fromIndex = 0;
  
  for(i=0; i< mvfSize; i++){
    if ((fromAigArray[i] == bAig_One) && (toAigArray[i] == bAig_One)){
      return;   /* the clause is always true */
    }
  }
  clause  = array_alloc(int, 0);
  for(i=0; i< mvfSize; i++){
    if ((fromAigArray[i] != bAig_Zero) && (toAigArray[i] != bAig_Zero)){
      if (toAigArray[i] != bAig_One){ 
         /* to State */

         toIndex = BmcGenerateCnfFormulaForAigFunction(manager,toAigArray[i],
						       toState,cnfClauses);
      }
      if (fromAigArray[i] != bAig_One){ 
         /* from State */
         fromIndex = BmcGenerateCnfFormulaForAigFunction(manager,
							 fromAigArray[i],
							 fromState,
							 cnfClauses);
      }
     /*
      Create new var for the output of this node. We don't create variable for this node, we only
      use its index number.
     */
     cnfIndex = cnfClauses->cnfGlobalIndex++;  /* index of the output of the OR of T(from, to) */
     if (toAigArray[i] == bAig_One){       
       tmpclause  = array_alloc(int, 2);
       array_insert(int, tmpclause, 0, -fromIndex);
       array_insert(int, tmpclause, 1, cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause); 

       tmpclause  = array_alloc(int, 2);
       array_insert(int, tmpclause, 0, fromIndex);
       array_insert(int, tmpclause, 1, -cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause);       

     } else if (fromAigArray[i] == bAig_One){
       tmpclause  = array_alloc(int, 2);
       array_insert(int, tmpclause, 0, -toIndex);
       array_insert(int, tmpclause, 1, cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause);

       tmpclause  = array_alloc(int, 2);
       array_insert(int, tmpclause, 0, toIndex);
       array_insert(int, tmpclause, 1, -cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause);
       
     } else {
       tmpclause  = array_alloc(int, 3);
       array_insert(int, tmpclause, 0, -toIndex);
       array_insert(int, tmpclause, 1, -fromIndex);
       array_insert(int, tmpclause, 2,  cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause);

       tmpclause  = array_alloc(int, 2);
       array_insert(int, tmpclause, 0, toIndex);
       array_insert(int, tmpclause, 1, -cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause);

       tmpclause  = array_alloc(int, 2);
       array_insert(int, tmpclause, 0, fromIndex);
       array_insert(int, tmpclause, 1, -cnfIndex);
       BmcCnfInsertClause(cnfClauses, tmpclause);
       array_free(tmpclause);
     }
     array_insert_last(int, clause, cnfIndex);
    } /* if */
  } /* for i loop */
  if (outIndex != 0 ){
    array_insert_last(int, clause, -outIndex);
  }
  BmcCnfInsertClause(cnfClauses, clause);
  array_free(clause);
  
  return;
}

/**Function********************************************************************

  Synopsis [Write the set of clauses in diamacs format to the output file.]

  SideEffects []

******************************************************************************/
void
BmcWriteClauses(
   mAig_Manager_t  *maigManager,
   FILE            *cnfFile,
   BmcCnfClauses_t *cnfClauses)
{
  st_generator *stGen;
  char         *name;
  int          cnfIndex, i, k;
  
  st_foreach_item_int(cnfClauses->cnfIndexTable, stGen, &name, &cnfIndex) {
    fprintf(cnfFile, "c %s %d\n",name, cnfIndex);
  }
  (void) fprintf(cnfFile, "p cnf %d %d\n", cnfClauses->cnfGlobalIndex-1,
		 cnfClauses->noOfClauses);
  if (cnfClauses->clauseArray != NIL(array_t)) {
    for (i = 0; i < cnfClauses->nextIndex; i++) {
      k = array_fetch(int, cnfClauses->clauseArray, i);
      (void) fprintf(cnfFile, "%d%c", k, (k == 0) ? '\n' : ' ');
    }
  }
  return;
}

/**Function********************************************************************

  Synopsis [Check the satisfiability of CNF formula written in file]

  Description [Run SAT solver on the input file]

  SideEffects []

******************************************************************************/
array_t *
BmcCheckSAT(BmcOption_t *options)
{
  FILE        *fp;
  static char parseBuffer[1024];
  int         satStatus;
  char        line[MAX_LENGTH];
  int         num;
  array_t     *result = NIL(array_t);
  char        *tmpStr, *tmpStr1, *tmpStr2;
  long        solverStart;
  int         satTimeOutPeriod = 0;

  strcpy(parseBuffer,"zchaff ");
  if (options->timeOutPeriod > 0) {
     /* Compute the residual CPU time and subtract a little time to
	give vis a chance to clean up before its own time out expires.
     */
    satTimeOutPeriod = options->timeOutPeriod - 1 -
      (util_cpu_ctime() - options->startTime) / 1000;
    if (satTimeOutPeriod <= 0){ /* no time left to run SAT solver*/
       return NIL(array_t);
    }
    tmpStr2 = util_inttostr(satTimeOutPeriod);
    tmpStr1 = util_strcat3(options->satInFile," ", tmpStr2);
    tmpStr = util_strcat3(tmpStr1, " >", options->satOutFile);
    FREE(tmpStr1);
    FREE(tmpStr2);
  } else {
    tmpStr = util_strcat3(options->satInFile, " >", options->satOutFile);
  }
  strcat(parseBuffer, tmpStr);
  FREE(tmpStr);

  if (options->verbosityLevel == BmcVerbosityMax_c) {
    (void)fprintf(vis_stdout,"Calling SAT solver ...");
    (void) fflush(vis_stdout);
    solverStart = util_cpu_ctime();
  } else { /* to remove uninitialized variables warning */
    solverStart = 0;
  }
  /* Call Sat Solver*/
  satStatus = system(parseBuffer);
  if (satStatus != 0){
    (void) fprintf(vis_stderr, "Can't run zchaff. zchaff may not be installed in your system\n");
    return NIL(array_t);
  }

  if (options->verbosityLevel == BmcVerbosityMax_c) {
    (void) fprintf(vis_stdout," done ");
    (void) fprintf(vis_stdout, "(%g s)\n",
		   (double) (util_cpu_ctime() - solverStart)/1000.0);
  }
  fp = Cmd_FileOpen(options->satOutFile, "r", NIL(char *), 0);
  if (fp == NIL(FILE)) {
    (void) fprintf(vis_stderr, "** bmc error: Cannot open the file %s\n",
		   options->satOutFile);
    return NIL(array_t);
  }
  /* Skip the first two lines */
  fgets(line, MAX_LENGTH - 1, fp);
  fgets(line, MAX_LENGTH - 1, fp);
  fgets(line, MAX_LENGTH - 1, fp);

  if(strstr(line,"unsatisfiable") != NIL(char)) {
    if (options->verbosityLevel != BmcVerbosityNone_c){
      (void) fprintf(vis_stdout, "# BMC: Counterexample not found\n");
    }
  } else if(strstr(line,"satisfiable") != NIL(char)) {
    if (options->verbosityLevel != BmcVerbosityNone_c){
      (void) fprintf(vis_stdout, "# BMC: Counterexample found\n");
    }
    result = array_alloc(int, 0);
    while (fscanf(fp,"%d ", &num) != feof(fp)){
      array_insert_last(int, result, num);
    }
  } else if(strstr(line,"Memory out") != NIL(char)) {
    (void) fprintf(vis_stdout,"# BMC: SAT Solver Memory out\n");
  } else if(strstr(line,"Time out") != NIL(char)) {
    (void) fprintf(vis_stdout,
		   "# BMC: SAT Solver Time out occurred after %d seconds.\n",
		   satTimeOutPeriod);
  } else {
    (void) fprintf(vis_stdout, "# BMC: SAT Solver failed, try again\n");
  }
  (void) fflush(vis_stdout);
  (void) fclose(fp);

  /* Adjust alarm if timeout in effect.  This is necessary because the
   * alarm may have gone off while the SAT solver is running.  Since
   * the CPU time of a child process is charged to the parent only when
   * the child terminates, the SIGALRM handler assumes that more time
   * is left than it is in reality.  We could do this adjustment right
   * after calling the SAT solver, but we prefer to give ourselves the
   * extra time to report the result even if this means using more time
   * than we are allotted.
   */
  if (options->timeOutPeriod > 0) {
    int residualTime = options->timeOutPeriod -
      (util_cpu_ctime() - options->startTime) / 1000;
    /* Make sure we do not cancel the alarm if no time is left. */
    if (residualTime <= 0) {
      residualTime = 1;
    }
    (void) alarm(residualTime);
  }

  return result;
}

/**Function********************************************************************

  Synopsis [Build multi-valued function for a given node.]

  SideEffects [The mvf of this node is a function of combinational input
  nodes.]

******************************************************************************/

MvfAig_Function_t *
BmcNodeBuildMVF(
  Ntk_Network_t *network,
  Ntk_Node_t    *node)
{
  MvfAig_Function_t * MvfAig; 
  lsGen tmpGen;
  Ntk_Node_t *tmpNode;
  array_t    *mvfArray;
  array_t    *tmpRoots = array_alloc(Ntk_Node_t *, 0);
  st_table   *tmpLeaves = st_init_table(st_ptrcmp, st_ptrhash);
  array_insert_last(Ntk_Node_t *, tmpRoots, node);

  Ntk_NetworkForEachCombInput(network, tmpGen, tmpNode) {
    st_insert(tmpLeaves, (char *) tmpNode, (char *) ntmaig_UNUSED);
  }

  mvfArray = ntmaig_NetworkBuildMvfAigs(network, tmpRoots, tmpLeaves);
  MvfAig   = array_fetch(MvfAig_Function_t *, mvfArray, 0);
  array_free(tmpRoots);
  st_free_table(tmpLeaves);
  array_free(mvfArray);
  return MvfAig;
}

/**Function********************************************************************

  Synopsis [Print CounterExample.]

  SideEffects [The mvf of this node is a function of combinational input nodes.]

******************************************************************************/
void
BmcPrintCounterExample(
  Ntk_Network_t   *network,
  st_table        *nodeToMvfAigTable,
  BmcCnfClauses_t *cnfClauses,
  array_t         *result,
  int             length,
  st_table        *CoiTable,
  BmcOption_t     *options)
{
  mAig_Manager_t    *manager = Ntk_NetworkReadMAigManager(network);
  lsGen             gen;
  Ntk_Node_t        *node;
  int               k;
  array_t           *latches = array_alloc(int, 0);
  int               *prevValueLatches;
  array_t           *inputs  = array_alloc(int, 0);
  int               *prevValueInputs;
  int               tmp;

  /* sort latches by name */
  Ntk_NetworkForEachLatch(network, gen, node){
    /* skip the latch if it is not in COI of property*/
    if (st_lookup_int(CoiTable, (char *) node, &tmp)){
      array_insert_last(char*, latches, Ntk_NodeReadName(node));
    }
  }
  array_sort(latches, nameCompare);
  prevValueLatches = ALLOC(int, array_n(latches));

  /* sort inputs by name */
  Ntk_NetworkForEachInput(network, gen, node){
    array_insert_last(char*, inputs, Ntk_NodeReadName(node));
  }
  array_sort(inputs, nameCompare);
  prevValueInputs = ALLOC(int, array_n(inputs));
 
  /*
  Ntk_NetworkForEachPrimaryOutput(network, gen, node){
    array_insert_last(char*, outputs, Ntk_NodeReadName(node));
  }
  array_sort(outputs, nameCompare);
  */
  for (k=0; k<= length; k++){
    if (k == 0){
      (void) fprintf(vis_stdout, "\n--State %d:\n", k);
    } else {
      (void) fprintf(vis_stdout, "\n--Goes to state %d:\n", k);
    }
    printValue(manager, network, nodeToMvfAigTable, cnfClauses,
	       latches, result,  k, prevValueLatches);  
#if 0
    (void) fprintf(vis_stdout, "--Primary output:\n");
    printValue(manager, network, nodeToMvfAigTable, cnfClauses, outputs, result, k);
#endif
    if((options->printInputs == TRUE) && (k !=0)) {
      (void) fprintf(vis_stdout, "--On input:\n");
      printValue(manager, network, nodeToMvfAigTable, cnfClauses,
		 inputs, result,   k-1, prevValueInputs);   
    }
  } /* for k loop */
  return;
} /* BmcPrintCounterExample() */

/**Function********************************************************************

  Synopsis    [Generate CNF formula for a path from state 'from' to state 'to']

  Description [Unfold the transition relation 'k' states (k = to-from +1), and
  generate clauses for each state.
           For a multi-valued latch of 4 vlaues. Two binary variables are used 
           to rpresent X, x1 and x0. For this latch, there exist three multi-valued 
           functions.
           One for the binary reoresentation of the variable. For example the
           second entry of the mvf = 1, iff ~x1 and x0. 
           The second mfv is for the data input of the latch. If the And/Inv graph 
           attached to an entry of this mvf equal to 1, X equal to the binary 
           representation corresponding to this entry. For example, if the And/
           INV graph attached to the first entry =1, then X = ~x1 & ~x0. 
           To generate the CNF to the transition relation, first generate CNF to
           next state varaible using mvf of the latch.  Then, generate CNF for 
           latch data input using current state variavles. Finaly, generate CNF
           for the AND of these two MVF.  This for every entry of the MVF. Then 
           OR the results.
           The third MVF is for the initial value of the latch.  It is treat the
           same as the latch data input except if the initial value is constant. 
        
           ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
void
BmcGenerateCnfForPath(
   Ntk_Network_t   *network,
   int             from,
   int             to,
   BmcCnfClauses_t *cnfClauses,
   st_table        *nodeToMvfAigTable,
   st_table        *CoiTable)
{
  mAig_Manager_t  *manager   = Ntk_NetworkReadMAigManager(network);
  lsGen           gen;
  
  Ntk_Node_t         *latch, *latchData, *latchInit;
  MvfAig_Function_t  *initMvfAig, *dataMvfAig, *latchMvfAig;
  bAigEdge_t         *initBAig, *latchBAig, *dataBAig;
  
  int        i, k, mvfSize;

  Ntk_NetworkForEachLatch(network, gen, latch) { 
    int tmp;
    /* skip the latch if it is not in COI of property*/
    if (!st_lookup_int(CoiTable, (char *) latch, &tmp)){
      continue;
    }
   latchInit  = Ntk_LatchReadInitialInput(latch);
   latchData  = Ntk_LatchReadDataInput(latch);

   /* Get the multi-valued function for each node*/
   initMvfAig = BmcReadMvfAig(latchInit, nodeToMvfAigTable);
   if (initMvfAig ==  NIL(MvfAig_Function_t)){
     (void) fprintf(vis_stdout, "No multi-valued function for this node %s \n", Ntk_NodeReadName(latchInit));
     return;
   } 
   dataMvfAig = BmcReadMvfAig(latchData, nodeToMvfAigTable);
   if (dataMvfAig ==  NIL(MvfAig_Function_t)){
     (void) fprintf(vis_stdout, "No multi-valued function for this node %s \n", Ntk_NodeReadName(latchData));
     return;
   }
   latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);
   if (latchMvfAig ==  NIL(MvfAig_Function_t)){
     latchMvfAig = BmcNodeBuildMVF(network, latch);
     array_free(latchMvfAig);
     latchMvfAig = BmcReadMvfAig(latch, nodeToMvfAigTable);
   }
      
   mvfSize   = array_n(initMvfAig);
   initBAig  = ALLOC(bAigEdge_t, mvfSize);
   dataBAig  = ALLOC(bAigEdge_t, mvfSize);
   latchBAig = ALLOC(bAigEdge_t, mvfSize);   

   for(i=0; i< mvfSize; i++){
     dataBAig[i]  = MvfAig_FunctionReadComponent(dataMvfAig,  i);
     latchBAig[i] = MvfAig_FunctionReadComponent(latchMvfAig, i);
     initBAig[i]  = MvfAig_FunctionReadComponent(initMvfAig,  i);
   }   
   if (from == 0){
   /* Generate the CNF for the initial state of the latch */
     BmcGenerateCnfFromStateTostate(manager, initBAig, latchBAig, mvfSize, -1, 0, cnfClauses, 0);
   }
   /* Generate the CNF for the transition functions */   
   for (k=from; k < to; k++){
     BmcGenerateCnfFromStateTostate(manager, dataBAig, latchBAig, mvfSize, k, k+1, cnfClauses, 0);
   } /* for k state loop */
   FREE(initBAig);
   FREE(dataBAig);
   FREE(latchBAig);
  } /* ForEachLatch loop*/
  return;
}

/**Function********************************************************************

  Synopsis [Alloc Memory for BmcCnfClauses_t]

  SideEffects []

  SeeAlso []
******************************************************************************/
BmcCnfClauses_t *
BmcCnfClausesAlloc(void)
{
  BmcCnfClauses_t *result = ALLOC(BmcCnfClauses_t, 1);
  
  if (!result){
    return result;
  }
  result->clauseArray    = array_alloc(int, 0);
  result->cnfIndexTable  = st_init_table(strcmp, st_strhash);
  result->freezedKeys    = array_alloc(nameType_t *, 0);
  
  result->nextIndex      = 0;
  result->noOfClauses    = 0;
  result->cnfGlobalIndex = 1;  
  result->emptyClause    = FALSE;
  result->freezed        = FALSE;

  return result;
} /*BmcCnfClausesAlloc()*/

/**Function********************************************************************

  Synopsis [Free Memory for BmcCnfClauses_t]

  SideEffects []

  SeeAlso []
******************************************************************************/
void 
BmcCnfClausesFree(
  BmcCnfClauses_t *cnfClauses)
{
  st_generator *stGen;
  char         *name;
  int          cnfIndex;
  
  array_free(cnfClauses->clauseArray);
  array_free(cnfClauses->freezedKeys);

  if (cnfClauses->cnfIndexTable != NIL(st_table)){
    st_foreach_item_int(cnfClauses->cnfIndexTable, stGen, (char **) &name, &cnfIndex) {
      FREE(name);
    }
    st_free_table(cnfClauses->cnfIndexTable);
  }
  FREE(cnfClauses);
} /* BmcCnfClausesFree() */

/**Function********************************************************************

  Synopsis [Freeze the current state of CNF]

  Description [The current state of CNF is stored in the structure BmcCnfStates_t.
  This information may use later to store CNF to this state by calling
  BmcCnfClausesRestore().]

  SideEffects []

  SeeAlso [BmcCnfClausesRestore() BmcCnfClausesUnFreeze() ]
******************************************************************************/
BmcCnfStates_t *
BmcCnfClausesFreeze(
  BmcCnfClauses_t * cnfClauses)
{
  BmcCnfStates_t *state = ALLOC(BmcCnfStates_t, 1);

  state->nextIndex      = cnfClauses->nextIndex;
  state->noOfClauses    = cnfClauses->noOfClauses;
  state->cnfGlobalIndex = cnfClauses->cnfGlobalIndex;

  /* This variable is used when deleting any new entries in cnfClauses->freezedKeys
   that will be added after CNF is freezed.*/
  state->freezedKeySize = array_n(cnfClauses->freezedKeys);
  
  state->emptyClause    = cnfClauses->emptyClause;
  state->freezed        = cnfClauses->freezed;
  cnfClauses->freezed   = TRUE;
  return state;
} /* mcCnfClausesFreeze() */

/**Function********************************************************************

  Synopsis [Unfreeze CNF]

  Description [Keeps the current state of CNF]

  SideEffects []

  SeeAlso []
******************************************************************************/
void 
BmcCnfClausesUnFreeze(
  BmcCnfClauses_t *cnfClauses,
  BmcCnfStates_t  *state)
{
  int i;

  cnfClauses->freezed = FALSE;
  if (array_n(cnfClauses->freezedKeys) != 0){
    int freezedKeySize = array_n(cnfClauses->freezedKeys);
    
    for (i=0; i< (freezedKeySize-state->freezedKeySize); i++){
      (cnfClauses->freezedKeys)->num--;
    }
  }
} /* BmcCnfClausesUnFreeze() */

/**Function********************************************************************

  Synopsis [Restore the CNF to its previouse state]

  Description [Restore the CNF to its previouse state that CNF was when
  BmcCnfClausesFreeze() was last called]

  SideEffects []
  
  SeeAlso []
******************************************************************************/
void 
BmcCnfClausesRestore(
  BmcCnfClauses_t *cnfClauses,
  BmcCnfStates_t  *state)
{
  int        i, index;
  nameType_t *key;

  cnfClauses->nextIndex      = state->nextIndex;
  cnfClauses->noOfClauses    = state->noOfClauses;
  cnfClauses->cnfGlobalIndex = state->cnfGlobalIndex;
  cnfClauses->emptyClause    = state->emptyClause;
  cnfClauses->freezed        = state->freezed;
  
  if (array_n(cnfClauses->freezedKeys) != 0){
    int freezedKeySize = array_n(cnfClauses->freezedKeys);
		       
    for (i=0; i< (freezedKeySize-state->freezedKeySize); i++){
      key = array_fetch_last(nameType_t *, cnfClauses->freezedKeys);
      if (st_delete_int(cnfClauses->cnfIndexTable, &key, &index)){
	FREE(key);
      } 
      (cnfClauses->freezedKeys)->num--;
    }
  }
} /* BmcCnfClausesRestore() */

/**Function********************************************************************

  Synopsis [Add clause to the clauseArray]

  Description [Add a clause to the clause array.  clause is of type array_t.
  The user must free clause.]

  SideEffects []
  
  SeeAlso []
******************************************************************************/
void 
BmcCnfInsertClause(
  BmcCnfClauses_t *cnfClauses,
  array_t         *clause)
{
  int  i, lit;

  if (clause != NIL(array_t)){
    if (array_n(clause) == 0){ /* empty clause */
      cnfClauses->emptyClause = TRUE;
      return;
    }
    for (i=0; i< array_n(clause); i++){
      lit = array_fetch(int, clause, i);
      array_insert(int, cnfClauses->clauseArray, cnfClauses->nextIndex++, lit);
    }
    array_insert(int, cnfClauses->clauseArray, cnfClauses->nextIndex++, 0); /*End Of clause*/
    cnfClauses->noOfClauses++;
    cnfClauses->emptyClause = FALSE;
  }
  return;
}/* BmcCnfInsertClause() */

/**Function********************************************************************

  Synopsis [Add an empty clause]

  SideEffects []

  SeeAlso []
******************************************************************************/
void
BmcAddEmptyClause(
  BmcCnfClauses_t *cnfClauses)
{
  cnfClauses->emptyClause = TRUE;
}/* BmcAddEmptyClause() */

/**Function********************************************************************

  Synopsis [Return the cnfIndex of the node]

  Description [If CNF was generated for this node, return its cnfIndex, otherwise
  insert the name of this node in the cnfIndexTable, and return its cnfIndex. The
  key to the cnfIndexTable is (nodeName_state).] 

  SideEffects []
******************************************************************************/
int
BmcCnfReadOrInsertNode(
   BmcCnfClauses_t *cnfClauses,
   nameType_t      *nodeName,
   int             state,
   int             *nodeIndex)
{
  nameType_t *varName;
  int        index;
  char       *stateStr = util_inttostr(state);

  varName  = util_strcat3(nodeName, "_", stateStr);
  FREE(stateStr);
  if (!st_lookup_int(cnfClauses->cnfIndexTable, varName, &index)) {
    index = cnfClauses->cnfGlobalIndex++;
    st_insert(cnfClauses->cnfIndexTable, varName, (char*) (long) index);
    if(cnfClauses->freezed == TRUE){
      array_insert_last(nameType_t *, cnfClauses->freezedKeys, varName);
    }
    *nodeIndex = index;
    return 0; /* Inserted */
  }
  else { /* The node has been visited */
    FREE(varName);
    *nodeIndex = index;
    return 1;     
  }
}


/**Function********************************************************************

  Synopsis    [Find the Cone of Influnce (COI) for an LTL formula]

  Description [Return a list of state variables (latches) that are in the COI of the
  LTL formula.]

  SideEffects []

******************************************************************************/
void
BmcGetCoiForLtlFormula(
  Ntk_Network_t   *network,
  Ctlsp_Formula_t *formula,
  st_table        *ltlCoiTable)
{
  st_table *visited = st_init_table(st_ptrcmp, st_ptrhash);
  
  BmcGetCoiForLtlFormulaRecursive(network, formula, ltlCoiTable, visited);
  st_free_table(visited);
  return;
} /* BmcGetCoiForLtlFormula() */

/**Function********************************************************************

  Synopsis    [Recursive function to find the COI of a network node.]

  Description []

  SideEffects []

******************************************************************************/
void
BmcGetCoiForLtlFormulaRecursive(
  Ntk_Network_t   *network,
  Ctlsp_Formula_t *formula,
  st_table        *ltlCoiTable,
  st_table        *visited)
{
  if (formula == NIL(Ctlsp_Formula_t)) {
    return;
  }
  /* leaf node */
  if (formula->type == Ctlsp_ID_c){
    char       *name = Ctlsp_FormulaReadVariableName(formula);
    Ntk_Node_t *node = Ntk_NetworkFindNodeByName(network, name);
    int        tmp;

    if (st_lookup_int(visited, (char *) node, &tmp)){
      /* Node already visited */
      return;
    }
    BmcGetCoiForNtkNode(node, ltlCoiTable, visited);
    return;
  }
  BmcGetCoiForLtlFormulaRecursive(network, formula->left,  ltlCoiTable, visited);
  BmcGetCoiForLtlFormulaRecursive(network, formula->right, ltlCoiTable, visited);

  return;
} /* BmcGetCoiForLtlFormulaRecursive() */

/**Function********************************************************************

  Synopsis    [Genrate COI for a non-latch network node]

  Description [For each fanins of the given node, if its latch, add it
  to the CoiTable and return. If it is not latch, call this function to
  look for any latches in the fanins of this node.]

  SideEffects []

******************************************************************************/
void
BmcGetCoiForNtkNode(
  Ntk_Node_t  *node,
  st_table    *CoiTable,
  st_table    *visited)
{
  int        i, j;
  Ntk_Node_t *faninNode;

  if(node == NIL(Ntk_Node_t)){
    return;
  }
  if (st_lookup_int(visited, (char *) node, &j)){
    /* Node already visited */
    return;
  }
  st_insert(visited, (char *) node, (char *) 0);
  if (Ntk_NodeTestIsLatch(node)){
    st_insert(CoiTable, (char *) node, (char *) 0);
  }
  Ntk_NodeForEachFanin(node, i, faninNode) {
    BmcGetCoiForNtkNode(faninNode, CoiTable, visited);
  }
  return;
} /* BmcGetCoiForNtkNode() */

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis []

  SideEffects []
******************************************************************************/
static void 
printTable(
  st_table  *mappingLatchesTable)
{
  char *key; 
  char *data = NIL(char);
  st_generator *stGen;

  st_foreach_item(mappingLatchesTable, stGen, &key, &data) {
    printf("%s  %s \n", key, data);
  }
}


/**Function********************************************************************

  Synopsis [Test that the given string is an integer. Returns 0 if string is
  not an integer, 1 if the integer is too big for int, and 2 if integer fits
  in int.]

  SideEffects [Sets the pointer value if the string is an integer small enough
  for int.]

******************************************************************************/
static int
StringCheckIsInteger(
  char *string,
  int *value)
{
  char *ptr;
  long l;
  
  l = strtol (string, &ptr, 0) ;
  if(*ptr != '\0')
    return 0;
  if ((l > MAXINT) || (l < -1 - MAXINT))
    return 1 ;
  *value = (int) l;
  return 2 ;
}

/**Function********************************************************************

  Synopsis [Search the given array for a given value. Return -1 if not found]

  SideEffects []

******************************************************************************/
static int
searchArray(
  array_t *result,
  int     value)
{
  int i;

  for(i=0; i< array_n(result); i++){
    if (array_fetch(int, result, i) == value){
      return i; /* found at index i*/
    }
  }
  return -1; /* not found */
}

/**Function********************************************************************

  Synopsis    [Compare procedure for name comparison.]

  Description [Compare procedure for name comparison.]

  SideEffects []

******************************************************************************/
static int
nameCompare(
  const void * name1,
  const void * name2)
{
    return(strcmp(*(char**)name1, *(char **)name2));

} /* end of signatureCompare */

/**Function********************************************************************

  Synopsis    []

  Description []

  SideEffects []

******************************************************************************/
static void
printValue(
  mAig_Manager_t  *manager,
  Ntk_Network_t   *network,
  st_table        *nodeToMvfAigTable,
  BmcCnfClauses_t *cnfClauses,
  array_t         *varNames,
  array_t         *result,
  int             state,
  int             *prevValue)
{
  Ntk_Node_t        *node;
  int               i, j;
  bAigEdge_t        bAigId;
  nameType_t        *varName, *nodeName;
  int               value, index;
  MvfAig_Function_t *MvfAig;
  int               changed = 0; 

  for (j=0; j< array_n(varNames); j++) {
    if (state == 0){
      prevValue[j] = -1;
    }
    nodeName = array_fetch(char *, varNames, j);
    node = Ntk_NetworkFindNodeByName(network, nodeName);
    /* Get the multi-valued function for each node*/
    MvfAig = BmcReadMvfAig(node, nodeToMvfAigTable);
    if (MvfAig ==  NIL(MvfAig_Function_t)){
      continue;
    }
    value = -1;
    for (i=0; i< array_n(MvfAig); i++) {
      bAigId = MvfAig_FunctionReadComponent(MvfAig, i);
      /*
	constant value
       */
      if (bAigId == bAig_One){
	value = i;
	break;
      }
      if (bAigId != bAig_Zero){
	nodeName = bAig_NodeReadName(manager, bAigId);
	varName  = util_strcat3(nodeName, "_", util_inttostr(state));
	if (st_lookup_int(cnfClauses->cnfIndexTable, varName, &index)) {
	  if (bAig_IsInverted(bAigId)){
	    index = -index;
	  }
	  if (searchArray(result, index) > -1){
	    value = i;
	    break;
	  }
	} /* if st_lookup()  */
	FREE(varName);
      } /* if (bAigId != bAig_Zero) */
    }
    if (value >= 0){
      if (value != prevValue[j]){
	Var_Variable_t *nodeVar = Ntk_NodeReadVariable(node);
	
	prevValue[j] = value;
	changed = 1;
	if (Var_VariableTestIsSymbolic(nodeVar)) {
	  char *symbolicValue = Var_VariableReadSymbolicValueFromIndex(nodeVar, value);
	  
	  (void) fprintf(vis_stdout,"%s:%s\n",  Ntk_NodeReadName(node), symbolicValue);
	} 
	else {
	  (void) fprintf(vis_stdout,"%s:%d\n",  Ntk_NodeReadName(node), value);
	}
      }
    }
  } /* for j loop */
  if (changed == 0){
    fprintf( vis_stdout, "<Unchanged>\n");
  }
}/* end of printValue() */
