/**CFile***********************************************************************

  FileName    [mcMc.c]

  PackageName [mc]

  Synopsis    [Fair CTL model checker.]

  Description [This file contains the functions for both backward and
  forward model checking. Forward model checking is based on Iwashta's
  ICCAD96 paper. Future tense CTLs are automatically converted to past
  tense ones as much as possible in forward model checking.]

  Author [Adnan Aziz, Tom Shiple, Rajeev Ranjan, In-Ho Moon, Roderick Bloem]

  Copyright   [Copyright (c) 1994-1996 The Regents of the Univ. of California.
  All rights reserved.

  Permission is hereby granted, without written agreement and without license
  or royalty fees, to use, copy, modify, and distribute this software and its
  documentation for any purpose, provided that the above copyright notice and
  the following two paragraphs appear in all copies of this software.

  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
  DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
  OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
  CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
  FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN
  "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE
  MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.]

******************************************************************************/

#include "ctlp.h"
#include "mcInt.h"

static char rcsid[] UNUSED = "$Id: mcMc.c,v 1.213 2002/09/18 01:46:05 wangc Exp $";

#ifdef DEBUG_MC
void _McPrintSatisfyingMinterms(mdd_t *aMdd, Fsm_Fsm_t *fsm);
#endif

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
static jmp_buf timeOutEnv;


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

static mdd_t * EvaluateFormulaRecur(Fsm_Fsm_t *modelFsm, Ctlp_Formula_t *ctlFormula, mdd_t *fairStates, Fsm_Fairness_t *fairCondition, array_t *modelCareStatesArray, Mc_EarlyTermination_t *earlyTermination, Fsm_HintsArray_t *hintsArray, Mc_GuidedSearch_t hintType, Ctlp_Approx_t TRMinimization, Ctlp_Approx_t *resultApproxType, Mc_VerbosityLevel verbosity, Mc_DcLevel dcLevel, int buildOnionRings, Mc_GSHScheduleType GSHschedule);
static int CommandMc(Hrc_Manager_t **hmgr, int argc, char **argv);
static mdd_t * ModelCheckAtomicFormula(Fsm_Fsm_t *modelFsm, Ctlp_Formula_t *ctlFormula);
static McOptions_t * ParseMcOptions(int argc, char **argv);
static int CommandLe(Hrc_Manager_t **hmgr, int argc, char **argv);
static McOptions_t * ParseLeOptions(int argc, char **argv);
static int CommandInv(Hrc_Manager_t **hmgr, int argc, char **argv);
static McOptions_t * ParseInvarOptions(int argc, char **argv);
static void TimeOutHandle(void);
static mdd_t * McForwardReachable(Fsm_Fsm_t *modelFsm, mdd_t *targetMdd, mdd_t *invariantMdd, mdd_t *fairStates, array_t *careStatesArray, array_t *onionRings, Mc_VerbosityLevel verbosity, Mc_DcLevel dcLevel);
static void InvarPrintDebugInformation(Fsm_Fsm_t *modelFsm, array_t *invarFormulaArray, array_t *invarStatesArray, int *resultArray, McOptions_t *options, array_t *hintsStatesArray);
static array_t * SortFormulasByFsm(Fsm_Fsm_t *totalFsm, array_t *invarFormulaArray, array_t *fsmArray, McOptions_t *options);
static int UpdateResultArray(mdd_t *reachableStates, array_t *invarStatesArray, int *resultArray);
static void PrintMcPassFail(mdd_manager *mddMgr, Fsm_Fsm_t *modelFsm, Mc_FwdBwdAnalysis traversalDirection, Ctlp_Formula_t *ctlFormula, mdd_t *ctlFormulaStates, mdd_t *modelInitialStates, array_t *modelCareStatesArray, McOptions_t *options, Mc_VerbosityLevel verbosity);
static void PrintInvPassFail(array_t *invarFormulaArray, int *resultArray);
static int TestInvariantsInTotalFsm(Fsm_Fsm_t *totalFsm, array_t *invarStatesArray, int shellFlag);
static Ctlp_Approx_t ComputeResultingAproximation(Ctlp_FormulaType formulaType, Ctlp_Approx_t leftApproxType, Ctlp_Approx_t rightApproxType, Ctlp_Approx_t TRMinimization, Mc_EarlyTermination_t *earlyTermination, boolean fixpoint);

/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis [Initialize mc package]

  SideEffects []

******************************************************************************/
void
Mc_Init(void)
{
  Cmd_CommandAdd("model_check", CommandMc, /* doesn't changes_network */ 0);
  Cmd_CommandAdd("check_invariant", CommandInv, /* doesn't changes_network */ 0);
  Cmd_CommandAdd("lang_empty", CommandLe, /* doesn't changes_network */ 0);
  Cmd_CommandAdd("_init_state_formula", McCommandInitState, /* doesn't changes_network */ 0);
}


/**Function********************************************************************

  Synopsis [End mc package]

  SideEffects []

******************************************************************************/
void
Mc_End(void)
{
}

/**Function********************************************************************

  Synopsis [Compute all the states that lead to a fair cycle]

  Comment  [Used by LE/LTL/CTL* model checking. Different dicision procedures 
  are applied according to the "strength of the  buechi automaton": For the 
  strong automaton, we need to compute EG(true) under the fairness constraints;
  If the automaton is weak, we will check !EFEG(fair); If the automaton is 
  terminal, we only need to check !EF(fair). This is based on the CAV'99 paper
  of Bloem, Ravi & Somenzi: "Efficient Decision Procedures for Model Checking 
  of Linear Time Logic Properties".
  
  Debugging information will be print out if dbgLevel is non-zero.

  Return the set of initial states from where fair paths exist.]
  
  SideEffects [The caller should free the returned mdd_t.]

******************************************************************************/
mdd_t *
Mc_FsmCheckLanguageEmptiness(
  Fsm_Fsm_t *modelFsm,
  array_t *modelCareStatesArray,
  int automatonStrength,
  int dcLevel,
  int dbgLevel,
  int printInputs,
  int verbosity,
  Mc_GSHScheduleType GSHschedule,
  Mc_FwdBwdAnalysis GSHdirection,
  int lockstep,
  FILE *dbgFile,
  int UseMore,
  int SimValue,
  char *driverName)
{
  mdd_t *result;
  mdd_t *mddOne;
  mdd_t *badStates;
  mdd_t *aBadState;
  mdd_t *initialStates;
  mdd_t *fairStates;
  mdd_manager *mddMgr;
  Ctlp_Formula_t *ctlFair, *ctlFormula;
  Mc_EarlyTermination_t *earlyTermination;
  array_t *arrayOfOnionRings;
  array_t *stemArray;
  array_t *cycleArray;
  McPath_t *fairPath;
  McPath_t *normalPath;
  FILE *tmpFile = vis_stdout;
  extern FILE *vis_stdpipe;
  int i;
  
  /**************************************************************************
   * CASE1 & CASE2: TERMINAL/WEAK AUTOMATON
   *                problem is reduced to checking !EF (fair) for TERMINAL
   *                or !EF EG (fair) for WEAK
   **************************************************************************/
  if (automatonStrength == 0 || automatonStrength == 1) {
    Fsm_Fairness_t *modelFairness;
    modelFairness = Fsm_FsmReadFairnessConstraint(modelFsm);
    assert(Fsm_FairnessTestIsBuchi(modelFairness));
    ctlFair = Ctlp_FormulaCreate(Ctlp_ID_c,
				 (void *)util_strsav("fair1$AUT$NTK2"),
				 (void *)util_strsav("1"));
    /*    Dup(Fsm_FairnessReadFinallyInfFormula(modelFairness, 0, 0));    */
    if (automatonStrength == 0) {
      Ctlp_Formula_t *ctlTRUE;
      ctlTRUE = Ctlp_FormulaCreate(Ctlp_TRUE_c, NIL(Ctlp_Formula_t),
				   NIL(Ctlp_Formula_t));
      ctlFormula = Ctlp_FormulaCreate(Ctlp_EU_c, ctlTRUE, ctlFair);
      if (dbgLevel || verbosity) 
	fprintf(vis_stdout, "# %s: Terminal automaton -- check EF(states satisfy fairness)\n",
		driverName);
    }else {
      Ctlp_Formula_t *ctlTRUE, *ctlEGfair;
      ctlTRUE = Ctlp_FormulaCreate(Ctlp_TRUE_c, NIL(Ctlp_Formula_t), NIL(Ctlp_Formula_t));
      ctlEGfair = Ctlp_FormulaCreate(Ctlp_EG_c, ctlFair, NIL(Ctlp_Formula_t));
      ctlFormula = Ctlp_FormulaCreate(Ctlp_EU_c, ctlTRUE, ctlEGfair);
      if (dbgLevel || verbosity) 
	fprintf(vis_stdout, "# %s: Weak automaton -- check EF EG (states satisfy fairness)\n",
		driverName);
    }

    /* Evaluate ctlFormula */
    mddMgr = Fsm_FsmReadMddManager(modelFsm);
    mddOne = mdd_one(mddMgr);
    initialStates = Fsm_FsmComputeInitialStates(modelFsm);
    earlyTermination = Mc_EarlyTerminationAlloc(McSome_c, initialStates);
    fairStates = 
      Mc_FsmEvaluateFormula(modelFsm, ctlFormula, mddOne, modelFairness, 
			    modelCareStatesArray,
			    earlyTermination,
			    NIL(Fsm_HintsArray_t), Mc_None_c,
			    (Mc_VerbosityLevel) verbosity,
			    (Mc_DcLevel) dcLevel, 
			    (dbgLevel != McDbgLevelNone_c || verbosity),
			    GSHschedule);
    Mc_EarlyTerminationFree(earlyTermination);
    mdd_free(mddOne);

    /* If there is at least one fair initial state, language is not empty. */
    if (!mdd_lequal(initialStates, fairStates, 1, 0)) {
      if (strcmp(driverName, "LE") == 0)
	fprintf(vis_stdout, "# LE: language is not empty\n");
      else
	fprintf(vis_stdout, "# %s: formula failed\n", driverName);
      result = mdd_and(fairStates, initialStates, 1, 1);
    } else {
      if (strcmp(driverName, "LE") == 0)
	fprintf(vis_stdout, "# LE: language emptiness check passes\n");
      else
	fprintf(vis_stdout, "# %s: formula passed\n", driverName);
      Ctlp_FormulaFree(ctlFormula);
      mdd_free(fairStates);
      mdd_free(initialStates);
      return NIL(mdd_t);
    }

    /* If no wintness is requested, clean up and return. */
    if (dbgLevel == McDbgLevelNone_c) {
      Ctlp_FormulaFree(ctlFormula);
      mdd_free(initialStates);
      mdd_free(fairStates);
      return result;
    }else {
      Ctlp_Formula_t *F;
      array_t *EGArrayOfOnionRings, *EGonionRings, *EFonionRings;
      array_t *newOnionRings;
      mdd_t *mdd1;
      int j;
      arrayOfOnionRings = array_alloc(array_t *, 0);
      EFonionRings = (array_t *) Ctlp_FormulaReadDebugData(ctlFormula);
      if (automatonStrength == 0) {
	array_insert_last(array_t *, arrayOfOnionRings,
			  mdd_array_duplicate(EFonionRings));
      }else if (automatonStrength == 1) {
        F = Ctlp_FormulaReadRightChild(ctlFormula);
	/* EG (states labelled fair) */
	EGArrayOfOnionRings = (array_t *)Ctlp_FormulaReadDebugData(F);
	arrayForEachItem(array_t *, EGArrayOfOnionRings, i, EGonionRings) {
	  newOnionRings = mdd_array_duplicate(EGonionRings);
	  arrayForEachItem(mdd_t *, EFonionRings, j, mdd1) {
	    if (j != 0)
	      array_insert_last(mdd_t *, newOnionRings, mdd_dup(mdd1));
	  }
	  array_insert_last(array_t *, arrayOfOnionRings, newOnionRings);
	}
      }
      Ctlp_FormulaFree(ctlFormula);
      fprintf(vis_stdout, "# %s: generating path to fair cycle\n", driverName);
    }
    
  }else {
    /*************************************************************************
     * CASE3:     STRONG AUTOMATON 
     *            need to solve the problem by computing EG_fair (true) 
     *************************************************************************/
    mddMgr = Fsm_FsmReadMddManager(modelFsm);
    initialStates = Fsm_FsmComputeInitialStates(modelFsm);

    /* Compute fair states. */
    if (lockstep == MC_LOCKSTEP_OFF) {
      fairStates = Fsm_FsmComputeFairStates(modelFsm, modelCareStatesArray,
					    verbosity, dcLevel, GSHschedule,
					    GSHdirection, FALSE);
      arrayOfOnionRings = Fsm_FsmReadDebugArray(modelFsm);
    } else {
      /* For lockstep, the computed set of states is a subset of the
	 reachable fair states, with the property that the subset is
	 empty only if there are no reachable fair states.  Since the
	 onion rings are built for the fairness constraint restricted to
	 the subset of the fair states, they are not valid in general
	 for the FSM.  Hence, they are not saved.  */
      array_t *SCCs = array_alloc(mdd_t *, 0);
      arrayOfOnionRings = array_alloc(array_t *, 0);
      fairStates = McFsmComputeFairSCCsByLockStep(modelFsm, lockstep, SCCs,
						  arrayOfOnionRings,
						  (Mc_VerbosityLevel)
						  verbosity,
						  (Mc_DcLevel) dcLevel);
      mdd_array_free(SCCs);
    }

    
    /* Lockstep guarantees that all fair SCCs are reachable.  However,
       it does not extend the fair states backward unless a
       counterexample is requested.  For the SCC hull approach with backward
       operators, the language is not empty if there is at least one
       fair initial state.  Finally, For SCC hull with forward operators,
       currently we do not extend the hull backward, but we enforce
       reachability of the hull.  Hence, nonemptiness of the hull signals that
       the language is not empty. */
    if ((lockstep == MC_LOCKSTEP_OFF &&
	 (!mdd_lequal(initialStates, fairStates, 1, 0) ||
	  (GSHdirection == McFwd_c &&
	   !mdd_lequal_array(fairStates, modelCareStatesArray, 1, 0)))) ||
	(lockstep != MC_LOCKSTEP_OFF &&
	 !mdd_is_tautology(fairStates, 0))) {
      if (strcmp(driverName, "LE") == 0)
	fprintf(vis_stdout, "# LE: language is not empty\n");
      else
	fprintf(vis_stdout, "# %s: formula failed\n", driverName);
      result = mdd_and(fairStates, initialStates, 1, 1);
    } else {
      if (strcmp(driverName, "LE") == 0)
	fprintf(vis_stdout, "# LE: language emptiness check passes\n");
      else
	fprintf(vis_stdout, "# %s: formula passed\n", driverName);
      /* Dispose of onion rings if produced by lockstep. */
      if (lockstep != MC_LOCKSTEP_OFF) {
	mdd_array_array_free(arrayOfOnionRings);
      }
      mdd_free(fairStates);
      mdd_free(initialStates);
      return NIL(mdd_t);
    }

    /* If no witness is requested, clean up and return. */
    if (dbgLevel == McDbgLevelNone_c) {
      /* Dispose of onion rings if produced by lockstep. */
      if (lockstep != MC_LOCKSTEP_OFF) {
	mdd_array_array_free(arrayOfOnionRings);
      }
      mdd_free(fairStates);
      mdd_free(initialStates);
      return result;
    } else {
      (void) fprintf(vis_stdout, "# %s: generating path to fair cycle\n",
		     driverName);
    }
  }

  /* Generate witness. */
  badStates = mdd_and(initialStates, fairStates, 1, 1);
  mdd_free(fairStates);
  mdd_free(initialStates);
  aBadState = Mc_ComputeAState(badStates, modelFsm);
  mdd_free(badStates);

  mddOne = mdd_one(mddMgr);
  fairPath = McBuildFairPath(aBadState, mddOne, arrayOfOnionRings, modelFsm,
			     modelCareStatesArray,
			     (Mc_VerbosityLevel) verbosity,
			     (Mc_DcLevel) dcLevel,
			     McFwd_c);
  mdd_free(mddOne);
  mdd_free(aBadState);
  
  /* Dispose of onion rings if produced by lockstep. */
  if (lockstep != MC_LOCKSTEP_OFF || automatonStrength != 2) {
    mdd_array_array_free(arrayOfOnionRings);
  }

  /* Print witness. */
  normalPath = McPathNormalize(fairPath);
  McPathFree(fairPath);

  stemArray = McPathReadStemArray(normalPath);
  cycleArray = McPathReadCycleArray(normalPath);

  /* we should pass dbgFile/UseMore as parameters 
     dbgFile = McOptionsReadDebugFile(options);*/
  if (dbgFile != NIL(FILE)) {
    vis_stdpipe = dbgFile;
  } else if (UseMore == TRUE) {
    vis_stdpipe = popen("more","w");
  } else {
    vis_stdpipe = vis_stdout;
  }
  vis_stdout = vis_stdpipe;

  /* We should also pass SimValue as a parameter */
  if (SimValue == FALSE) {
      (void) fprintf(vis_stdout, "# %s: path to fair cycle:\n\n", driverName);
      Mc_PrintPath(stemArray, modelFsm, printInputs);
      fprintf (vis_stdout, "\n");

      (void) fprintf(vis_stdout, "# %s: fair cycle:\n\n", driverName);
      Mc_PrintPath(cycleArray, modelFsm, printInputs);
      fprintf (vis_stdout, "\n");
  }else {
      array_t *completePath = McCreateMergedPath(stemArray, cycleArray);
      (void) fprintf(vis_stdout, "# %s: counter-example\n", driverName);
      McPrintSimPath(vis_stdout, completePath, modelFsm);
      mdd_array_free(completePath);
  }

  if (dbgFile == NIL(FILE) && UseMore == TRUE) {
    (void) pclose(vis_stdpipe);
  }  
  vis_stdout = tmpFile;

  McPathFree(normalPath);

  return result;
}


/**Function********************************************************************

  Synopsis    [Model check formula on given FSM under fairness]

  Description [Model check formula on given FSM under fairness.]

  Comment [ Recursive evaluation using Emerson-Lei LICS 1986 algorithms. The
  states satisfying subformulas are stored and used later if there is sharing
  of subformulas. Therefore, each subformula is computed just once. Large
  switch - base case is atomic formula, TRUE, and FALSE.  Boolean operators -
  straightforward. Path modalities are reduced to one of the three - EX, EU,
  EG which are evaluated under fairness. Note the semantics for atomic
  formulas carefully - in addition to generating the prescribed output, the
  state must also be fair. Don't cares from unreached, unfair states are both
  used. The proof of correctness of using all these dont cares hinges on the
  invariant that restricted to the reached state set, the computed sets are
  exactly as they should be. Both backward and forward formulas are handled.

  This function sets the states and debug info of the formula for use
  with the debugger.
  
  ARGUMENTS:

  ctlFormula: a ctl formula in simple(!) existential form, using
  forward and/or backward operators.
  
  EarlyTermination: If earlyTermination.mode is McAll_c
  (McSome_c), the algorithm can stop as soon as

  1. it is sure that all (some) of earlyTermination.states are
  included in the evaluation of the formula, or

  2. It is sure that some (all) of these states are not included in
  the evaluation.
  
  If earlyTermination.mode is McCare_c, then the algorithm can stop as
  soon as it is guaranteed that either all or none of
  earlyTermination.states are included in the evaluation.

  These checks are done modulo the care set, but not modulo the fair
  states, which means that they will not be effective if you
  require unfair states to be in the evaluation.
  Typically {ALL, initialStates} will be good. NULL means the option has no
  effect.

  Debugging only works if this function is called with {ALL,
  initialStates}.  Otherwise, debugging of the EG formulae may go
  wrong.

  hintsArray is an array of hints.  Pass NULL and set hintType to
  Mc_None_c to ignore.

  If hintType is Mc_Global_c, then the formula is evaluated once for
  each hint, where every hint defines an underapproximation of the
  formula.  This only works for pure A/E CTL formulae.  If hintType is
  Mc_Local_c, then every fixpoint goes through the sequence of hints
  until completion, before moving up to the next temporal operator. 
  
  ]

  SideEffects []

******************************************************************************/
mdd_t *
Mc_FsmEvaluateFormula(
  Fsm_Fsm_t *fsm,
  Ctlp_Formula_t *ctlFormula,
  mdd_t *fairStates,
  Fsm_Fairness_t *fairCondition,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  Fsm_HintsArray_t *hintsArray,
  Mc_GuidedSearch_t hintType,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  int buildOnionRing,
  Mc_GSHScheduleType GSHschedule)
{
  mdd_t *hint;      /* to iterate over hintsArray */
  int   hintnr;     /* idem                       */
  Ctlp_Approx_t resultApproxType;

  assert(hintType == Mc_None_c || hintsArray != NIL(Fsm_HintsArray_t));
  
  if(hintType != Mc_Global_c)
    return EvaluateFormulaRecur(fsm, ctlFormula, fairStates,
				fairCondition, careStatesArray,
				earlyTermination, hintsArray, hintType,
				Ctlp_Exact_c, &resultApproxType,
				verbosity, dcLevel,
				buildOnionRing, GSHschedule);
  else{
    mdd_t *result = NIL(mdd_t);
    Ctlp_Approx_t approx;

    if(verbosity >= McVerbosityMax_c)
      fprintf(vis_stdout, "--Using global hints\n");
    
    arrayForEachItem(mdd_t *, hintsArray, hintnr, hint){
      if(result != NIL(mdd_t))
	mdd_free(result);
      
      if(verbosity >= McVerbosityMax_c)
	fprintf(vis_stdout, "--Instantiating global hint %d\n", hintnr);

      Fsm_InstantiateHint(fsm, hint, FALSE, TRUE, Ctlp_Underapprox_c,
			  (verbosity >= McVerbosityMax_c));

      approx = mdd_is_tautology(hint, 1) ? Ctlp_Exact_c : Ctlp_Underapprox_c;

      result = EvaluateFormulaRecur(fsm, ctlFormula, fairStates,
				    fairCondition, careStatesArray,
				    earlyTermination, hintsArray, hintType,
				    approx, &resultApproxType, verbosity,
				    dcLevel, buildOnionRing, GSHschedule); 
      /* TBD: take into account (early) termination here */
    }
    Fsm_CleanUpHints(fsm, FALSE, TRUE, (verbosity >= McVerbosityMax_c));

    return result;
  }

} /* Mc_FsmEvaluateFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EX target]

  Description [Evaluate states satisfying EX target.  Basically, calls
  the image package -- all the smarts are in there, parameters for it
  can be set from the VIS prompt.  In the presence of fairness, a
  state satisifies EX foo just in case it has a sucessor where foo is
  true AND that successor can be continued to an infinite path which
  is fair.]

  SideEffects []

******************************************************************************/
mdd_t *
Mc_FsmEvaluateEXFormula(
  Fsm_Fsm_t *modelFsm,
  mdd_t *targetMdd,
  mdd_t *fairStates,
  array_t *careStatesArray,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel)
{
  mdd_t *fromLowerBound;
  mdd_t *fromUpperBound;
  mdd_t *result;

  Img_ImageInfo_t * imageInfo = Fsm_FsmReadOrCreateImageInfo(modelFsm, 1, 0);

  /*
   * The lower bound is the conjunction of the fair states,
   * and the target states.
   */
  fromLowerBound = mdd_and(targetMdd, fairStates, 1, 1);
  /*
   * The upper bound is the same as the lower bound.
   */
  fromUpperBound = mdd_dup(fromLowerBound);

  result = Img_ImageInfoComputePreImageWithDomainVars(imageInfo,
		fromLowerBound, fromUpperBound, careStatesArray);
  mdd_free(fromLowerBound);
  mdd_free(fromUpperBound);

  if (verbosity == McVerbosityMax_c) {
    static int refCount=0;
    mdd_manager *mddMgr = Fsm_FsmReadMddManager(modelFsm);
    array_t *psVars = Fsm_FsmReadPresentStateVars(modelFsm);
    mdd_t *tmpMdd = careStatesArray ?
	mdd_and_array(result, careStatesArray, 1, 1) : mdd_dup(result);
    fprintf(vis_stdout, "--EX called %d (bdd_size - %d)\n", ++refCount,
	    mdd_size(result));
    fprintf(vis_stdout, "--There are %.0f care states satisfying EX formula\n",
	    mdd_count_onset(mddMgr, tmpMdd, psVars));
    mdd_free(tmpMdd);
  }

  return result;

} /* Mc_FsmEvaluateEXFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EY target]

  Description [Evaluate states satisfying EY target.]

  SideEffects []

******************************************************************************/
mdd_t *
Mc_FsmEvaluateEYFormula(
  Fsm_Fsm_t *modelFsm,
  mdd_t *targetMdd,
  mdd_t *fairStates,
  array_t *careStatesArray,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel)
{
  mdd_t *fromLowerBound;
  mdd_t *fromUpperBound;
  mdd_t *result;

  Img_ImageInfo_t * imageInfo = Fsm_FsmReadOrCreateImageInfo(modelFsm, 0, 1);

  /*
   * The lower bound is the conjunction of the fair states,
   * and the target states.
   */
  fromLowerBound = mdd_and(targetMdd, fairStates, 1, 1);
  /*
   * The upper bound is the same as the lower bound.
   */
  fromUpperBound = fromLowerBound;

  result = Img_ImageInfoComputeImageWithDomainVars(imageInfo,
		fromLowerBound, fromUpperBound, careStatesArray);
  mdd_free(fromLowerBound);

  if (verbosity == McVerbosityMax_c) {
    static int refCount=0;
    mdd_manager *mddMgr = Fsm_FsmReadMddManager(modelFsm);
    array_t *psVars = Fsm_FsmReadPresentStateVars(modelFsm);
    mdd_t *tmpMdd = careStatesArray ?
	mdd_and_array(result, careStatesArray, 1, 1) : mdd_dup(result);
    fprintf(vis_stdout, "--EY called %d (bdd_size - %d)\n", ++refCount,
	    mdd_size(result));
    fprintf(vis_stdout, "--There are %.0f care states satisfying EY formula\n",
	    mdd_count_onset(mddMgr, tmpMdd, psVars));
    mdd_free(tmpMdd);
  }

  return result;

} /* Mc_FsmEvaluateEYFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying E(invariant U target)]

  Description [Evaluate states satisfying E(invariant U target).
  Performed by a least fixed point computation -- start with target
  AND fair, see what can get to this set by a path solely in
  invariant.  If the onionRings array is not NIL(array_t *), the
  ``onion rings'' arising in the least fixed point computation are
  stored in the array, starting from the target.  The onion rings
  explain how to get from any state in the result to the target.

  The evaluation stops once all (some) of earlyTermination.states
  have been found (depending on earlyTermination.mode).  See
  Mc_FsmEvaluateFormula()

  If underapprox is not NIL, it must hold a valid underapproximation
  of the fixpoint.  This is used to kick off the new computation.

  The arguments hintsArray and hints are used for guided search.  Set
  hintType to Mc_None_c to ignore.  If hintType is Mc_Local_c,
  hintsArray should be an array of mdds, of which the last entry is 1.
  This array will be used to obtain underapproximations of the
  transition system.  (See Mc_FsmEvaluateFormula.)

  fixpoint is a return argument.  It is true if the formula is
  evaluated exactly, false if early termination kicked in.  If
  earlyTermination is NIL, the formula is evaluated exacty, and this
  argument may be NIL too.

  OnionRings is an array of mdd_t *.  It can be NULL, in which case no
  onion rings are returned.  If it is not NULL, there are two
  possibilities.  In the simple case, the onion rings are empty, and
  underapprox is NULL.  In this case, the onion rings returned explain
  how to get to the target from any state in the result returned (r).
  If underapprox u is not NULL, the onion rings that are passed in
  should explain how to get from any state in u to any state in the
  target.  Rings will be added by the procudure to explain how to get
  from the r to a state in underapprox, resulting in an explanation of
  how to get from a state in r to the target.  ]

  Comment     [Check that evaluation mod reached AND fair will lead to
  faster convergence.  Also, it is possible there is more (?)
  flexibility in dont care minimization - use only the layer of the
  onion rings as set for evaluating EX.]

  SideEffects []

  SeeAlso [Mc_FsmEvaluateFormula, McEvaluateEUFormulaWithGivenTR]

******************************************************************************/
mdd_t *
Mc_FsmEvaluateEUFormula(
  Fsm_Fsm_t *fsm,
  mdd_t *invariant,
  mdd_t *target,
  mdd_t *underapprox,
  mdd_t *fairStates,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  Fsm_HintsArray_t *hintsArray,
  Mc_GuidedSearch_t hintType,
  array_t *onionRings,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint)
{
  /* should be strengthened to check that rings supply a correct explanation.
   */
  assert(underapprox == NULL || onionRings == NULL ||
	 array_n(onionRings) != 0);  
 
  if(hintType != Mc_Local_c)
    /* postcondition: rings should explain how to get from return value to
     * target.
     */
    return McEvaluateEUFormulaWithGivenTR(fsm, invariant, target,
					  underapprox, fairStates,
					  careStatesArray,
					  earlyTermination, onionRings,
					  verbosity, dcLevel, fixpoint);  
  else{
    mdd_t *iterate, *newiterate;
    mdd_t *hint;
    int hintnr;

    if(verbosity >= McVerbosityMax_c)
	fprintf(vis_stdout, "--Using local hints\n");

    if(underapprox != NIL(mdd_t))
      iterate = mdd_dup(underapprox);
    else
      iterate = mdd_zero(Fsm_FsmReadMddManager(fsm));

    arrayForEachItem(mdd_t *, hintsArray, hintnr, hint){
	if(verbosity >= McVerbosityMax_c)
	    fprintf(vis_stdout, "--Instantiating local hint %d\n", hintnr);
      Fsm_InstantiateHint(fsm, hint, FALSE, TRUE, Ctlp_Underapprox_c,
 			  (verbosity >= McVerbosityMax_c));
      
      newiterate =
	McEvaluateEUFormulaWithGivenTR(fsm, invariant, target, iterate,
				       fairStates, careStatesArray,
				       earlyTermination, onionRings,
				       verbosity, dcLevel, fixpoint);    
      mdd_free(iterate);
      iterate = newiterate;

      /* check if we can terminate without considering all hints */ 
      if(mdd_is_tautology(iterate, 1)){
	*fixpoint = 1;
	break;
      }
      if(McCheckEarlyTerminationForUnderapprox(earlyTermination, iterate,
					       careStatesArray)){ 
	*fixpoint = 0;
	break;
      }

    }/* for all hints */
    Fsm_CleanUpHints(fsm, FALSE, TRUE, (verbosity >= McVerbosityMax_c));

    /* postcondition: rings should explain how to get from return value to
     * target.
     */
    return iterate;
  }/* if hinttype */

} /* Mc_FsmEvaluateEUFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying E(invariant S source)]

  Description [Evaluate states satisfying E(invariant U target).
  Performed by a least fixed point computation -- start with source
  AND fair, see what can be reached from this set by a path solely in
  invariant.  If the onionRings array is not NIL(array_t *), the
  ``onion rings'' arising in the least fixed point computation are
  stored in the array, starting from the target.

  The evaluation stops once all (some) of earlyTermination.states
  have been found (depending on earlyTermination.mode).  See
  Mc_FsmEvaluateFormula()

  If underapprox is not NIL, it must hold a valid underapproximation
  of the fixpoint.  This is used to kick off the new computation.

  The arguments hintsArray and hints are used for guided search.  Set
  hintType to Mc_None_c to ignore.  If hintType is Mc_Local_c,
  hintsArray should be an array of mdds, of which the last entry is 1.
  This array will be used to obtain underapproximations of the
  transition system.  (See Mc_FsmEvaluateFormula.)

  fixpoint is a return argument.  It is true if the formula is
  evaluated exactly, false if early termination kicked in.  If
  earlyTermination is NIL, the formula is evaluated exacty, and this
  argument may be NIL too.

  OnionRIngs is an array of mdd_t *.  It should be empty on call if
  you want debug info, NULL otherwise.  After the call, it contains
  the onion rings (frontier sets) that occured while computing the
  ES.  This is used for debugging.]

  Comment [Check that evaluation mod reached AND fair will lead to
  faster convergence.  Also, it is possible there is more (?)
  flexibility in don't care minimization - use only the layer of the
  onion rings as set for evaluating EX.]

  SideEffects []

  SeeAlso [Mc_FsmEvaluateFormula, McEvaluateESFormulaWithGivenTR]

******************************************************************************/
mdd_t *
Mc_FsmEvaluateESFormula(
  Fsm_Fsm_t *fsm,
  mdd_t *invariant,
  mdd_t *source,
  mdd_t *underapprox,
  mdd_t *fairStates,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  Fsm_HintsArray_t *hintsArray,
  Mc_GuidedSearch_t hintType,
  array_t *onionRings,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint)
{
  /* should be strengthened to check that rings supply a correct explanation.
   */
  assert(underapprox == NULL || onionRings == NULL ||
	 array_n(onionRings) != 0);

  if(hintType != Mc_Local_c)
    /* postcondition: rings should explain how to get from return value to
     * target.
     */
    return McEvaluateESFormulaWithGivenTR(fsm, invariant, source,
					  underapprox, fairStates,
					  careStatesArray,
					  earlyTermination, onionRings,
					  verbosity, dcLevel, fixpoint);  
  else{
    mdd_t *iterate, *newiterate;
    mdd_t *hint;
    int hintnr;

    if(verbosity >= McVerbosityMax_c)
	fprintf(vis_stdout, "--Using local hints\n");

    if(underapprox != NIL(mdd_t))
      iterate = mdd_dup(underapprox);
    else
      iterate = mdd_zero(Fsm_FsmReadMddManager(fsm));

    arrayForEachItem(mdd_t *, hintsArray, hintnr, hint){
	if(verbosity >= McVerbosityMax_c)
	    fprintf(vis_stdout, "--Instantiating local hint %d\n", hintnr);
      Fsm_InstantiateHint(fsm, hint, FALSE, TRUE, Ctlp_Underapprox_c,
 			  (verbosity >= McVerbosityMax_c));
      
      newiterate =
	McEvaluateESFormulaWithGivenTR(fsm, invariant, source, iterate,
				       fairStates, careStatesArray,
				       earlyTermination, onionRings,
				       verbosity, dcLevel, fixpoint);    
      mdd_free(iterate);
      iterate = newiterate;

      /* check if we can terminate without considering all hints */ 
      if(mdd_is_tautology(iterate, 1)){
	*fixpoint = 1;
	break;
      }
      if(McCheckEarlyTerminationForUnderapprox(earlyTermination, iterate,
					       careStatesArray)){ 
	*fixpoint = 0;
	break;
      }

    }/* for all hints */
    Fsm_CleanUpHints(fsm, FALSE, TRUE, (verbosity >= McVerbosityMax_c));

    /* postcondition: rings should explain how to get from return value to
     * target.
     */

    return iterate;
  }/* if hinttype */

} /* Mc_FsmEvaluateESFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EG invariant]

  Description [Evaluate states satisfying EG invariant.  This is done
  by starting with all states marked with the invariant.  From this
  initial set, recursively remove states that can not reach Buechi
  contraints (through paths entirely within the current set).

  If overapprox is not nil, then it is a valid overapproximation to the
  satisfiable set of the formula that will be used to kick off the
  computation.

  pOnionRingsArrayForDbg is a pointer to (a pointer to) an array of
  arrays of mdd_t *s.  Upon calling this formula, it should point to
  an empty array if you want debug information, or be NIL (or a
  pointer to NIL) if you do not.  After the call, it points to an
  array that contains one entry for every fairness constraint: entry i
  contains the onionrings for the last EU computation that was
  performed for fairness constraint i.  There are two exceptions: if
  the satisfying set is 0, or if early termination kicks in, the onion
  ring array is not changed.

  The arguments hintsArray and hints are used for guided search.  Set
  hintType to Mc_None_c to ignore.  If hintType is Mc_Local_c,
  hintsArray should be an array of mdds, of which the last entry is 1.
  This array will be used to obtain overapproximations of the
  transition system (See Mc_FsmEvaluateFormula).

  fixpoint is a return argument.  It is true if the formula is
  evaluated exactly, false if early termination kicked in.  If
  earlyTermination is NIL, the formula is evaluated exacty, and this
  argument may be NIL too.  See Mc_FsmEvaluateFormula for a discussion
  of early termination.]

  SideEffects []

  SeeAlso     [Mc_FsmEvaluateFormula]

******************************************************************************/
mdd_t *
Mc_FsmEvaluateEGFormula(
  Fsm_Fsm_t *fsm,
  mdd_t *invariant,
  mdd_t *overapprox,
  mdd_t *fairStates,
  Fsm_Fairness_t *modelFairness,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  Fsm_HintsArray_t *hintsArray,
  Mc_GuidedSearch_t hintType,
  array_t **pOnionRingsArrayForDbg,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint,
  Mc_GSHScheduleType GSHschedule)
{
  /* Illegal to pass in onion rings. */
  assert(pOnionRingsArrayForDbg == NIL(array_t *) ||
	 *pOnionRingsArrayForDbg == NIL(array_t) ||
	 array_n(*pOnionRingsArrayForDbg) == 0);

  if(hintType != Mc_Local_c)
    if (GSHschedule == McGSH_old_c) {
      return McEvaluateEGFormulaWithGivenTR(fsm, invariant, overapprox,
					    fairStates,
					    modelFairness, careStatesArray,
					    earlyTermination,
					    pOnionRingsArrayForDbg,
					    verbosity,
					    dcLevel, fixpoint);
    } else {
      return McFsmEvaluateEGFormulaUsingGSH(fsm, invariant, overapprox,
					    fairStates,
					    modelFairness, careStatesArray,
					    earlyTermination,
					    pOnionRingsArrayForDbg, verbosity,
					    dcLevel, fixpoint, GSHschedule);
    }
  else {
    mdd_t *iterate, *newiterate;
    mdd_t *hint;
    int hintnr;
    boolean useRings;

    useRings = (pOnionRingsArrayForDbg != NIL(array_t *) &&
		*pOnionRingsArrayForDbg != NIL(array_t));
    
    if(verbosity >= McVerbosityMax_c)
	fprintf(vis_stdout, "** mc Info: Using local hints\n");

    if(overapprox != NIL(mdd_t))
      iterate = mdd_dup(overapprox);
    else
      iterate = mdd_one(Fsm_FsmReadMddManager(fsm));
    
    arrayForEachItem(mdd_t *, hintsArray, hintnr, hint){
	if(verbosity >= McVerbosityMax_c)
	    fprintf(vis_stdout, "--Instantiating local hint %d\n", hintnr);
      Fsm_InstantiateHint(fsm, hint, FALSE, TRUE, Ctlp_Overapprox_c,
 			  (verbosity >= McVerbosityMax_c));

      /* old onionrings go stale.  Fry 'em */
      if(useRings){
	mdd_array_array_free(*pOnionRingsArrayForDbg);
	*pOnionRingsArrayForDbg = array_alloc(array_t *, 0);
      }

      if (GSHschedule == McGSH_old_c) {
	newiterate =
	  McEvaluateEGFormulaWithGivenTR(fsm, invariant, iterate, fairStates,
					 modelFairness, careStatesArray,
					 earlyTermination,
					 pOnionRingsArrayForDbg,
					 verbosity, dcLevel, fixpoint);
      } else {
	newiterate =
	  McFsmEvaluateEGFormulaUsingGSH(fsm, invariant, iterate,
					 fairStates,
					 modelFairness, careStatesArray,
					 earlyTermination,
					 pOnionRingsArrayForDbg, verbosity,
					 dcLevel, fixpoint, GSHschedule);
      }
      mdd_free(iterate);
      iterate = newiterate;

      /* check if we can terminate without considering all hints */ 
      if(mdd_is_tautology(iterate, 0)){
	*fixpoint = 1;
	break;
      }
      if(McCheckEarlyTerminationForOverapprox(earlyTermination, iterate,
					      careStatesArray)){  
	*fixpoint = 0;
	break;
      }

    } /* For all hints */
    Fsm_CleanUpHints(fsm, FALSE, TRUE, (verbosity >= McVerbosityMax_c));

    return iterate;
  }/* if hinttype */

} /* Mc_FsmEvaluateEGFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EH invariant.]

  Description [Evaluate states satisfying EH invariant.  This is done
  by starting with all states marked with the invariant.  From this
  initial set, recursively remove states that cannot reach Buechi
  contraints (through paths entirely within the current set).

  If overapprox is not nil, then it is a valid overapproximation to the
  satisfiable set of the formula that will be used to kick off the
  computation.

  pOnionRingsArrayForDbg is a pointer to (a pointer to) an array of
  arrays of mdd_t *s.  Upon calling this formula, it should point to
  an empty array if you want debug information, or be NIL (or a
  pointer to NIL) if you do not.  After the call, it points to an
  array that contains one entry for every fairness constraint: entry i
  contains the onionrings for the last ES computation that was
  performed for fairness constraint i.  There are two exceptions: if
  the satisfying set is 0, or if early termination kicks in, the onion
  ring array is not changed.

  The arguments hintsArray and hints are used for guided search.  Set
  hintType to Mc_None_c to ignore.  If hintType is Mc_Local_c,
  hintsArray should be an array of mdds, of which the last entry is 1.
  This array will be used to obtain overapproximations of the
  transition system (See Mc_FsmEvaluateFormula).

  fixpoint is a return argument.  It is true if the formula is
  evaluated exactly, false if early termination kicked in.  If
  earlyTermination is NIL, the formula is evaluated exacty, and this
  argument may be NIL too.  See Mc_FsmEvaluateFormula for a discussion
  of early termination.]

  SideEffects []

  SeeAlso [Mc_FsmEvaluateFormula]

******************************************************************************/
mdd_t *
Mc_FsmEvaluateEHFormula(
  Fsm_Fsm_t *fsm,
  mdd_t *invariant,
  mdd_t *overapprox,
  mdd_t *fairStates,
  Fsm_Fairness_t *modelFairness,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  Fsm_HintsArray_t *hintsArray,
  Mc_GuidedSearch_t hintType,
  array_t **pOnionRingsArrayForDbg,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint,
  Mc_GSHScheduleType GSHschedule)
{
  /* Illegal to pass in onion rings. */
  assert(pOnionRingsArrayForDbg == NIL(array_t *) ||
	 *pOnionRingsArrayForDbg == NIL(array_t) ||
	 array_n(*pOnionRingsArrayForDbg) == 0);

  if(hintType != Mc_Local_c)
    if (GSHschedule == McGSH_old_c) {
      return McEvaluateEHFormulaWithGivenTR(fsm, invariant, overapprox,
					    fairStates,
					    modelFairness, careStatesArray,
					    earlyTermination,
					    pOnionRingsArrayForDbg,
					    verbosity,
					    dcLevel, fixpoint);
    } else {
      return McFsmEvaluateEHFormulaUsingGSH(fsm, invariant, overapprox,
					    fairStates,
					    modelFairness, careStatesArray,
					    earlyTermination,
					    pOnionRingsArrayForDbg, verbosity,
					    dcLevel, fixpoint, GSHschedule);
    }
  else {
    mdd_t *iterate, *newiterate;
    mdd_t *hint;
    int hintnr;
    boolean useRings;

    useRings = (pOnionRingsArrayForDbg != NIL(array_t *) &&
		*pOnionRingsArrayForDbg != NIL(array_t));
    
    if(verbosity >= McVerbosityMax_c)
	fprintf(vis_stdout, "** mc Info: Using local hints\n");

    if(overapprox != NIL(mdd_t))
      iterate = mdd_dup(overapprox);
    else
      iterate = mdd_one(Fsm_FsmReadMddManager(fsm));
    
    arrayForEachItem(mdd_t *, hintsArray, hintnr, hint){
	if(verbosity >= McVerbosityMax_c)
	    fprintf(vis_stdout, "--Instantiating local hint %d\n", hintnr);
      Fsm_InstantiateHint(fsm, hint, FALSE, TRUE, Ctlp_Overapprox_c,
 			  (verbosity >= McVerbosityMax_c));

      /* old onionrings go stale.  Fry 'em */
      if(useRings){
	mdd_array_array_free(*pOnionRingsArrayForDbg);
	*pOnionRingsArrayForDbg = array_alloc(array_t *, 0);
      }

      if (GSHschedule == McGSH_old_c) {
	newiterate =
	  McEvaluateEHFormulaWithGivenTR(fsm, invariant, iterate, fairStates,
					 modelFairness, careStatesArray,
					 earlyTermination,
					 pOnionRingsArrayForDbg,
					 verbosity, dcLevel, fixpoint);
      } else {
	newiterate =
	  McFsmEvaluateEHFormulaUsingGSH(fsm, invariant, iterate,
					 fairStates,
					 modelFairness, careStatesArray,
					 earlyTermination,
					 pOnionRingsArrayForDbg, verbosity,
					 dcLevel, fixpoint, GSHschedule);
      }
      mdd_free(iterate);
      iterate = newiterate;

      /* check if we can terminate without considering all hints */ 
      if(mdd_is_tautology(iterate, 0)){
	*fixpoint = 1;
	break;
      }
      if(McCheckEarlyTerminationForOverapprox(earlyTermination, iterate,
					      careStatesArray)){  
	*fixpoint = 0;
	break;
      }

    } /* For all hints */
    Fsm_CleanUpHints(fsm, FALSE, TRUE, (verbosity >= McVerbosityMax_c));

    return iterate;
  }/* if hinttype */

} /* Mc_FsmEvaluateEHFormula */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying FwdUntil]

  Description [Evaluate states satisfying FwdUntil.]

  Comment []

  SideEffects []

******************************************************************************/
mdd_t *
Mc_FsmEvaluateFwdUFormula(
  Fsm_Fsm_t *modelFsm,
  mdd_t *targetMdd,
  mdd_t *invariantMdd,
  mdd_t *fairStates,
  array_t *careStatesArray,
  array_t *onionRings,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel)
{
  mdd_t *aMdd;
  mdd_t *bMdd;
  mdd_t *cMdd;
  mdd_t *resultMdd;
  mdd_t *ringMdd;
  int nImgComps;

  /*							    t
  ** E[p U q]      = lfp Z[q V (p ^ EX(Z))]   :   p p ... p q
  ** FwdUntil(p,q) = lfp Z[p V EY(Z ^ q)]     :		    pq q q ... q
  */

  nImgComps = Img_GetNumberOfImageComputation(Img_Forward_c);
  resultMdd = mdd_and(targetMdd, fairStates, 1, 1);
  ringMdd = mdd_dup(resultMdd);

  /* if until is trivially satisfied, */
  if (onionRings) {
    array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
  }
  if (mdd_is_tautology(fairStates, 1)) {
    if (mdd_equal_mod_care_set_array(invariantMdd, resultMdd, careStatesArray))
    { 
      bdd_free(ringMdd);
      return(resultMdd);
    }
  } else {
    bdd_t *trivialCheck = mdd_and(invariantMdd, fairStates, 1, 1);
    if (mdd_equal_mod_care_set_array(trivialCheck, resultMdd, careStatesArray)) {
      bdd_free(trivialCheck);
      bdd_free(ringMdd);
      return(resultMdd);
    }
    bdd_free(trivialCheck);
  }

  while (TRUE) {
    aMdd = mdd_and(ringMdd, invariantMdd, 1, 1);
    mdd_free(ringMdd);
    bMdd = Mc_FsmEvaluateEYFormula(modelFsm, aMdd, fairStates, careStatesArray,
				   verbosity, dcLevel);
    mdd_free(aMdd);
    cMdd = mdd_or(resultMdd, bMdd, 1, 1);
    mdd_free(bMdd);

    if (mdd_equal_mod_care_set_array(cMdd, resultMdd, careStatesArray)) {
      mdd_free(cMdd);
      break;
    }

    if (dcLevel >= McDcLevelRchFrontier_c) {
      mdd_t *tmpMdd = mdd_and(resultMdd, cMdd, 0, 1);
      ringMdd = bdd_between(tmpMdd, cMdd);
      if (verbosity == McVerbosityMax_c) {
	fprintf(vis_stdout,
	  "-- FwdU |A(n+1)-A(n)| = %d\t|A(n+1)| = %d\t|between-dc| = %d\n",
	  mdd_size(tmpMdd), mdd_size(resultMdd), mdd_size(ringMdd));
      }
      mdd_free(tmpMdd);
    }
    else {
      ringMdd = mdd_dup(cMdd);
      if (verbosity == McVerbosityMax_c) {
	fprintf(vis_stdout, "-- FwdU |ringMdd| = %d\n", mdd_size(ringMdd));
      }
    }

    mdd_free(resultMdd);
    resultMdd = cMdd;
    if (onionRings) {
      array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
    }
  } /* while(!termination) for LFP */

  /* Print some debug info */
  if (verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    static int refCount=0;
    mdd_manager *mddMgr = Fsm_FsmReadMddManager(modelFsm);
    array_t *psVars = Fsm_FsmReadPresentStateVars(modelFsm);

    if (verbosity == McVerbosityMax_c) {
      mdd_t *tmpMdd = careStatesArray ?
		      mdd_and_array(resultMdd, careStatesArray, 1, 1) :
		      mdd_dup(resultMdd);
      fprintf(vis_stdout, "--FwdU called %d (bdd_size - %d)\n",
		++refCount, mdd_size(resultMdd));
      fprintf(vis_stdout,
		"--There are %.0f care states satisfying FwdU formula\n",
		mdd_count_onset(mddMgr, tmpMdd, psVars));
#ifdef DEBUG_MC
      /* The following 2 lines are just for debug */
      fprintf(vis_stdout, "FwdU satisfying minterms :\n");
      (void)_McPrintSatisfyingMinterms(tmpMdd, modelFsm);
#endif
      mdd_free(tmpMdd);
    } else {
      fprintf(vis_stdout,
		"--There are %.0f states satisfying FwdU formula\n",
		mdd_count_onset(mddMgr, resultMdd, psVars));
    }
    fprintf(vis_stdout, "--FwdU: %d image computations\n",
      Img_GetNumberOfImageComputation(Img_Forward_c) - nImgComps);
  }

  return resultMdd;
}


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying FwdG invariant]

  Description [Evaluate states satisfying FwdG invariant.]

  SideEffects []

******************************************************************************/
mdd_t *
Mc_FsmEvaluateFwdGFormula(
  Fsm_Fsm_t *modelFsm,
  mdd_t *targetMdd,
  mdd_t *invariantMdd,
  mdd_t *fairStates,
  Fsm_Fairness_t *modelFairness,
  array_t *careStatesArray,
  array_t *onionRingsArrayForDbg,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel)
{
  mdd_t *resultMdd, *invariant;
  array_t *onionRings;
  int nImgComps;

  nImgComps = Img_GetNumberOfImageComputation(Img_Forward_c);
  onionRings = NIL(array_t);

  invariant = McForwardReachable(modelFsm, targetMdd, invariantMdd, fairStates,
				 careStatesArray, onionRings, verbosity,
				 dcLevel);
  resultMdd = Mc_FsmEvaluateFwdEHFormula(modelFsm, invariant, fairStates,
					 modelFairness, careStatesArray,
					 onionRingsArrayForDbg, verbosity,
					 dcLevel);
  mdd_free(invariant);

  if (verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    fprintf(vis_stdout, "--FwdG: %d image computations\n",
      Img_GetNumberOfImageComputation(Img_Forward_c) - nImgComps);
  }
#ifdef DEBUG_MC
  /* The following 2 lines are just for debug */
  if ((verbosity == McVerbosityMax_c)) {
    fprintf(vis_stdout, "FwdG satisfying minterms :\n");
    (void)_McPrintSatisfyingMinterms(resultMdd, modelFsm);
  }
#endif
  return resultMdd;
}



/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EH invariant]

  Description [Evaluate states satisfying EH invariant.]

  SideEffects []

******************************************************************************/
mdd_t *
Mc_FsmEvaluateFwdEHFormula(
  Fsm_Fsm_t *modelFsm,
  mdd_t *invariantMdd,
  mdd_t *fairStates,
  Fsm_Fairness_t *modelFairness,
  array_t *careStatesArray,
  array_t *onionRingsArrayForDbg,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel)
{
  int i;
  array_t *onionRings = NIL(array_t);
  array_t *tmpOnionRingsArrayForDbg = NIL(array_t);
  mdd_manager *mddManager = Fsm_FsmReadMddManager(modelFsm);
  mdd_t *mddOne = mdd_one(mddManager);
  mdd_t *Zmdd;
  int nIterations = 0;
  int nImgComps = Img_GetNumberOfImageComputation(Img_Forward_c);

  /*
  **  EG(f) = gfp Z[f ^ EX(Z)]
  **  EH(f) = gfp Z[f ^ EY(Z)]
  **
  **  EcG(f) = gfp Z[f ^ EX( ^ E[Z U Z ^ c])]
  **                        c<C
  **  EcH(f) = gfp Z[f ^ EY( ^ Reachable(c,Z))]
  **                        c<C
  **  Reachable(p,q) = FwdUntil(p,q) ^ q
  */

  array_t *buchiFairness = array_alloc(mdd_t *, 0);
  if (modelFairness) {
    if (!Fsm_FairnessTestIsBuchi(modelFairness)) {
      (void) fprintf(vis_stdout,
	 "** mc error: non-Buechi fairness constraints not supported\n");
      return NIL(mdd_t);
    }
    else {
      int j;
      int numBuchi = Fsm_FairnessReadNumConjunctsOfDisjunct(modelFairness, 0);
      for (j = 0; j < numBuchi; j++) {
	mdd_t *tmpMdd = Fsm_FairnessObtainFinallyInfMdd(modelFairness, 0, j,
						careStatesArray, dcLevel);
	array_insert_last(mdd_t *, buchiFairness, tmpMdd);
      }
    }
  }
  else {
    array_insert_last(mdd_t *, buchiFairness, mdd_one(mddManager));
  }

  if (onionRingsArrayForDbg !=NIL(array_t)) {
    tmpOnionRingsArrayForDbg = array_alloc(array_t *, 0);
  }

  Zmdd = mdd_dup(invariantMdd);
  while (TRUE) {
    mdd_t *ZprimeMdd = mdd_dup(Zmdd);
    mdd_t *conjunctsMdd = NIL(mdd_t);
    mdd_t *AAmdd = mdd_dup(Zmdd);

    nIterations++;
    for (i = 0; i < array_n(buchiFairness); i++) {
      mdd_t *aMdd, *bMdd;
      mdd_t *FiMdd = array_fetch(mdd_t *, buchiFairness, i);

      if (tmpOnionRingsArrayForDbg) {
	onionRings = array_alloc(mdd_t *, 0);
	array_insert_last(array_t *, tmpOnionRingsArrayForDbg, onionRings);
      }

      aMdd = McForwardReachable(modelFsm, FiMdd, AAmdd, mddOne, careStatesArray,
				onionRings, verbosity, dcLevel);
      bMdd = Mc_FsmEvaluateEYFormula(modelFsm, aMdd, mddOne, careStatesArray,
				     verbosity, dcLevel);
      mdd_free(aMdd);

      if (conjunctsMdd == NIL(mdd_t)) {
	conjunctsMdd = mdd_dup(bMdd);
      }
      else {
	mdd_t *tmpMdd = mdd_and(bMdd, conjunctsMdd, 1, 1);
	mdd_free(conjunctsMdd);
	conjunctsMdd = tmpMdd;
      }
      mdd_free(AAmdd); AAmdd = mdd_and(conjunctsMdd, invariantMdd, 1, 1);
      mdd_free(bMdd);
    }
    mdd_free(AAmdd);

    mdd_free(ZprimeMdd);
    ZprimeMdd = mdd_and(invariantMdd, conjunctsMdd, 1, 1);
    mdd_free(conjunctsMdd);

    if (mdd_equal_mod_care_set_array(ZprimeMdd, Zmdd, careStatesArray)) {
      mdd_free(ZprimeMdd);
      break;
    }

    mdd_free(Zmdd);
    Zmdd = ZprimeMdd;

    if (tmpOnionRingsArrayForDbg) {
      arrayForEachItem(array_t *, tmpOnionRingsArrayForDbg, i, onionRings) {
	mdd_array_free(onionRings);
      }
      array_free(tmpOnionRingsArrayForDbg);
      tmpOnionRingsArrayForDbg = array_alloc(array_t *, 0);
    }
  }

  if (onionRingsArrayForDbg != NIL(array_t)) {
    arrayForEachItem(array_t *, tmpOnionRingsArrayForDbg, i, onionRings) {
      array_insert_last(array_t *, onionRingsArrayForDbg, onionRings);
    }
    array_free(tmpOnionRingsArrayForDbg);
  }

  if (verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    if (onionRingsArrayForDbg) {
      for (i = 0; i < array_n(onionRingsArrayForDbg); i++) {
	int j;
	mdd_t *Fi = array_fetch(mdd_t *, buchiFairness, i);
	array_t *onionRings = array_fetch(array_t *, onionRingsArrayForDbg, i);
	for (j = 0; j < array_n(onionRings); j++) {
	  mdd_t *ring = array_fetch(mdd_t *, onionRings, j);
	  if (j == 0) {
	    if (!mdd_lequal(ring, Fi, 1, 1)) {
	      fprintf(vis_stderr, "** mc error: Problem w/ dbg in EH - ");
	      fprintf(vis_stderr,
		      "inner most ring not in Fi(fairness constraint).\n");
	    }
	  }
	  if (!mdd_lequal(ring, Zmdd, 1, 1)) {
	    fprintf(vis_stderr, "** mc error: Problem w/ dbg in EH - ");
	    fprintf(vis_stderr, "onion ring of last FwdU fails invariant\n");
	  }
	}
      }
    }

    if (verbosity == McVerbosityMax_c) {
      mdd_t *tmpMdd = careStatesArray ?
		      mdd_and_array(Zmdd, careStatesArray, 1, 1) :
		      mdd_dup(Zmdd);
      fprintf(vis_stdout,
	      "--There are %.0f care states satisfying EH formula\n",
	      mdd_count_onset(Fsm_FsmReadMddManager(modelFsm), tmpMdd,
			      Fsm_FsmReadPresentStateVars(modelFsm)));
#ifdef DEBUG_MC
      /* The following 2 lines are just for debug */
      fprintf(vis_stdout, "EH satisfying minterms :\n");
      (void)_McPrintSatisfyingMinterms(tmpMdd, modelFsm);
#endif
      mdd_free(tmpMdd);
    } else {
      fprintf(vis_stdout, "--There are %.0f states satisfying EH formula\n",
	      mdd_count_onset(Fsm_FsmReadMddManager(modelFsm), Zmdd,
			      Fsm_FsmReadPresentStateVars(modelFsm)));
    }

    fprintf(vis_stdout, "--EH: %d iterations, %d image computations\n",
      nIterations, Img_GetNumberOfImageComputation(Img_Forward_c) - nImgComps);
  }

  mdd_array_free(buchiFairness);
  mdd_free(mddOne);

  return Zmdd;

} /* Mc_FsmEvaluateFwdEHFormula */


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis    [Evaluate states satisfying E(invariant U target), given
  a transition relation]

  Description [Does the same as Mc_FsmEvaluateEUFormula, minus the hints.]

  SeeAlso     [Mc_FsmEvaluateEUFormula]
  
  SideEffects []

******************************************************************************/
mdd_t *
McEvaluateEUFormulaWithGivenTR(
  Fsm_Fsm_t *modelFsm,
  mdd_t *invariantMdd,
  mdd_t *targetMdd,
  mdd_t *underapprox,
  mdd_t *fairStates,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  array_t *onionRings,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint
  )
{
  mdd_t *aMdd;
  mdd_t *bMdd;
  mdd_t *cMdd;
  mdd_t *resultMdd;
  mdd_t *ringMdd;
  int nPreComps;
  boolean term_tautology = FALSE; /* different termination conditions */
  boolean term_early     = FALSE; 
  boolean term_fixpoint  = FALSE; 

  nPreComps = Img_GetNumberOfImageComputation(Img_Backward_c);
  resultMdd = mdd_and(targetMdd, fairStates, 1, 1);

  if(underapprox != NIL(mdd_t)){
    mdd_t *tmp = mdd_or(resultMdd, underapprox, 1, 1);
    mdd_free(resultMdd);
    resultMdd = tmp;
  }

  ringMdd = mdd_dup(resultMdd);

  if (onionRings) {
    array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
  }
  
  /* if until is trivially satisfied, or
     the early termination condition holds*/ 
  {
    bdd_t *fairInvariantStates = mdd_and(invariantMdd, fairStates, 1, 1);
    boolean trivial;

    trivial = mdd_lequal_mod_care_set_array(fairInvariantStates, resultMdd,
					    1, 1, careStatesArray);
    /* the lequal also takes care of the case where result = 1 */
    if(trivial ||
       McCheckEarlyTerminationForUnderapprox(earlyTermination,
					     resultMdd,
					     careStatesArray)){
      bdd_free(fairInvariantStates);
      bdd_free(ringMdd);

      /* trivially satisfied means that the fixpoint is complete.  If this
	 information is requested, give it. */
      if(fixpoint != NIL(boolean))
	*fixpoint = trivial;
      return(resultMdd);
    }
    bdd_free(fairInvariantStates);
  }

  /* The LFP loop */
  while (!term_fixpoint && !term_tautology && !term_early) {
    /* If dclevel is maximal, ringbdd is between the frontier and the
       reached set */
    aMdd = Mc_FsmEvaluateEXFormula(modelFsm, ringMdd, fairStates,
				   careStatesArray, verbosity, dcLevel);
    mdd_free(ringMdd);

    bMdd = mdd_and(aMdd, invariantMdd, 1, 1);
    mdd_free(aMdd);

    cMdd = mdd_or(resultMdd, bMdd, 1, 1);
    mdd_free(bMdd);

    /* Can we stop?  The tautology condition can be weakened to
       cMdd <= careStatesArray */
    term_fixpoint = mdd_equal_mod_care_set_array(cMdd, resultMdd,
						 careStatesArray);
    if(!term_fixpoint)
      term_tautology = mdd_is_tautology(cMdd, 1);
    if(!term_fixpoint && !term_tautology)
      term_early = McCheckEarlyTerminationForUnderapprox(earlyTermination, 
							 cMdd,
							 careStatesArray);

    /* Print some debug info, and set ringmdd */
    if (dcLevel >= McDcLevelRchFrontier_c) {
      mdd_t *tmpMdd = mdd_and(resultMdd, cMdd, 0, 1);
      ringMdd = bdd_between(tmpMdd, cMdd);
      if (verbosity == McVerbosityMax_c) {
	fprintf(vis_stdout,
		"--EU |A(n+1)-A(n)|= %d\t|A(n+1)| = %d\t|between-dc| = %d\n",
		mdd_size(tmpMdd), mdd_size(resultMdd), mdd_size(ringMdd));
      }
      mdd_free(tmpMdd);
    }
    else {
      ringMdd = mdd_dup(cMdd);
      if (verbosity == McVerbosityMax_c) {
	  fprintf(vis_stdout, "-- EU |ringMdd| = %d\n", mdd_size(ringMdd));
      }
    }

    mdd_free(resultMdd);
    resultMdd = cMdd;
    /* add onion ring unless fixpoint reached */
    if (!term_fixpoint && onionRings != NIL(array_t)) {
      array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
    }
  } /* while(!termination) for LFP */

  mdd_free(ringMdd);

  /* Print some debug info */
  if (verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    static int refCount=0;
    mdd_manager *mddMgr = Fsm_FsmReadMddManager(modelFsm);
    array_t *psVars = Fsm_FsmReadPresentStateVars(modelFsm);

    if (verbosity == McVerbosityMax_c) {
      mdd_t *tmpMdd = careStatesArray ?
		      mdd_and_array(resultMdd, careStatesArray, 1, 1) :
		      mdd_dup(resultMdd);
      fprintf(vis_stdout, "--EU called %d (bdd_size - %d)\n", ++refCount,
	      mdd_size(resultMdd));
      fprintf(vis_stdout,
	      "--There are %.0f care states satisfying EU formula\n",
	      mdd_count_onset(mddMgr, tmpMdd, psVars));
#ifdef DEBUG_MC
      /* The following 2 lines are just for debug */
      fprintf(vis_stdout, "EU satisfying minterms :\n");
      (void)_McPrintSatisfyingMinterms(tmpMdd, modelFsm);
#endif
      mdd_free(tmpMdd);
    } else {
      fprintf(vis_stdout,
	      "--There are %.0f states satisfying EU formula\n",
	      mdd_count_onset(mddMgr, resultMdd, psVars));
    }

    fprintf(vis_stdout, "--EU: %d preimage computations\n",
      Img_GetNumberOfImageComputation(Img_Backward_c) - nPreComps);
  }

  if(fixpoint != NIL(boolean))
    *fixpoint = term_fixpoint || term_tautology;
  return resultMdd;

} /* McEvaluateEUFormulaWithGivenTR */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying E(invariant S target), given
  a transition relation]

  Description [Does the same as Mc_FsmEvaluateESFormula, minus the hints.]

  SeeAlso [Mc_FsmEvaluateESFormula]
  
  SideEffects []

******************************************************************************/
mdd_t *
McEvaluateESFormulaWithGivenTR(
  Fsm_Fsm_t *modelFsm,
  mdd_t *invariantMdd,
  mdd_t *sourceMdd,
  mdd_t *underapprox,
  mdd_t *fairStates,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  array_t *onionRings,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint
  )
{
  mdd_t *aMdd;
  mdd_t *bMdd;
  mdd_t *cMdd;
  mdd_t *resultMdd;
  mdd_t *ringMdd;
  int nImgComps;
  boolean term_tautology = FALSE; /* different termination conditions */
  boolean term_early     = FALSE; 
  boolean term_fixpoint  = FALSE; 

  nImgComps = Img_GetNumberOfImageComputation(Img_Forward_c);
  resultMdd = mdd_and(sourceMdd, fairStates, 1, 1);

  if(underapprox != NIL(mdd_t)){
    mdd_t *tmp = mdd_or(resultMdd, underapprox, 1, 1);
    mdd_free(resultMdd);
    resultMdd = tmp;
  }

  ringMdd = mdd_dup(resultMdd);

  if (onionRings) {
    array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
  }
  
  /* if Since is trivially satisfied, or
     the early termination condition holds*/ 
  {
    bdd_t *fairInvariantStates = mdd_and(invariantMdd, fairStates, 1, 1);
    boolean trivial;

    trivial = mdd_lequal_mod_care_set_array(fairInvariantStates, resultMdd,
					    1, 1, careStatesArray);
    /* the lequal also takes care of the case where result = 1 */
    if(trivial ||
       McCheckEarlyTerminationForUnderapprox(earlyTermination,
					     resultMdd,
					     careStatesArray)){
      bdd_free(fairInvariantStates);
      bdd_free(ringMdd);

      /* trivially satisfied means that the fixpoint is complete.  If this
	 information is requested, give it. */
      if(fixpoint != NIL(boolean))
	*fixpoint = trivial;
      return(resultMdd);
    }
    bdd_free(fairInvariantStates);
  }

  /* The LFP loop */
  while (!term_fixpoint && !term_tautology && !term_early) {
    /* If dclevel is maximal, ringbdd is between the frontier and the
       reached set */
    aMdd = Mc_FsmEvaluateEYFormula(modelFsm, ringMdd, fairStates,
				   careStatesArray, verbosity, dcLevel);
    mdd_free(ringMdd);

    bMdd = mdd_and(aMdd, invariantMdd, 1, 1);
    mdd_free(aMdd);

    cMdd = mdd_or(resultMdd, bMdd, 1, 1);
    mdd_free(bMdd);

    /* Can we stop?  The tautology condition can be weakened to
       cMdd <= careStatesArray */
    term_fixpoint = mdd_equal_mod_care_set_array(cMdd, resultMdd,
						 careStatesArray);
    if(!term_fixpoint)
      term_tautology = mdd_is_tautology(cMdd, 1);
    if(!term_fixpoint && !term_tautology)
      term_early = McCheckEarlyTerminationForUnderapprox(earlyTermination, 
							 cMdd,
							 careStatesArray);

    /* Print some debug info, and set ringmdd */
    if (dcLevel >= McDcLevelRchFrontier_c) {
      mdd_t *tmpMdd = mdd_and(resultMdd, cMdd, 0, 1);
      ringMdd = bdd_between(tmpMdd, cMdd);
      if (verbosity == McVerbosityMax_c) {
	fprintf(vis_stdout,
		"--ES |A(n+1)-A(n)|= %d\t|A(n+1)| = %d\t|between-dc| = %d\n",
		mdd_size(tmpMdd), mdd_size(resultMdd), mdd_size(ringMdd));
      }
      mdd_free(tmpMdd);
      }
    else {
      ringMdd = mdd_dup(cMdd);
      if (verbosity == McVerbosityMax_c) {
	  fprintf(vis_stdout, "-- ES |ringMdd| = %d\n", mdd_size(ringMdd));
      }
    }

    mdd_free(resultMdd);
    resultMdd = cMdd;
    /* add onion ring unless fixpoint reached */
    if (!term_fixpoint && onionRings != NIL(array_t)) {
      array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
    }
  } /* while(!termination) for LFP */

  mdd_free(ringMdd);

  /* Print some debug info */
  if (verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    static int refCount=0;
    mdd_manager *mddMgr = Fsm_FsmReadMddManager(modelFsm);
    array_t *psVars = Fsm_FsmReadPresentStateVars(modelFsm);

    if (verbosity == McVerbosityMax_c) {
      mdd_t *tmpMdd = careStatesArray ?
		      mdd_and_array(resultMdd, careStatesArray, 1, 1) :
		      mdd_dup(resultMdd);
      fprintf(vis_stdout, "--ES called %d (bdd_size - %d)\n", ++refCount,
	      mdd_size(resultMdd));
      fprintf(vis_stdout,
	      "--There are %.0f care states satisfying ES formula\n",
	      mdd_count_onset(mddMgr, tmpMdd, psVars));
#ifdef DEBUG_MC
      /* The following 2 lines are just for debug */
      fprintf(vis_stdout, "ES satisfying minterms :\n");
      (void)_McPrintSatisfyingMinterms(tmpMdd, modelFsm);
#endif
      mdd_free(tmpMdd);
    } else {
      fprintf(vis_stdout,
	      "--There are %.0f states satisfying ES formula\n",
	      mdd_count_onset(mddMgr, resultMdd, psVars));
    }

    fprintf(vis_stdout, "--ES: %d image computations\n",
      Img_GetNumberOfImageComputation(Img_Forward_c) - nImgComps);
  }

  if(fixpoint != NIL(boolean))
    *fixpoint = term_fixpoint || term_tautology;
  return resultMdd;

} /* McEvaluateESFormulaWithGivenTR */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EG(invariantMdd), given
  a transition relation]

  Description [Does the same as Mc_FsmEvaluateEGFormula, minus the hints.]

  SeeAlso [Mc_FsmEvaluateEGFormula]
  
  SideEffects []

******************************************************************************/
mdd_t *
McEvaluateEGFormulaWithGivenTR(
  Fsm_Fsm_t *modelFsm,
  mdd_t *invariantMdd,
  mdd_t *overapprox,
  mdd_t *fairStates,
  Fsm_Fairness_t *modelFairness,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  array_t **pOnionRingsArrayForDbg,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint)
{
  int i;
  array_t *onionRings;
  boolean useRings;
  mdd_manager *mddManager;
  mdd_t *mddOne;
  mdd_t *iterate;
  array_t *buchiFairness;
  int nPreComps;
  int nIterations;
  boolean term_fixpoint = FALSE;        /* different termination conditions */
  boolean term_tautology = FALSE;
  boolean term_early = FALSE;

  /* Here's how the onionRingsArrayForDbg works.  It is an array of
     arrays of mdds.  It is filled in anew for every pass of the
     algorithm, that is for every /\_{c inC} EXE(Y U Y*c).  There is
     one entry for every fairness constraint, and this entry contains
     the onionrings of the EU computation that corresponds to this
     constraint. */
  assert(pOnionRingsArrayForDbg == NIL(array_t *) ||
	 *pOnionRingsArrayForDbg == NIL(array_t) ||
	 array_n(*pOnionRingsArrayForDbg) == 0);  

  useRings = (pOnionRingsArrayForDbg != NIL(array_t *) &&
	      *pOnionRingsArrayForDbg != NIL(array_t));

  nIterations = 0;
  nPreComps  = Img_GetNumberOfImageComputation(Img_Backward_c);
  onionRings = NIL(array_t);
  mddManager = Fsm_FsmReadMddManager(modelFsm);
  mddOne  = mdd_one(mddManager);

  /* If an overapproximation of the greatest fixpoint is given, use it. */
  if(overapprox != NIL(mdd_t)){
    iterate = mdd_and(invariantMdd, overapprox, 1, 1);
  } else {
    iterate = mdd_dup(invariantMdd);
  }

  /* See if we need to enter the loop at all.  If we wanted to test for
   * early termination here, we should fix the onion rings.  In the current
   * case, we do not need to, since the EG does not hold, and hence a
   * counterexample does not exist.
   */
  if( mdd_is_tautology(iterate, 0)){
    mdd_free(mddOne);
    if(fixpoint != NIL(boolean)) *fixpoint = mdd_is_tautology(iterate, 0);
    return(iterate);
  }

  /* read fairness constraints */
  buchiFairness = array_alloc(mdd_t *, 0);
  if(modelFairness != NIL(Fsm_Fairness_t)) {
    if(!Fsm_FairnessTestIsBuchi(modelFairness)) {
      (void)fprintf(vis_stdout,
	       "** mc error: non-Buechi fairness constraints not supported\n");
      array_free(buchiFairness);
      mdd_free(iterate);
      mdd_free(mddOne);

      if(fixpoint != NIL(boolean)) *fixpoint = FALSE;
      return NIL(mdd_t);
    }
    else {
      int j;
      int numBuchi = Fsm_FairnessReadNumConjunctsOfDisjunct(modelFairness, 0);
      for (j = 0 ; j < numBuchi; j++) {
	mdd_t *tmpMdd = Fsm_FairnessObtainFinallyInfMdd(modelFairness, 0, j,
						careStatesArray, dcLevel);
	array_insert_last(mdd_t *, buchiFairness, tmpMdd);
      }
    }
  }
  else {
    array_insert_last(mdd_t *, buchiFairness, mdd_one(mddManager));
  }


  /* GFP.  If we know that the result is a subset of a set S, we can conjoin
   the iterate with that set.  We use this to converge faster.  */
  while (TRUE) {
    mdd_t *newIterate;           /* the new iterate */

    nIterations++;
    newIterate = mdd_dup(iterate);
    
    /* for all fairness constraints */
    for (i = 0 ; i < array_n(buchiFairness) ; i++) {
      mdd_t *aMdd, *bMdd, *cMdd, *dMdd;
      mdd_t *FiMdd = array_fetch(mdd_t *, buchiFairness, i);
      
      if(useRings) {
	onionRings = array_alloc(mdd_t *, 0);
	array_insert_last(array_t *, *pOnionRingsArrayForDbg, onionRings);
      }

      aMdd = mdd_and(FiMdd, newIterate, 1, 1);

      bMdd = Mc_FsmEvaluateEUFormula(modelFsm, newIterate, aMdd, NIL(mdd_t),
				     mddOne, careStatesArray,
				     MC_NO_EARLY_TERMINATION, NIL(array_t),
				     Mc_None_c, onionRings, verbosity, dcLevel,
				     NIL(boolean));  
      mdd_free(aMdd);
      cMdd = Mc_FsmEvaluateEXFormula(modelFsm, bMdd, mddOne, careStatesArray,
				     verbosity, dcLevel);
      mdd_free(bMdd);

      dMdd  = mdd_and(newIterate, cMdd, 1, 1);

      mdd_free(cMdd);
      mdd_free(newIterate);
      newIterate = dMdd;

      /* invariants that hold here: newiterate <= iterate <= invariant. */
    }/* for all fairness constraints */

    /* termination */
    term_tautology = mdd_is_tautology(newIterate, 0);
    if(!term_tautology)
      term_fixpoint =  mdd_equal_mod_care_set_array(newIterate, iterate,
						    careStatesArray);
    if(!term_tautology && !term_fixpoint)
      term_early = McCheckEarlyTerminationForOverapprox(earlyTermination,
							newIterate, 
							careStatesArray);
    if(term_tautology || term_fixpoint || term_early){
      mdd_free(iterate);
      iterate = newIterate;
      break; /* from the GFP loop */
    }

    /* update iterate */
    mdd_free(iterate);
    iterate = newIterate;

    if(useRings){
      mdd_array_array_free(*pOnionRingsArrayForDbg);
      *pOnionRingsArrayForDbg = array_alloc(array_t *, 0);
    }
  } /* while(true) for gfp */

  /* Check if onionrings are OK */
  if(verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    if(useRings){
      for (i = 0 ; i < array_n(*pOnionRingsArrayForDbg); i++) {
	int j;
	mdd_t *Fi = array_fetch(mdd_t *,  buchiFairness, i);
	array_t *onionRings = array_fetch(array_t *, *pOnionRingsArrayForDbg,
					  i); 
	for (j = 0 ; j < array_n(onionRings); j++) {
	  mdd_t *ring = array_fetch(mdd_t *, onionRings, j);
	  if(j == 0) {
	    if(!mdd_lequal_mod_care_set_array(ring, Fi, 1, 1,
					       careStatesArray)) {
	      fprintf(vis_stderr, "** mc error: Problem w/ dbg in EG - ");
	      fprintf(vis_stderr,
		      "inner most ring not in Fi (fairness constraint).\n");
	    }
	  }
	  if(!mdd_lequal_mod_care_set_array(ring, invariantMdd, 1, 1,
					     careStatesArray)) {
	    fprintf(vis_stderr, "** mc error: Problem w/ dbg in EG - ");
	    fprintf(vis_stderr, "An onion ring of last EU fails invariant\n");
	  }
	} /* for all onionrings in array i */ 
      } /* for all onionringarrays in onionringsarray */
    }/* if useRings */

    if(verbosity == McVerbosityMax_c) {
      mdd_t *tmpMdd = careStatesArray ?
	mdd_and_array(iterate, careStatesArray, 1, 1) : mdd_dup(iterate);
      fprintf(vis_stdout,
	      "--There are %.0f care states satisfying EG formula\n",
	      mdd_count_onset(Fsm_FsmReadMddManager(modelFsm), tmpMdd,
				Fsm_FsmReadPresentStateVars(modelFsm)));
#ifdef DEBUG_MC
      /* The following 2 lines are just for debug */
      fprintf(vis_stdout, "EG satisfying minterms :\n");
      (void)_McPrintSatisfyingMinterms(tmpMdd, modelFsm);
#endif
      mdd_free(tmpMdd);
    } else {
      fprintf(vis_stdout, "--There are %.0f states satisfying EG formula\n",
	      mdd_count_onset(Fsm_FsmReadMddManager(modelFsm), iterate,
				Fsm_FsmReadPresentStateVars(modelFsm)));
    }

    fprintf(vis_stdout, "--EG: %d iterations, %d preimage computations\n",
      nIterations, Img_GetNumberOfImageComputation(Img_Backward_c) - nPreComps);
  }

  mdd_array_free(buchiFairness);
  mdd_free(mddOne);

  if(fixpoint != NIL(boolean))
     *fixpoint = term_fixpoint || term_tautology;
  return iterate;

} /* McEvaluateEGFormulaWithGivenTR */


/**Function********************************************************************

  Synopsis    [Evaluate states satisfying EH(invariantMdd), given
  a transition relation]

  Description [Does the same as Mc_FsmEvaluateEHFormula, minus the hints.]

  SeeAlso     [Mc_FsmEvaluateEHFormula]
  
  SideEffects []

******************************************************************************/
mdd_t *
McEvaluateEHFormulaWithGivenTR(
  Fsm_Fsm_t *modelFsm,
  mdd_t *invariantMdd,
  mdd_t *overapprox,
  mdd_t *fairStates,
  Fsm_Fairness_t *modelFairness,
  array_t *careStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  array_t **pOnionRingsArrayForDbg,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  boolean *fixpoint)
{
  int i;
  array_t *onionRings;
  boolean useRings;
  mdd_manager *mddManager;
  mdd_t *mddOne;
  mdd_t *iterate;
  array_t *buchiFairness;
  int nImgComps;
  int nIterations;
  boolean term_fixpoint = FALSE;        /* different termination conditions */
  boolean term_tautology = FALSE;
  boolean term_early = FALSE;

  /* Here's how the onionRingsArrayForDbg works.  It is an array of
     arrays of mdds.  It is filled in anew for every pass of the
     algorithm, that is for every /\_{c inC} EYE(Y S Y*c).  There is
     one entry for every fairness constraint, and this entry contains
     the onionrings of the ES computation that corresponds to this
     constraint. */
  assert(pOnionRingsArrayForDbg == NIL(array_t *) ||
	 *pOnionRingsArrayForDbg == NIL(array_t) ||
	 array_n(*pOnionRingsArrayForDbg) == 0);  

  useRings = (pOnionRingsArrayForDbg != NIL(array_t *) &&
	      *pOnionRingsArrayForDbg != NIL(array_t));

  nIterations = 0;
  nImgComps  = Img_GetNumberOfImageComputation(Img_Forward_c);
  onionRings = NIL(array_t);
  mddManager = Fsm_FsmReadMddManager(modelFsm);
  mddOne  = mdd_one(mddManager);

  /* If an overapproxiamtion of the greatest fixpoint is given, use it. */
  if(overapprox != NIL(mdd_t)){
    iterate = mdd_and(invariantMdd, overapprox, 1, 1);
  } else {
    iterate = mdd_dup(invariantMdd);
  }

  /* See if we need to enter the loop at all */
  if( mdd_is_tautology(iterate, 0) ||
      McCheckEarlyTerminationForOverapprox(earlyTermination,
					   iterate,
					   careStatesArray)){
    mdd_free(mddOne);

    if(fixpoint != NIL(boolean)) *fixpoint = mdd_is_tautology(iterate, 0);
    return(iterate);
  }

  /* read fairness constraints */
  buchiFairness = array_alloc(mdd_t *, 0);
  if(modelFairness != NIL(Fsm_Fairness_t)) {
    if(!Fsm_FairnessTestIsBuchi(modelFairness)) {
      (void)fprintf(vis_stdout,
	       "** mc error: non-Buechi fairness constraints not supported\n");
      array_free(buchiFairness);
      mdd_free(iterate);
      mdd_free(mddOne);

      if(fixpoint != NIL(boolean)) *fixpoint = FALSE;
      return NIL(mdd_t);
    }
    else {
      int j;
      int numBuchi = Fsm_FairnessReadNumConjunctsOfDisjunct(modelFairness, 0);
      for (j = 0 ; j < numBuchi; j++) {
	mdd_t *tmpMdd = Fsm_FairnessObtainFinallyInfMdd(modelFairness, 0, j,
						careStatesArray, dcLevel);
	array_insert_last(mdd_t *, buchiFairness, tmpMdd);
      }
    }
  }
  else {
    array_insert_last(mdd_t *, buchiFairness, mdd_one(mddManager));
  }


  /* GFP.  If we know that the result is a subset of a set S, we can conjoin
   the iterate with that set.  We use this to converge faster.  */
  while (TRUE) {
    mdd_t *newIterate;           /* the new iterate */

    nIterations++;
    newIterate = mdd_dup(iterate);
    
    /* for all fairness constraints */
    for (i = 0 ; i < array_n(buchiFairness) ; i++) {
      mdd_t *aMdd, *bMdd, *cMdd, *dMdd;
      mdd_t *FiMdd = array_fetch(mdd_t *, buchiFairness, i);
      
      if(useRings) {
	onionRings = array_alloc(mdd_t *, 0);
	array_insert_last(array_t *, *pOnionRingsArrayForDbg, onionRings);
      }

      aMdd = mdd_and(FiMdd, newIterate, 1, 1);

      bMdd = Mc_FsmEvaluateESFormula(modelFsm, newIterate, aMdd, NIL(mdd_t),
				     mddOne, careStatesArray,
				     MC_NO_EARLY_TERMINATION, NIL(array_t),
				     Mc_None_c, onionRings, verbosity, dcLevel,
				     NIL(boolean));  
      mdd_free(aMdd);
      cMdd = Mc_FsmEvaluateEYFormula(modelFsm, bMdd, mddOne, careStatesArray,
				     verbosity, dcLevel);
      mdd_free(bMdd);

      dMdd  = mdd_and(newIterate, cMdd, 1, 1);

      mdd_free(cMdd);
      mdd_free(newIterate);
      newIterate = dMdd;

      /* invariants that hold here: newiterate <= iterate <= invariant. */
    }/* for all fairness constraints */

    /* termination */
    term_tautology = mdd_is_tautology(newIterate, 0);
    if(!term_tautology)
      term_fixpoint =  mdd_equal_mod_care_set_array(newIterate, iterate,
						    careStatesArray);
    if(!term_tautology && !term_fixpoint)
      term_early = McCheckEarlyTerminationForOverapprox(earlyTermination,
							newIterate, 
							careStatesArray);
    if(term_tautology || term_fixpoint || term_early){
      mdd_free(iterate);
      iterate = newIterate;
      break; /* from the GFP loop */
    }

    /* update iterate */
    mdd_free(iterate);
    iterate = newIterate;

    if(useRings){
      mdd_array_array_free(*pOnionRingsArrayForDbg);
      *pOnionRingsArrayForDbg = array_alloc(array_t *, 0);
    }
  } /* while(true) for gfp */

  /* Check if onionrings are OK */
  if(verbosity == McVerbositySome_c || verbosity == McVerbosityMax_c) {
    if(useRings){
      for (i = 0 ; i < array_n(*pOnionRingsArrayForDbg); i++) {
	int j;
	mdd_t *Fi = array_fetch(mdd_t *,  buchiFairness, i);
	array_t *onionRings = array_fetch(array_t *, *pOnionRingsArrayForDbg,
					  i); 
	for (j = 0 ; j < array_n(onionRings); j++) {
	  mdd_t *ring = array_fetch(mdd_t *, onionRings, j);
	  if(j == 0) {
	    if(!mdd_lequal_mod_care_set_array(ring, Fi, 1, 1,
					       careStatesArray)) {
	      fprintf(vis_stderr, "** mc error: Problem w/ dbg in EH - ");
	      fprintf(vis_stderr,
		      "inner most ring not in Fi (fairness constraint).\n");
	    }
	  }
	  if(!mdd_lequal_mod_care_set_array(ring, invariantMdd, 1, 1,
					     careStatesArray)) {
	    fprintf(vis_stderr, "** mc error: Problem w/ dbg in EH - ");
	    fprintf(vis_stderr, "An onion ring of last ES fails invariant\n");
	  }
	} /* for all onionrings in array i */ 
      } /* for all onionringarrays in onionringsarray */
    }/* if useRings */

    if(verbosity == McVerbosityMax_c) {
      mdd_t *tmpMdd = careStatesArray ?
	mdd_and_array(iterate, careStatesArray, 1, 1) : mdd_dup(iterate);
      fprintf(vis_stdout,
	      "--There are %.0f care states satisfying EH formula\n",
	      mdd_count_onset(Fsm_FsmReadMddManager(modelFsm), tmpMdd,
				Fsm_FsmReadPresentStateVars(modelFsm)));
#ifdef DEBUG_MC
      /* The following 2 lines are just for debug */
      fprintf(vis_stdout, "EH satisfying minterms :\n");
      (void)_McPrintSatisfyingMinterms(tmpMdd, modelFsm);
#endif
      mdd_free(tmpMdd);
    } else {
      fprintf(vis_stdout, "--There are %.0f states satisfying EH formula\n",
	      mdd_count_onset(Fsm_FsmReadMddManager(modelFsm), iterate,
				Fsm_FsmReadPresentStateVars(modelFsm)));
    }

    fprintf(vis_stdout, "--EH: %d iterations, %d image computations\n",
      nIterations, Img_GetNumberOfImageComputation(Img_Forward_c) - nImgComps);
  }

  mdd_array_free(buchiFairness);
  mdd_free(mddOne);

  if(fixpoint != NIL(boolean))
     *fixpoint = term_fixpoint || term_tautology;
  return iterate;

} /* McEvaluateEHFormulaWithGivenTR */


/**Function********************************************************************

  Synopsis    [Frees memory stored at formula corresponding to debug data.]

  Description [Frees memory stored at formula corresponding to debug data.  It
  is assumed that the debug data is valid.  It is an error to call this
  function on a formula which is not of the type EG or EU.]

  SideEffects []

******************************************************************************/
void
McFormulaFreeDebugData(
  Ctlp_Formula_t *formula)
{
  Ctlp_FormulaType  type     = Ctlp_FormulaReadType(formula);
  array_t          *dbgArray = (array_t *) Ctlp_FormulaReadDebugData(formula);

  if (type == Ctlp_EU_c || type == Ctlp_FwdU_c) {
    mdd_array_free(dbgArray);
  }
  else if (type == Ctlp_EG_c || type == Ctlp_FwdG_c || type == Ctlp_EH_c) 
    mdd_array_array_free(dbgArray);
  else {
    fail("Error - called on non EG/EU formula.\n");
  }
}


/**Function********************************************************************

  Synopsis    [Print satisfying minterms.]

  Description [Print satisfying minterms. Just made for debug.]

  SideEffects []

******************************************************************************/
void
_McPrintSatisfyingMinterms(mdd_t *aMdd, Fsm_Fsm_t *fsm)
{
  array_t *support = Fsm_FsmReadPresentStateVars(fsm);
  mdd_manager *mddMgr = Fsm_FsmReadMddManager(fsm);
  Ntk_Network_t *network = Fsm_FsmReadNetwork(fsm);
  int i, n;
  array_t *mintermArray;
  mdd_t *aMinterm;

  n = (int) mdd_count_onset(mddMgr, aMdd, support);
  mintermArray = mdd_pick_arbitrary_minterms(mddMgr, aMdd, support, n);

  for (i = 0; i < n; i++) {
    aMinterm = array_fetch(mdd_t *, mintermArray, i);
    Mc_MintermPrint(aMinterm, support, network);
    mdd_free(aMinterm);
  }

  array_free(mintermArray);
}

/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/
  
/**Function********************************************************************

  Synopsis    [Model check formula on given fsm under fairness]

  Description [This is the recursive part of Mc_FsmEvaluateFormula]

  Comment [See Mc_EvaluateFormula.  The only difference is that
  EvaluateFormula takes care of global hints, and passes on an extra
  flag, TRMinimization that indicates how the current TR related to
  the exact TR.  Using this flag, the procedure decides in which field
  of the formula to store results (states, underapproxStates or
  overapproxStates) ]

  SideEffects []

******************************************************************************/
static mdd_t *
EvaluateFormulaRecur(
  Fsm_Fsm_t *modelFsm,
  Ctlp_Formula_t *ctlFormula,
  mdd_t *fairStates,
  Fsm_Fairness_t *fairCondition,
  array_t *modelCareStatesArray,
  Mc_EarlyTermination_t *earlyTermination,
  Fsm_HintsArray_t *hintsArray,
  Mc_GuidedSearch_t hintType,
  Ctlp_Approx_t TRMinimization,
  Ctlp_Approx_t *resultApproxType,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel,
  int buildOnionRings,
  Mc_GSHScheduleType GSHschedule)
{
  mdd_t *leftMdd  = NIL(mdd_t);
  mdd_t *rightMdd = NIL(mdd_t);
  mdd_t *result   = NIL(mdd_t);
  
  mdd_manager *mddMgr = Fsm_FsmReadMddManager(modelFsm);
  array_t     *careStatesArray = NIL(array_t);
  mdd_t       *tmpMdd;
  mdd_t       *approx = NIL(mdd_t); /* prev computed approx of sat set */ 
  mdd_t *ctlFormulaStates = Ctlp_FormulaObtainStates( ctlFormula );
  Ctlp_Approx_t leftApproxType = Ctlp_Exact_c;
  Ctlp_Approx_t rightApproxType = Ctlp_Exact_c;
  Ctlp_Approx_t thisApproxType = Ctlp_Exact_c;
  boolean fixpoint = FALSE;

  if ( ctlFormulaStates ) {
    return ctlFormulaStates;
  }
  
  {
    Ctlp_Formula_t *leftChild = Ctlp_FormulaReadLeftChild(ctlFormula);

    if (leftChild) {
      Mc_EarlyTermination_t *nextEarlyTermination =
	McObtainUpdatedEarlyTerminationCondition(
	  earlyTermination, NIL(mdd_t), Ctlp_FormulaReadType(ctlFormula));

      leftMdd = EvaluateFormulaRecur(modelFsm, leftChild, fairStates,
				     fairCondition, modelCareStatesArray,
				     nextEarlyTermination,
				     hintsArray, hintType, TRMinimization,
				     &leftApproxType, verbosity, dcLevel,
				     buildOnionRings, GSHschedule);
      Mc_EarlyTerminationFree(nextEarlyTermination);
      if (!leftMdd) {
	return NIL(mdd_t);
      }
    }
  }

  {/* read right */
    Ctlp_Formula_t *rightChild = Ctlp_FormulaReadRightChild(ctlFormula);

    if (rightChild) {
      Mc_EarlyTermination_t *nextEarlyTermination =
	McObtainUpdatedEarlyTerminationCondition(
	  earlyTermination, leftMdd, Ctlp_FormulaReadType(ctlFormula));

      rightMdd = EvaluateFormulaRecur(modelFsm, rightChild, fairStates,
				      fairCondition, modelCareStatesArray,
				      nextEarlyTermination,
				      hintsArray, hintType, TRMinimization,
				      &rightApproxType, verbosity, dcLevel,
				      buildOnionRings, GSHschedule);
      Mc_EarlyTerminationFree(nextEarlyTermination);
      if (!rightMdd) {
	if (leftMdd)
	  mdd_free(leftMdd);
	return NIL(mdd_t);
      }
    }
  }/* read right */

  if (modelCareStatesArray != NIL(array_t)) {
    careStatesArray = modelCareStatesArray;
  } else {
    careStatesArray = array_alloc(mdd_t *, 0);
    array_insert_last(mdd_t *, careStatesArray, mdd_one(mddMgr));
  }
  switch (Ctlp_FormulaReadType(ctlFormula)) {
  case Ctlp_ID_c:
    result = ModelCheckAtomicFormula(modelFsm, ctlFormula);
    break;

  case Ctlp_TRUE_c: result = mdd_one(mddMgr); break;
  case Ctlp_FALSE_c: result = mdd_zero(mddMgr); break;

  case Ctlp_NOT_c: result = mdd_not(leftMdd); break;

  case Ctlp_AND_c: result = mdd_and(leftMdd, rightMdd, 1, 1); break;

  case Ctlp_EQ_c: result = mdd_xnor(leftMdd, rightMdd); break;

  case Ctlp_XOR_c: result = mdd_xor(leftMdd, rightMdd); break;

  case Ctlp_THEN_c: result = mdd_or(leftMdd, rightMdd, 0, 1); break;

  case Ctlp_OR_c: result = mdd_or(leftMdd, rightMdd, 1, 1); break;

  case Ctlp_EX_c:
    result = Mc_FsmEvaluateEXFormula(modelFsm, leftMdd, fairStates,
				     careStatesArray, verbosity, dcLevel);
    break;

  case Ctlp_EU_c: {
    array_t *onionRings = NIL(array_t); /* array of mdd_t */
    boolean newrings;

    /* An underapproximation, together with its explanation may be stored. */
    approx = Ctlp_FormulaObtainApproxStates( ctlFormula, Ctlp_Underapprox_c );
    onionRings = (array_t *) Ctlp_FormulaReadDebugData(ctlFormula);
    newrings = onionRings == NIL(array_t);

    if (buildOnionRings){
      assert((approx != NIL(mdd_t) && onionRings != NIL(array_t)) ||
	     (approx == NIL(mdd_t) && onionRings == NIL(array_t)));
      if(onionRings == NIL(array_t))
	onionRings = array_alloc(mdd_t *, 0);
    }

    result = Mc_FsmEvaluateEUFormula(modelFsm, leftMdd, rightMdd,
				     approx, fairStates,
				     careStatesArray,
				     earlyTermination, hintsArray,
				     hintType, onionRings,
				     verbosity, dcLevel, &fixpoint);

    if(buildOnionRings && newrings)
      Ctlp_FormulaSetDbgInfo(ctlFormula, (void *) onionRings,
			     McFormulaFreeDebugData);
    if(approx != NIL(mdd_t))
      mdd_free(approx);

    break;
  }
  case Ctlp_EG_c: {
    array_t  *arrayOfOnionRings = NIL(array_t); /* array of array of mdd_t* */

    if(buildOnionRings)
      arrayOfOnionRings = array_alloc(array_t *, 0);

    approx = Ctlp_FormulaObtainApproxStates( ctlFormula, Ctlp_Overapprox_c );

    result = Mc_FsmEvaluateEGFormula(modelFsm, leftMdd, approx, fairStates, 
				     fairCondition, careStatesArray,
				     earlyTermination, hintsArray, hintType, 
				     &arrayOfOnionRings, verbosity,
				     dcLevel, &fixpoint, GSHschedule);

    if(buildOnionRings)
      Ctlp_FormulaSetDbgInfo(ctlFormula, (void *)arrayOfOnionRings,
			     McFormulaFreeDebugData);

    if(approx != NIL(mdd_t)) mdd_free(approx);

    break;
  }
  case Ctlp_Cmp_c: {
    if (Ctlp_FormulaReadCompareValue(ctlFormula) == 0)
      result = bdd_is_tautology(leftMdd, 0) ?
      		mdd_one(mddMgr) : mdd_zero(mddMgr);
    else
      result = bdd_is_tautology(leftMdd, 0) ?
      		mdd_zero(mddMgr) : mdd_one(mddMgr);
    break;
  }
  case Ctlp_Init_c:
    result = Fsm_FsmComputeInitialStates(modelFsm);
    break;
  case Ctlp_FwdU_c:
    if (buildOnionRings) {
      array_t *onionRings = array_alloc(mdd_t *, 0);
      result = Mc_FsmEvaluateFwdUFormula(modelFsm, leftMdd, rightMdd,
					 fairStates, careStatesArray,
					 onionRings, verbosity, dcLevel);
      Ctlp_FormulaSetDbgInfo(ctlFormula, (void *) onionRings,
			     McFormulaFreeDebugData);
    }
    else {
      if (Ctlp_FormulaReadLeftChild(ctlFormula) &&
	  Ctlp_FormulaReadType(Ctlp_FormulaReadLeftChild(ctlFormula)) ==
	  Ctlp_Init_c &&
	  Ctlp_FormulaReadRightChild(ctlFormula) &&
	  Ctlp_FormulaReadType(Ctlp_FormulaReadRightChild(ctlFormula)) ==
	  Ctlp_TRUE_c &&
	  Fsm_FsmTestIsReachabilityDone(modelFsm)) {
	result = mdd_dup(Fsm_FsmReadReachableStates(modelFsm));
	break;
      }
      result = Mc_FsmEvaluateFwdUFormula(modelFsm, leftMdd, rightMdd,
					 fairStates, careStatesArray,
					 NIL(array_t), verbosity, 
					 dcLevel);
    }
    break;
  case Ctlp_FwdG_c:
    if (buildOnionRings) {
      array_t *arrayOfOnionRings = array_alloc(array_t *, 0);
      result = Mc_FsmEvaluateFwdGFormula(modelFsm, leftMdd, rightMdd,
					 fairStates, fairCondition,
					 careStatesArray,
					 arrayOfOnionRings, verbosity,
					 dcLevel);   
      Ctlp_FormulaSetDbgInfo(ctlFormula, (void *)arrayOfOnionRings,
			     McFormulaFreeDebugData);
    } else {
      result = Mc_FsmEvaluateFwdGFormula(modelFsm, leftMdd, rightMdd,
					 fairStates, fairCondition,
					 careStatesArray,
					 NIL(array_t), verbosity,
					 dcLevel);
    }
    break;
  case Ctlp_EY_c:
    result = Mc_FsmEvaluateEYFormula(modelFsm, leftMdd, fairStates,
				     careStatesArray, verbosity, dcLevel);
    break;

  default: fail("Encountered unknown type of CTL formula\n");
  }

  tmpMdd = mdd_dup(result);
  thisApproxType = ComputeResultingAproximation(
    Ctlp_FormulaReadType(ctlFormula), leftApproxType,
    rightApproxType, TRMinimization, earlyTermination, fixpoint);
  Ctlp_FormulaSetApproxStates(ctlFormula, tmpMdd, thisApproxType);

#ifdef DEBUG_MC
  /* The following code is just for debugging */
  if ((verbosity == McVerbosityMax_c)) {
    char *type = Ctlp_FormulaGetTypeString(ctlFormula);
    if (Ctlp_FormulaReadType(ctlFormula) == Ctlp_Cmp_c) {
      fprintf(vis_stdout, "Info : result for [Cmp]\n");
      if (bdd_is_tautology(result, 1))
	fprintf(vis_stdout, "TRUE\n");
      else
	fprintf(vis_stdout, "FALSE\n");
    }
    else if (Ctlp_FormulaReadType(ctlFormula) == Ctlp_ID_c) {
      char *atomic = Ctlp_FormulaConvertToString(ctlFormula);
      fprintf(vis_stdout, "Info : satisfying minterms for [%s(%s)]:\n",
	      type, atomic);
      free(atomic);
      if (bdd_is_tautology(result, 1))
	fprintf(vis_stdout, "-- TAUTOLOGY --\n");
      else if (bdd_is_tautology(result, 0))
	fprintf(vis_stdout, "-- null --\n");
      else
	(void)_McPrintSatisfyingMinterms(result, modelFsm);
    }
    else {
      fprintf(vis_stdout, "Info : satisfying minterms for [%s]:\n", type);
      if (bdd_is_tautology(result, 1))
	fprintf(vis_stdout, "-- TAUTOLOGY --\n");
      else if (bdd_is_tautology(result, 0))
	fprintf(vis_stdout, "-- null --\n");
      else
	(void)_McPrintSatisfyingMinterms(result, modelFsm);
    }
    free(type);
  }
#endif
  
  if (modelCareStatesArray == NIL(array_t))
    mdd_array_free(careStatesArray);
  if (leftMdd)
    mdd_free(leftMdd);
  if (rightMdd)
    mdd_free(rightMdd);
  *resultApproxType = thisApproxType;
  return result;
}


/**Function********************************************************************

  Synopsis [Check CTL formulas given in file are modeled by flattened network]

  CommandName [model_check]

  CommandSynopsis [perform fair CTL model checking on a flattened network]

  CommandArguments [ \[-b\] \[-c\] \[-d &lt;dbg_level&gt;\]
  \[-f &lt;dbg_file&gt;\] \[-g &lt;hints_file&gt;\] \[-h\] \[-i\] \[-m\] \[-r\]
  \[-t &lt;time_out_period&gt;\]\[-v &lt;verbosity_level&gt;\]
  \[-D &lt;dc_level&gt;\] \[-F\] \[-S &lt;schedule&gt;\] &lt;ctl_file&gt;]

  CommandDescription [Performs fair CTL model checking on a flattened
  network.  Before calling this command, the user should have
  initialized the design by calling the command <A
  HREF="init_verifyCmd.html"> <code> init_verify</code></A>.
  Regardless of the options, no 'false positives' or 'false negatives'
  will occur: the result is correct for the given circuit.  <p>

  Properties to be verified should be provided as CTL formulas in the
  file <code>ctl_file</code>.  Note that the support of any wire
  referred to in a formula should consist only of latches.  For the
  precise syntax of CTL formulas, see the <A
  HREF="../ctl_ltl/ctl_ltl/ctl_ltl.html"> VIS CTL and LTL syntax manual</A>.  <p>

  Properties of the form <code> AG f</code>, where  <code>f</code> is a formula not
  involving path quantifiers are referred to as invariants; for such properties
  it may be substantially faster to use the <A HREF="check_invariantCmd.html">
  <code> check_invariant</code></A> command.
  <p>

  A fairness constraint can be specified by invoking the
  <A HREF="read_fairnessCmd.html"><code>read_fairness</code></A> command;
  if none is specified, all paths are taken to be fair.
  If some initial states
  do not lie on a fair path, the model checker prints a message to this effect.
  <p>

  A formula passes iff it is true for all initial states of the
  system.  Therefore, in the presence of multiple initial states, if a
  formula fails, the negation of the formula may also fail.<p>

  If a formula does not pass, a (potentially partial) proof of failure
  (referred to as a debug trace) is demonstrated. Fair paths are
  represented by a finite sequence of states (the stem) leading to a
  fair cycle, i.e. a cycle on which there is a state from each
  fairness condition. The level of detail of the proof can be
  specified (see option <code>-d</code>).  <p>

  Both backward(future tense CTL formulas) and forward(past tense CTL
  formulas) model checking can be performed. Forward model checking is
  based on Iwashita's ICCAD96 paper. Future tense CTL formulas are
  automatically converted to past tense ones as much as possible in
  forward model checking.  <p>

  Command options:
  <p>

  <dl>

  <dt> -b 
  <dd> Use backward analysis when performing debugging; the default is
  to use forward analysis. This should be tried when the debugger spends a
  large amount of time when creating a path to a fair cycle. This option is not
  compatible with forward model checking option(-F).
	   
  <p><dt> -c
  <dd> Use the formula tree so that there is no sharing of sub-formulae among
  the formulae in the input file. This option is useful in the following
  scenario - formulae A, B and C are being checked in order and there is
  sub-formula sharing between A and C. If the BDDs corresponding to the shared
  sub-formula is huge then computation for B might not be able to finish
  without using this option.
  </dd><p>

  <dt> -d <code> &lt;dbg_level&gt; </code> <dd> Specify the amount of
  debugging performed when the system fails a formula being checked.
  Note that it may not always be possible to give a simple
  counter-example to show that a formula is false, since this may
  require enumerating all paths from a state.  In such a case the
  model checker will print a message to this effect.  This option is
  incompatible with -F.  <p> <dd> <code> dbg_level</code> must be one
  of the following:

  <p><code>0</code>: No debugging performed.
  <code>dbg_level</code>=<code>0</code> is the default.

  <p><code>1</code>: Debugging with minimal output: generate counter-examples
  for universal formulas (formulas of the form <code>AX|AF|AU|AG</code>) and
  witnesses for existential formulas (formulas of the form
  <code>EX|EF|EU|EG</code>).  States on a path are not further analyzed.
  
  <p><code>2</code>: Same as  <code>dbg_level</code>=<code>1</code>, but more
  verbose. (The subformulas are printed, too.)

  <p><code>3</code>: Maximal automatic debugging: as for level <code>1</code>, except that states occurring on paths will be recursively analyzed.

  <p><code>4</code>: Manual debugging: at each state, the user is queried if more feedback is desired.

  <p>

  <dt> -f &lt;<code>dbg_file</code>&gt;
  <dd> Write the debugger output to <code>dbg_file</code>.
  This option is incompatible with -F.
  Notes: when you use -d4(interactive mode), -f is not recommended, since you
  can't see output of vis from the stdout.

  <dt> -g &lt;<code>hints_file</code>&gt; <dd> Use guided search.  The file
  <code>hints_file</code> contains a series of hints.  A hint is a formula that
  does not contain any temporal operators, so <code>hints_file</code> has the
  same syntax as a file of invariants used for check_invariant.  The hints are
  used in the order given to change the transition relation.  In the case of
  least fixpoints (EF, EU), the transition relation is conjoined with the hint,
  whereas with greatest fixpoints the transition relation is disjoined with the
  negation of the hint.  If the hints are cleverly chosen, this may speed up
  the computation considerably, because a search with the changed transition
  relation may be much simpler than one with the original transition relation,
  and results obtained can be reused, so that we may never have to do a
  complicated search with the original relation.  Note: hints in terms of
  primary inputs are not useful for greatest fixpoints.  See also: Ravi and
  Somenzi, Hints to accelerate symbolic traversal. CHARME'99; Bloem, Ravi, and
  Somenzi, Efficient Decision Procedures for Model Checking of Linear Time
  Logic Properties, CAV'99; Bloem, Ravi, and Somenzi, Symbolic Guided Search
  for CTL Model Checking, DAC'00.

  <p>For formulae that contain both least and greatest fixpoints, the
  behavior depends on the flag <code>guided_search_hint_type</code>.
  If it is set to local (default) then every subformula is evaluated
  to completion, using all hints in order, before the next subformula
  is started.  For pure ACTL or pure ECTL formulae, we can also set
  guided_search_hint_type to global, in which case the entire formula
  is evaluated for one hint before moving on to the next hint, using
  underapproximations.  The description of some options for guided
  search can be found in the help page for
  print_guided_search_options.

  <p>model_check will call reachability without any guided search, even
  if -g is used.  If you want to perform reachability with guided
  search, call rch directly.

  <p>Incompatible with -F.
  
  <dt> -h
  <dd> Print the command usage.
  <p>

  <dt> -i
  <dd> Print input values causing transitions between states during debugging.
  Both primary and pseudo inputs are printed.
  This option is incompatible with -F.
  <p>

  <dt> -m
  <dd> Pipe debugger output through the UNIX utility  more.
  This option is incompatible with -F.
  <p>

  <dt> -r
  <dd> Reduce the FSM derived from the flattened network with respect to each
  formula being checked. By default, the FSM is reduced with respect to the
  conjunction of the formulae in the input file. If this option is used and
  don't cares are being used for simplification, then subformula sharing is
  disabled (result might be incorrect otherwise).
  <p>

  <dd> The truth of  a formula may be independent  of  parts of the network (such as when wires have
  been abstracted; see <A HREF="flatten_hierarchyCmd.html"><code>flatten_hierarchy</code></A>).
  These parts are effectively removed when this option is invoked; this may result in
  more efficient model checking.
  <p>

  <dt> -t  <code>&lt;timeOutPeriod&gt;</code>
  <dd> Specify the time out period (in seconds) after which the command aborts. By default this option is set to infinity.
  <p>
  <dt> -v  <code>&lt;verbosity_level&gt;</code>
  <dd> Specify verbosity level. This sets the amount of feedback  on CPU usage and code status.

  <br><code>verbosity_level</code>  must be one of the following:<p>
  
  <code>0</code>: No feedback provided. This is the default.<p>

  <code>1</code>: Feedback on code location.<p>

  <code>2</code>: Feedback on code location and CPU usage.<p>

  <dt> -D <code> &lt;dc_level&gt; </code>

  <dd> Specify extent to which don't cares are used to simplify MDDs in model
  checking.  Don't cares are minterms on which the value taken by functions
  does not affect the computation; potentially, these minterms can be used to
  simplify MDDs and reduce the time taken to perform model checking.  The -g
  flag for guided search does not affect the way in which the don't-care
  conditions are computed.

  <br>
  <code> dc_level </code> must be one of the following:
  <p>

  <code> 0 </code>: No don't cares are used.

  <p> <code> 1 </code>: Use unreachable states as don't cares. This is the
  default.

  <p> <code> 2 </code>: Use unreachable states as don't cares and in the EU
  computation, use 'frontiers' for image computation.<p>

  <code> 3 </code>: First compute an overapproximation of the reachable states
  (ARDC), and use that as the cares set.  Use `frontiers' for image
  computation.  For help on controlling options for ARDC, look up help on the
  command: <A HREF="print_ardc_optionsCmd.html">print_ardc_options</A>. Refer
  to Moon, Jang, Somenzi, Pixley, Yuan, "Approximate Reachability Don't Cares
  for {CTL} Model Checking", ICCAD98, and to two papers by Cho et al, IEEE TCAD
  December 1996: one is for State Space Decomposition and the other is for
  Approximate FSM Traversal.<p>

  <dt> -F
  <dd> Use forward model checking based on Iwashita's method in ICCAD96.
       Future tense CTL formulas are automatically converted to past tense
       ones as much as possible. Converted forward formulas are printed when
       verbosity is greater than 0. Currently, debug options(-b, -d, -f, -i,
       and -m) are ignored with this option. We have seen that forward
       model checking was much faster than backward in many cases, also
       forward was much slower than backward in many cases.
  <p>

  <dt> -S <code> &lt;schedule&gt; </code>

  <dd> Specify schedule for GSH algorithm, which generalizes the
  Emerson-Lei algorithm and is used to compute greatest fixpoints.
  The choice of schedule affects the sequence in which EX and EU
  operators are applied.  It makes a difference only when fairness
  constraints are specified.

  <br>
  <code> &lt;schedule&gt; </code> must be one of the following:

  <p> <code> EL </code>: EU and EX operators strictly alternate.  This
  is the default.

  <p> <code> EL1 </code>: EX is applied once for every application of all EUs.

  <p> <code> EL2 </code>: EX is applied repeatedly after each application of
  all EUs.

  <p> <code> budget </code>: a hybrid of EL and EL2.

  <p> <code> random </code>: enabled operators are applied in
  (pseudo-)random order.

  <p> <code> off </code>: GSH is disabled, and the old algorithm is
  used instead.  The old algorithm uses the <code> EL </code>, but the
  termination checks are less sophisticated than in GSH.
  <p>

  <dt> <code> &lt;ctl_file&gt; </code>

  <dd> File containing CTL formulas to be model checked.
  </dl>

  Related "set" options:  
  <dl>
  <dt> ctl_change_bracket &lt;yes/no&gt;
  <dd> Vl2mv automatically converts "\[\]" to "&lt;&gt;" in node names,
  therefore CTL parser does the same thing. However, in some cases a user
  does not want to change node names in CTL parsing. Then, use this set
  option by giving "no". Default is "yes".
  <p>
  <dt>guided_search_hint_type</dt> <dd>Switches between local and
  global hints (see the -g option, or the help page for set).
  </dl>
  
  See also commands : approximate_model_check, incremental_ctl_verification 
  ]

  Description [First argument is a file containing a set of CTL formulas - see
  ctlp package for grammar. Second argument is an FSM that we will check the
  formulas on. Formulas are checked by calling the recursive function
  Mc_ModelCheckFormula. When the formula fails, the debugger is invoked.]

  Comment [Ctlp creates duplicate formulas when converting to
  existential form; e.g.  when converting AaUb. We aren't using this
  fact, which leads to some performance degradation.

  A system satisfies a formula if all its initial states are in the satosfying
  set of the formula.  Hence, we do not need to continue the computation if we
  know that all initial states are in the satisfying set, or if there are
  initial states that we are sure are not in the satisfying set.  This is what
  early termination does: it supplies an extra termination condition for the
  fixpoints that kicks in when we can decide the truth of the formula.  Not
  ethat this leads to some nasty consequences in storing the satisfying sets.
  A computation that has terminated early does not yield the exact satisfying
  set, and hence we can not always reuse this result when there is subformula
  sharing.]

  SideEffects []

  SeeAlso [CommandInv]

******************************************************************************/

static int
CommandMc(
  Hrc_Manager_t **hmgr,
  int argc,
  char **argv)
{
 /* options */
  McOptions_t         *options;
  Mc_VerbosityLevel   verbosity;
  Mc_DcLevel          dcLevel;
  FILE                *ctlFile;
  int                 timeOutPeriod     = 0;
  Mc_FwdBwdAnalysis   traversalDirection;
  int                 buildOnionRings   = 0;
  FILE                *guideFile;
  Mc_GuidedSearch_t   guidedSearchType  = Mc_None_c;
  Ctlp_FormulaArray_t *hintsArray       = NIL(Fsm_HintsArray_t);
  array_t             *hintsStatesArray = NIL(array_t); /* array of mdd_t* */
  boolean             noShare           = 0;
  Mc_GSHScheduleType  GSHschedule;
  
  /* CTL formulae */
  array_t *ctlArray;
  array_t *ctlNormalFormulaArray;
  int i;
  int numFormulae;

  /* FSM, network and image */
  Fsm_Fsm_t       *totalFsm = NIL(Fsm_Fsm_t);
  Fsm_Fsm_t       *modelFsm = NIL(Fsm_Fsm_t);
  Fsm_Fsm_t       *reducedFsm = NIL(Fsm_Fsm_t);
  Ntk_Network_t   *network;
  mdd_t           *modelCareStates = NIL(mdd_t);
  array_t         *modelCareStatesArray = NIL(array_t);
  mdd_t           *modelInitialStates;
  mdd_t           *fairStates;
  Fsm_Fairness_t  *fairCond;
  mdd_manager     *mddMgr;
  array_t         *bddIdArray;
  Img_ImageInfo_t *imageInfo;
  Mc_EarlyTermination_t *earlyTermination;

  /* Early termination is only partially implemented right now.  It needs
     distribution over all operators, including limited cases of temporal
     operators.  That should be relatively easy to implement. */
  
  /* time keeping */
  long totalInitialTime; /* for model checking */
  long initialTime, finalTime; /* for model checking */
  long iTime; /* starting time for reachability analysis */

  error_init();
  Img_ResetNumberOfImageComputation(Img_Both_c);

  /* read options */
  if (!(options = ParseMcOptions(argc, argv))) {
    return 1;
  }
  verbosity = McOptionsReadVerbosityLevel(options);
  dcLevel = McOptionsReadDcLevel(options);
  ctlFile = McOptionsReadCtlFile(options);
  timeOutPeriod = McOptionsReadTimeOutPeriod(options);
  traversalDirection = McOptionsReadTraversalDirection(options);
  buildOnionRings =
    (McOptionsReadDbgLevel(options) != McDbgLevelNone_c || verbosity);
  noShare = McOptionsReadUseFormulaTree(options);
  GSHschedule = McOptionsReadSchedule(options);

  guideFile =  McOptionsReadGuideFile(options);
 
  if(guideFile != NIL(FILE) ){
    guidedSearchType = Mc_ReadGuidedSearchType();
    if(guidedSearchType == Mc_None_c){  /* illegal setting */
      fprintf(vis_stderr, "** mc error: Unknown  hint type\n");
      fclose(guideFile);
      McOptionsFree(options);
      return 1;
    }
    
    if(traversalDirection == McFwd_c){  /* illegal combination */
      fprintf(vis_stderr, "** mc error: -g is incompatible with -F\n");
      fclose(guideFile);
      McOptionsFree(options);
      return 1;
    }

    if(Img_UserSpecifiedMethod() != Img_Iwls95_c &&
       Img_UserSpecifiedMethod() != Img_Monolithic_c &&
       Img_UserSpecifiedMethod() != Img_Mlp_c){
      fprintf(vis_stderr, "** mc error: -g only works with iwls95, MLP, or monolithic image methods.\n");
      fclose(guideFile);
      McOptionsFree(options);
      return 1;
    }

    hintsArray = Mc_ReadHints(guideFile);
    fclose(guideFile); guideFile = NIL(FILE);
    if( hintsArray == NIL(array_t) ){
      McOptionsFree(options);
      return 1;
    }
    
  } /* if guided search */

  /* If don't-cares are used, -r implies -c.  Note that the satisfying
     sets of a subformula are only in terms of propositions therein
     and their cone of influence.  Hence, we can share satisfying sets
     among formulae.  I don't quite understand what the problem with
     don't-cares is (RB) */
  if (McOptionsReadReduceFsm(options))
      if (dcLevel != McDcLevelNone_c)
	  McOptionsSetUseFormulaTree(options, TRUE);

  if (traversalDirection == McFwd_c &&
      McOptionsReadDbgLevel(options) != McDbgLevelNone_c) {
    McOptionsSetDbgLevel(options, McDbgLevelNone_c);
    (void)fprintf(vis_stderr, "** mc warning : option -d is ignored.\n");
  }

  /* Read CTL formulae */
  ctlArray = Ctlp_FileParseFormulaArray(ctlFile);
  fclose(ctlFile); ctlFile = NIL(FILE);
  if (ctlArray == NIL(array_t)) {
    (void) fprintf(vis_stderr,
		   "** mc error: error in parsing formulas from file\n");
    McOptionsFree(options);
    return 1;
  }
  
  /* read network */
  network = Ntk_HrcManagerReadCurrentNetwork(*hmgr);
  if (network == NIL(Ntk_Network_t)) {
    fprintf(vis_stdout, "%s\n", error_string());
    error_init();
    Ctlp_FormulaArrayFree(ctlArray);
    McOptionsFree(options);
    return 1;
  }

  /* read fsm */
  totalFsm = Fsm_NetworkReadOrCreateFsm(network);
  if (totalFsm == NIL(Fsm_Fsm_t)) {
    fprintf(vis_stdout, "%s\n", error_string());
    error_init();
    Ctlp_FormulaArrayFree(ctlArray);
    McOptionsFree(options);
    return 1;
  }

  if (verbosity > McVerbosityNone_c)
    totalInitialTime = util_cpu_time();
  else /* to remove uninitialized variable warning */
    totalInitialTime = 0;

  if(traversalDirection == McFwd_c){
    mdd_t *totalInitialStates; 
    double nInitialStates;
    
    totalInitialStates = Fsm_FsmComputeInitialStates(totalFsm);
    nInitialStates = mdd_count_onset(Fsm_FsmReadMddManager(totalFsm),
				     totalInitialStates,
				     Fsm_FsmReadPresentStateVars(totalFsm));
    mdd_free(totalInitialStates);
    
    /* If the number of initial states is only one, we can use both
     * conversion formulas(init ^ f != FALSE and init ^ !f == FALSE),
     * however, if we have multiple initial states, we should use
     * p0 ^ !f == FALSE.
     */
    ctlNormalFormulaArray =
      Ctlp_FormulaArrayConvertToForward(ctlArray, (nInitialStates == 1.0),
					noShare);
    /* end conversion for forward traversal */
  } else if (noShare) { /* conversion for backward, no sharing */
    ctlNormalFormulaArray =
      Ctlp_FormulaArrayConvertToExistentialFormTree(ctlArray);  
  }else{ /* conversion for backward, with sharing */
    /* Note that converting to DAG after converting to existential form would
       lead to more sharing, but it cannot be done since equal subformula that
       are converted from different formulae need different pointers back to
       their originals */
    array_t *temp = Ctlp_FormulaArrayConvertToDAG( ctlArray );
    array_free( ctlArray );
    ctlArray = temp;
    ctlNormalFormulaArray =
      Ctlp_FormulaDAGConvertToExistentialFormDAG(ctlArray);  
  }
  /* At this point, ctlNormalFormulaArray contains the formulas that are
     actually going to be checked, and ctlArray contains the formulas from
     which the conversion has been done.  Both need to be kept around until the
     end, for debugging purposes. */
  
  numFormulae = array_n(ctlNormalFormulaArray);
  
  /* time out */
  if (timeOutPeriod > 0) {
    (void) signal(SIGALRM, (void(*)(int))TimeOutHandle);
    (void) alarm(timeOutPeriod);
    if (setjmp(timeOutEnv) > 0) {
      (void) fprintf(vis_stdout,
		"# MC: timeout occurred after %d seconds.\n", timeOutPeriod);
      (void) fprintf(vis_stdout, "# MC: data may be corrupted.\n");
      if (verbosity > McVerbosityNone_c) {
	fprintf(vis_stdout, "-- total %d image computations and %d preimage computations\n",
		Img_GetNumberOfImageComputation(Img_Forward_c),
		Img_GetNumberOfImageComputation(Img_Backward_c));
      }
      alarm(0);
      return 1;
    }
  }

  /* Create reduced fsm, if necessary */
  if (!McOptionsReadReduceFsm(options)){
    /* We want to minimize only when "-r" option is not specified */
    /* reduceFsm would be NIL, if there is no reduction observed */
    assert (reducedFsm == NIL(Fsm_Fsm_t));
    reducedFsm = McConstructReducedFsm(network, ctlNormalFormulaArray);
    if (reducedFsm != NIL(Fsm_Fsm_t) && verbosity != McVerbosityNone_c) {
      mddMgr = Fsm_FsmReadMddManager(reducedFsm);
      bddIdArray = mdd_id_array_to_bdd_id_array(mddMgr,
		   Fsm_FsmReadPresentStateVars(reducedFsm));
      (void)fprintf(vis_stdout,"Local system includes ");
      (void)fprintf(vis_stdout,"%d (%d) multi-value (boolean) variables.\n",
		    array_n(Fsm_FsmReadPresentStateVars(reducedFsm)),
		    array_n(bddIdArray));
      array_free(bddIdArray);
    }
  }


  /************** for all formulae **********************************/
  for(i = 0; i < numFormulae; i++) {
    int nImgComps, nPreComps;
    mdd_t *ctlFormulaStates;
    Ctlp_Formula_t *ctlFormula = array_fetch(Ctlp_Formula_t *,
					     ctlNormalFormulaArray, i);
    modelFsm = NIL(Fsm_Fsm_t);
    
    /* do a check */
    if (!Mc_FormulaStaticSemanticCheckOnNetwork(ctlFormula, network, FALSE)) {
      (void) fprintf(vis_stdout,
		     "** mc error: error in parsing Atomic Formula:\n%s\n",
		     error_string());
      error_init();
      Ctlp_FormulaFree(ctlFormula);
      continue;
    }

    /* Create reduced fsm */
    if (McOptionsReadReduceFsm(options)) {
      /* We have not done top level reduction. */
      /* Using the same variable reducedFsm here   */
      array_t *oneFormulaArray = array_alloc(Ctlp_Formula_t *, 1);

      assert(reducedFsm == NIL(Fsm_Fsm_t));
      array_insert_last(Ctlp_Formula_t *, oneFormulaArray, ctlFormula);
      reducedFsm = McConstructReducedFsm(network, oneFormulaArray);
      array_free(oneFormulaArray);
      
      if (reducedFsm && verbosity != McVerbosityNone_c) {
	mddMgr = Fsm_FsmReadMddManager(reducedFsm);
	bddIdArray = mdd_id_array_to_bdd_id_array(mddMgr,
		       Fsm_FsmReadPresentStateVars(reducedFsm));
	(void)fprintf(vis_stdout,"Local system includes ");
	(void)fprintf(vis_stdout,"%d (%d) multi-value (boolean) variables.\n",
		      array_n(Fsm_FsmReadPresentStateVars(reducedFsm)),
		      array_n(bddIdArray));
	array_free(bddIdArray);
      }
    }/* if readreducefsm */

    /* Let us see if we got any reduction via top level or via "-r" */
    if (reducedFsm == NIL(Fsm_Fsm_t)) 
      modelFsm = totalFsm; /* no reduction */
    else
      modelFsm = reducedFsm; /* some reduction at some point */
    
    /* compute initial states */
    modelInitialStates = Fsm_FsmComputeInitialStates(modelFsm);
    if (modelInitialStates == NIL(mdd_t)) {
      int j;
      (void) fprintf(vis_stdout,
      "** mc error: Cannot build init states (mutual latch dependency?)\n%s\n",
		     error_string());
      if (modelFsm != totalFsm)
	Fsm_FsmFree(reducedFsm);

      alarm(0);
  
      for(j = i; j < numFormulae; j++)
	Ctlp_FormulaFree(
	  array_fetch( Ctlp_Formula_t *, ctlNormalFormulaArray, j ) );
      array_free( ctlNormalFormulaArray );
     
      Ctlp_FormulaArrayFree( ctlArray );
	  
      McOptionsFree(options);

      return 1;
    }

    earlyTermination = Mc_EarlyTerminationAlloc(McAll_c, modelInitialStates);

    if(hintsArray != NIL(Ctlp_FormulaArray_t)){
      hintsStatesArray = Mc_EvaluateHints(modelFsm, hintsArray);
      if( hintsStatesArray == NIL(array_t) ||
	  (guidedSearchType == Mc_Global_c &&
	   Ctlp_CheckClassOfExistentialFormula(ctlFormula) == Ctlp_Mixed_c)){
	int j;

	if( guidedSearchType == Mc_Global_c &&
	    Ctlp_CheckClassOfExistentialFormula(ctlFormula) == Ctlp_Mixed_c)
	  fprintf(vis_stderr, "** mc error: global hints incompatible with "
		  "mixed formulae\n");

	Mc_EarlyTerminationFree(earlyTermination);
	mdd_free(modelInitialStates);
	if (modelFsm != totalFsm)
	  Fsm_FsmFree(reducedFsm);
	alarm(0);
	for(j = i; j < numFormulae; j++)
	  Ctlp_FormulaFree(
	    array_fetch( Ctlp_Formula_t *, ctlNormalFormulaArray, j ) );
	array_free( ctlNormalFormulaArray );
	Ctlp_FormulaArrayFree( ctlArray );
	McOptionsFree(options);
	return 1;
      }/* problem with hints */
    }/* hints exist */
    
    /* stats */
    if (verbosity > McVerbosityNone_c) {
      initialTime = util_cpu_time();
      nImgComps = Img_GetNumberOfImageComputation(Img_Forward_c);
      nPreComps = Img_GetNumberOfImageComputation(Img_Backward_c);
    } else { /* to remove uninitialized variable warnings */
      initialTime = 0;
      nImgComps = 0;
      nPreComps = 0;
    }
    mddMgr = Fsm_FsmReadMddManager(modelFsm);

    /* compute don't cares. */
    if (modelCareStatesArray == NIL(array_t)) {
      if (verbosity > McVerbosityNone_c && i == 0)
	iTime = util_cpu_time();
      else /* to remove uninitialized variable warnings */
	iTime = 0;

      /* ardc */
      if (dcLevel == McDcLevelArdc_c) {
	Fsm_ArdcOptions_t *ardcOptions = McOptionsReadArdcOptions(options);
	
	modelCareStatesArray = Fsm_ArdcComputeOverApproximateReachableStates(
	  modelFsm, 0, verbosity, 0, 0, 0, 0, 0, 0, ardcOptions);
	if (verbosity > McVerbosityNone_c && i == 0)
	  Fsm_ArdcPrintReachabilityResults(modelFsm, util_cpu_time() - iTime);

      /* rch dc */
      } else if (dcLevel >= McDcLevelRch_c) {
	modelCareStates =
	  Fsm_FsmComputeReachableStates(modelFsm, 0, 1, 0, 0, 0, 0, 0,
					Fsm_Rch_Default_c, 0, 0,
					NIL(array_t), FALSE, NIL(array_t));  
	if (verbosity > McVerbosityNone_c && i == 0) {
	  Fsm_FsmReachabilityPrintResults(modelFsm, util_cpu_time() - iTime,
					  Fsm_Rch_Default_c);
	}

	modelCareStatesArray = array_alloc(mdd_t *, 0);
	array_insert(mdd_t *, modelCareStatesArray, 0, modelCareStates);
      } else {
	modelCareStates = mdd_one(mddMgr);
	modelCareStatesArray = array_alloc(mdd_t *, 0);
	array_insert(mdd_t *, modelCareStatesArray, 0, modelCareStates);
      }
    }
    
    Fsm_MinimizeTransitionRelationWithReachabilityInfo(
      modelFsm, (traversalDirection == McFwd_c) ? Img_Both_c : Img_Backward_c,
      verbosity > 1);
    
    /* fairness conditions */
    fairStates = Fsm_FsmComputeFairStates(modelFsm, modelCareStatesArray,
					  verbosity, dcLevel, GSHschedule,
					  McBwd_c, FALSE);
    fairCond = Fsm_FsmReadFairnessConstraint(modelFsm);

    if (mdd_lequal(modelInitialStates, fairStates, 1, 0)) {
      (void)fprintf(vis_stdout,
		    "** mc warning: There are no fair initial states\n");
    }
    else if (!mdd_lequal(modelInitialStates, fairStates, 1, 1)) {
      (void)fprintf(vis_stdout,
		    "** mc warning: Some initial states are not fair\n");
    }

    /* some user feedback */
    if (verbosity != McVerbosityNone_c) {
      (void)fprintf(vis_stdout, "Checking formula[%d] : ", i + 1);
      Ctlp_FormulaPrint(vis_stdout,
			Ctlp_FormulaReadOriginalFormula(ctlFormula));
      (void)fprintf (vis_stdout, "\n");
      if (traversalDirection == McFwd_c) {
	(void)fprintf(vis_stdout, "Forward formula : ");
	Ctlp_FormulaPrint(vis_stdout, ctlFormula);
	(void)fprintf(vis_stdout, "\n");
      }
    }

    /************** the actual computation **********************************/
    ctlFormulaStates =
      Mc_FsmEvaluateFormula(modelFsm, ctlFormula, fairStates,
			    fairCond, modelCareStatesArray,
			    earlyTermination, hintsStatesArray,
			    guidedSearchType, verbosity, dcLevel,
			    buildOnionRings, GSHschedule); 

    Mc_EarlyTerminationFree(earlyTermination);
    if(hintsStatesArray != NIL(array_t))
       mdd_array_free(hintsStatesArray);
    
    /* user feedback on succes/fail */
    PrintMcPassFail(mddMgr, modelFsm, traversalDirection, ctlFormula,
		    ctlFormulaStates, modelInitialStates, modelCareStatesArray,
		    options, verbosity);

    if (verbosity > McVerbosityNone_c) {
      finalTime = util_cpu_time();
      fprintf(vis_stdout, "-- mc time = %10g\n",
	(double)(finalTime - initialTime) / 1000.0);
      fprintf(vis_stdout,
	      "-- %d image computations and %d preimage computations\n",
	      Img_GetNumberOfImageComputation(Img_Forward_c) - nImgComps,
	      Img_GetNumberOfImageComputation(Img_Backward_c) - nPreComps);
    }

    Ctlp_FormulaFree(ctlFormula);
    mdd_free(modelInitialStates);
    mdd_free(fairStates);
    mdd_free(ctlFormulaStates);

    if ((McOptionsReadReduceFsm(options)) &&
	(reducedFsm != NIL(Fsm_Fsm_t))) {
      /*
      ** We need to free the reducedFsm only if it was created under "-r"
      ** option and was non-NIL.
      */
      Fsm_FsmFree(reducedFsm);
      reducedFsm = NIL(Fsm_Fsm_t);
      modelFsm = NIL(Fsm_Fsm_t); 
      if (modelCareStates) {
	mdd_array_free(modelCareStatesArray);
	modelCareStates = NIL(mdd_t);
	modelCareStatesArray = NIL(array_t);
      } else if (modelCareStatesArray) {
	modelCareStatesArray = NIL(array_t);
      }
    }
  }/* for all formulae */

  if (verbosity > McVerbosityNone_c) {
    finalTime = util_cpu_time();
    fprintf(vis_stdout, "-- total mc time = %10g\n",
      (double)(finalTime - totalInitialTime) / 1000.0);
    fprintf(vis_stdout,
	    "-- total %d image computations and %d preimage computations\n",
	    Img_GetNumberOfImageComputation(Img_Forward_c),
	    Img_GetNumberOfImageComputation(Img_Backward_c));
    /* Print tfm options if we have a global fsm. */
    if (!McOptionsReadReduceFsm(options) && modelFsm != NIL(Fsm_Fsm_t)) {
      imageInfo = Fsm_FsmReadImageInfo(modelFsm);
      if (Img_ImageInfoObtainMethodType(imageInfo) == Img_Tfm_c ||
	  Img_ImageInfoObtainMethodType(imageInfo) == Img_Hybrid_c) {
	Img_TfmPrintStatistics(imageInfo, Img_Both_c);
      }
    }
  }

  if (modelCareStates) {
    mdd_array_free(modelCareStatesArray);
  }

  if(hintsArray)
    Ctlp_FormulaArrayFree(hintsArray);
  
  if ((McOptionsReadReduceFsm(options) == FALSE) &&
      (reducedFsm != NIL(Fsm_Fsm_t))) {
    /* If "-r" was not specified and we did some reduction at top
       level, we need to free it */
    Fsm_FsmFree(reducedFsm);
    reducedFsm = NIL(Fsm_Fsm_t);
    modelFsm = NIL(Fsm_Fsm_t);
  }

  array_free(ctlNormalFormulaArray);
  (void) fprintf(vis_stdout, "\n");

  Ctlp_FormulaArrayFree(ctlArray);
  McOptionsFree(options);
  alarm(0);
  return 0;
}


/**Function********************************************************************

  Synopsis    [Find Mdd for states satisfying Atomic Formula.]

  Description [An atomic formula defines a set of states in the following
  way: it states a designated ``net'' (specified by the full path name)
  takes a certain value. The net should be purely a function of latches;
  as a result an evaluation of the net yields a set of states.]

  SideEffects []

******************************************************************************/

static mdd_t *
ModelCheckAtomicFormula(
  Fsm_Fsm_t *modelFsm,
  Ctlp_Formula_t *ctlFormula)
{
  mdd_t * resultMdd;
  mdd_t *tmpMdd;
  Ntk_Network_t *network = Fsm_FsmReadNetwork(modelFsm);
  char *nodeNameString = Ctlp_FormulaReadVariableName(ctlFormula);
  char *nodeValueString = Ctlp_FormulaReadValueName(ctlFormula);
  Ntk_Node_t *node = Ntk_NetworkFindNodeByName(network, nodeNameString);

  Var_Variable_t *nodeVar;
  int nodeValue;

  graph_t *modelPartition;
  vertex_t *partitionVertex;
  Mvf_Function_t *MVF;

  nodeVar = Ntk_NodeReadVariable(node);
  if (Var_VariableTestIsSymbolic(nodeVar)) {
    nodeValue = Var_VariableReadIndexFromSymbolicValue(nodeVar, nodeValueString);
  }
  else {
    nodeValue = atoi(nodeValueString);
  }

  modelPartition = Part_NetworkReadPartition(network);
  if (!(partitionVertex = Part_PartitionFindVertexByName(modelPartition,
							 nodeNameString))) {
    lsGen tmpGen;
    Ntk_Node_t *tmpNode;
    array_t *mvfArray;
    array_t *tmpRoots = array_alloc(Ntk_Node_t *, 0);
    st_table *tmpLeaves = st_init_table(st_ptrcmp, st_ptrhash);
    array_insert_last(Ntk_Node_t *, tmpRoots, node);

    Ntk_NetworkForEachCombInput(network, tmpGen, tmpNode) {
      st_insert(tmpLeaves, (char *) tmpNode, (char *) NTM_UNUSED);
    }

    mvfArray =  Ntm_NetworkBuildMvfs(network, tmpRoots, tmpLeaves,
				      NIL(mdd_t));
    MVF = array_fetch(Mvf_Function_t *, mvfArray, 0);
    array_free(tmpRoots);
    st_free_table(tmpLeaves);
    array_free(mvfArray);

    tmpMdd = Mvf_FunctionReadComponent(MVF, nodeValue);
    resultMdd = mdd_dup(tmpMdd);
    Mvf_FunctionFree(MVF);
  }
  else {
    MVF = Part_VertexReadFunction(partitionVertex);
    tmpMdd = Mvf_FunctionReadComponent(MVF, nodeValue);
    resultMdd = mdd_dup(tmpMdd);
  }
  if (Part_PartitionReadMethod(modelPartition) == Part_Frontier_c &&
      Ntk_NodeTestIsCombOutput(node)) {
    array_t	*psVars = Fsm_FsmReadPresentStateVars(modelFsm);
    mdd_manager	*mgr = Ntk_NetworkReadMddManager(Fsm_FsmReadNetwork(modelFsm));
    array_t	*supportList;
    st_table	*supportTable = st_init_table(st_numcmp, st_numhash);
    int		i, j;
    int		existIntermediateVars;
    int		mddId;
    Mvf_Function_t	*mvf;
    vertex_t	*vertex;
    array_t	*varBddRelationArray, *varArray;
    mdd_t	*iv, *ivMdd, *relation;

    for (i = 0; i < array_n(psVars); i++) {
      mddId = array_fetch(int, psVars, i);
      st_insert(supportTable, (char *)(long)mddId, NULL);
    }

    existIntermediateVars = 1;
    while (existIntermediateVars) {
      existIntermediateVars = 0;
      supportList = mdd_get_support(mgr, resultMdd);
      for (i = 0; i < array_n(supportList); i++) {
	mddId = array_fetch(int, supportList, i);
	if (!st_lookup(supportTable, (char *)(long)mddId, NULL)) {
	  vertex = Part_PartitionFindVertexByMddId(modelPartition, mddId);
	  mvf = Part_VertexReadFunction(vertex);
	  varBddRelationArray = mdd_fn_array_to_bdd_rel_array(mgr, mddId, mvf);
	  varArray = mdd_id_to_bdd_array(mgr, mddId);
	  assert(array_n(varBddRelationArray) == array_n(varArray));
	  for (j = 0; j < array_n(varBddRelationArray); j++) {
	    iv = array_fetch(mdd_t *, varArray, j);
	    relation = array_fetch(mdd_t *, varBddRelationArray, j);
	    ivMdd = bdd_cofactor(relation, iv);
	    mdd_free(relation);
	    tmpMdd = resultMdd;
	    resultMdd = bdd_compose(resultMdd, iv, ivMdd);
	    mdd_free(tmpMdd);
	    mdd_free(iv);
	    mdd_free(ivMdd);
	  }
	  array_free(varBddRelationArray);
	  array_free(varArray);
	  existIntermediateVars = 1;
	}
      }
      array_free(supportList);
    }
    st_free_table(supportTable);
  }
  return resultMdd;
}


/**Function********************************************************************

  Synopsis    [Parse command line options for mc.]

  SideEffects []

******************************************************************************/
static McOptions_t *
ParseMcOptions(
  int argc,
  char **argv)
{
  unsigned int i;
  int c;
  char *guideFileName = NIL(char);
  FILE *guideFile = NIL(FILE);
  boolean useMore = FALSE;
  boolean reduceFsm = FALSE;
  boolean printInputs = FALSE;
  boolean useFormulaTree = FALSE;
  FILE *inputFp=NIL(FILE);
  FILE *debugOut=NIL(FILE);
  char *debugOutName=NIL(char);
  Mc_DcLevel dcLevel = McDcLevelRch_c;
  McDbgLevel dbgLevel = McDbgLevelNone_c;
  Mc_VerbosityLevel verbosityLevel = McVerbosityNone_c;
  Mc_FwdBwdAnalysis fwdBwd = McFwd_c;
  McOptions_t *options = McOptionsAlloc();
  int timeOutPeriod = 0;
  Mc_FwdBwdAnalysis traversalDirection = McBwd_c;
  Fsm_ArdcOptions_t *ardcOptions;
  Mc_GSHScheduleType schedule = McGSH_EL_c;

  /*
   * Parse command line options.
   */

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "bcirmg:hv:d:D:f:t:FR:S:")) != EOF) {
    switch(c) {
    case 'g':
      guideFileName = util_strsav(util_optarg);
      break;
    case 'h':
      goto usage;
    case 'v':
      for (i = 0; i < strlen(util_optarg); i++) {
	if (!isdigit((int)util_optarg[i])) {
	  goto usage;
	}
      }
      verbosityLevel = (Mc_VerbosityLevel) atoi(util_optarg);
      break;
    case 'd':
      for (i = 0; i < strlen(util_optarg); i++) {
	if (!isdigit((int)util_optarg[i])) {
	  goto usage;
	}
      }
      dbgLevel = (McDbgLevel) atoi(util_optarg);
      break;
    case 'D':
      for (i = 0; i < strlen(util_optarg); i++) {
	if (!isdigit((int)util_optarg[i])) {
	  goto usage;
	}
      }
      dcLevel = (Mc_DcLevel) atoi(util_optarg);
      break;
    case 'b':
      fwdBwd = McBwd_c;
      break;
    case 'f':
      debugOutName = util_strsav(util_optarg);
      break;
    case 'm':
      useMore = TRUE;
      break;
    case 'r' :
      reduceFsm = TRUE;
      break;
    case 'c':
      useFormulaTree = TRUE;
      break;
    case 't' :
      timeOutPeriod  = atoi(util_optarg);
      break;
    case 'i' :
      printInputs = TRUE;
      break;
    case 'F' :
      traversalDirection = McFwd_c;
      break;
    case 'S' :
      schedule = McStringConvertToScheduleType(util_optarg);
      break;
    default:
      goto usage;
    }
  }
  
  /*
   * Open the ctl file.
   */
  if (argc - util_optind == 0) {
    (void) fprintf(vis_stderr, "** mc error: file containing ctl formulas not provided\n");
    goto usage;
  }
  else if (argc - util_optind > 1) {
    (void) fprintf(vis_stderr, "** mc error: too many arguments\n");
    goto usage;
  }

  McOptionsSetFwdBwd(options, fwdBwd);
  McOptionsSetUseMore(options, useMore);
  McOptionsSetReduceFsm(options, reduceFsm);
  McOptionsSetUseFormulaTree(options, useFormulaTree);
  McOptionsSetPrintInputs(options, printInputs);
  McOptionsSetTimeOutPeriod(options, timeOutPeriod);
  if (schedule == McGSH_Unassigned_c) {
    (void) fprintf(vis_stderr, "unknown schedule: %s\n", util_optarg);
    goto usage;
  } else {
    McOptionsSetSchedule(options, schedule);
  }
  inputFp = Cmd_FileOpen(argv[util_optind], "r", NIL(char *), 0);
  if (inputFp == NIL(FILE)) {
    McOptionsFree(options);
    if (guideFileName != NIL(char)) FREE(guideFileName);
    return NIL(McOptions_t);
  }
  McOptionsSetCtlFile(options, inputFp);

  if (debugOutName != NIL(char)) {
    debugOut = Cmd_FileOpen(debugOutName, "w", NIL(char *), 0);
    FREE(debugOutName);
    if (debugOut == NIL(FILE)) {
      McOptionsFree(options);
      if (guideFileName != NIL(char)) FREE(guideFileName);
      return NIL(McOptions_t);
    }
  }
  McOptionsSetDebugFile(options, debugOut);


  if (guideFileName != NIL(char)) {
    guideFile = Cmd_FileOpen(guideFileName, "r", NIL(char *), 0);
    if (guideFile == NIL(FILE)) {
      fprintf(vis_stderr, "** mc error: cannot open guided search file %s.\n",
	      guideFileName);
      FREE(guideFileName);
      guideFileName = NIL(char);
      McOptionsFree(options);
      return NIL(McOptions_t);
    }
    FREE(guideFileName);
    guideFileName = NIL(char);
  }
  McOptionsSetGuideFile(options, guideFile);

  if ((verbosityLevel != McVerbosityNone_c) &&
       (verbosityLevel != McVerbosityLittle_c) &&
       (verbosityLevel != McVerbositySome_c) &&
       (verbosityLevel != McVerbosityMax_c)) {
    goto usage;
  }
  else {
    McOptionsSetVerbosityLevel(options, verbosityLevel);
  }

  if ((dbgLevel != McDbgLevelNone_c) &&
       (dbgLevel != McDbgLevelMin_c) &&
       (dbgLevel != McDbgLevelMinVerbose_c) &&
       (dbgLevel != McDbgLevelMax_c) &&
       (dbgLevel != McDbgLevelInteractive_c)) {
    goto usage;
  }
  else {
    McOptionsSetDbgLevel(options, dbgLevel);
  }

  if ((dcLevel != McDcLevelNone_c) &&
       (dcLevel != McDcLevelRch_c ) &&
       (dcLevel != McDcLevelRchFrontier_c ) &&
       (dcLevel != McDcLevelArdc_c )) {
    goto usage;
  }
  else {
    McOptionsSetDcLevel(options, dcLevel);
  }

  McOptionsSetTraversalDirection(options, traversalDirection);
  if (dcLevel == McDcLevelArdc_c) {
    ardcOptions = Fsm_ArdcAllocOptionsStruct();
    Fsm_ArdcGetDefaultOptions(ardcOptions);
  } else
    ardcOptions = NIL(Fsm_ArdcOptions_t);
  McOptionsSetArdcOptions(options, ardcOptions);

  return options;

  usage:
  (void) fprintf(vis_stderr, "usage: model_check [-b][-c][-d dbg_level][-f dbg_file][-g <hintfile>][-h][-i][-m][-r][-t period][-v verbosity_level][-D dc_level][-F][-S <schedule>] <ctl_file>\n");
  (void) fprintf(vis_stderr, "    -b \tUse backward analysis in debugging\n");
  (void) fprintf(vis_stderr, "    -c \tNo sharing of CTL parse tree\n");
  (void) fprintf(vis_stderr, "    -d <dbg_level>\n");
  (void) fprintf(vis_stderr, "        debug_level = 0 => no debugging (Default)\n");
  (void) fprintf(vis_stderr, "        debug_level = 1 => automatic minimal debugging\n");
  (void) fprintf(vis_stderr, "        debug_level = 2 => automatic minimal debugging with extra verbosity\n");
  (void) fprintf(vis_stderr, "        debug_level = 3 => automatic maximal debugging\n");
  (void) fprintf(vis_stderr, "        debug_level = 4 => interactive debugging\n");
  (void) fprintf(vis_stderr, "    -f <dbg_out>\n");
  (void) fprintf(vis_stderr, "        write error trace to dbg_out\n");
  (void) fprintf(vis_stderr, "    -g <hint_file> \tSpecify file for guided search.\n");
  (void) fprintf(vis_stderr, "    -i \tPrint input values in debug traces.\n");
  (void) fprintf(vis_stderr, "    -m \tPipe debug output through more.\n");
  (void) fprintf(vis_stderr, "    -v <verbosity_level>\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 0 => no feedback (Default)\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 1 => code status\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 2 => code status and CPU usage profile\n");
  (void) fprintf(vis_stderr, "    -r  reduce FSM with respect to formula being verified\n");
  (void) fprintf(vis_stderr, "    -t <period> time out period\n");
  (void) fprintf(vis_stderr, "    -D <dc_level>\n");
  (void) fprintf(vis_stderr, "        dc_level = 0 => no use of don't cares\n");
  (void) fprintf(vis_stderr, "        dc_level = 1 => use unreached states as don't cares (Default)\n");
  (void) fprintf(vis_stderr, "        dc_level = 2 => use unreached states as don't cares\n");

  (void) fprintf(vis_stderr, "                        and aggressive use of DC's to simplify MDD's\n");
  (void) fprintf(vis_stderr, "        dc_level = 3 => use over-approximate unreached states as don't cares\n");
  (void) fprintf(vis_stderr, "                        and aggressive use of DC's to simplify MDD's\n");
  (void) fprintf(vis_stderr, "    -F \tUse forward model checking.\n");
  (void) fprintf(vis_stderr, "    -S <schedule>\n");
  (void) fprintf(vis_stderr, "       schedule is one of {EL,EL1,EL2,budget,random,off}\n");
  (void) fprintf(vis_stderr, "    <ctl_file> The input file containing CTL formula to be checked.\n");
  (void) fprintf(vis_stderr, "    -h \tPrints this usage message.\n");

  if (guideFileName != NIL(char)) FREE(guideFileName);
  McOptionsFree(options);

  return NIL(McOptions_t);
}


/**Function********************************************************************

  Synopsis [Check whether language of FSM is empty.]

  CommandName [lang_empty]

  CommandSynopsis [perform language emptiness check on a flattened network]

  CommandArguments [ \[-b\] \[-d &lt;dbg_level&gt;\] \[-f &lt;dbg_file&gt;\]
  \[-h\] \[-i\] \[-s\] \[-t &lt;time_out_period&gt;\]
  \[-v &lt;verbosity_level&gt;\] \[-D &lt;dc_level&gt;\]
  \[-S &lt;schedule&gt;\] \[-L &lt;lockstep_mode&gt;\] ]

  CommandDescription [
  Performs language emptiness check on a flattened network.  The
  language is not empty when there is a fair path starting at an
  initial state.  Before calling this command, the user should have
  initialized the design by calling the command <A
  HREF="init_verifyCmd.html"> <code> init_verify</code></A>.  <p>

  A fairness constraint can be read in by calling the
  <A HREF="read_fairnessCmd.html"><code>read_fairness</code></A> command;
  if none is specified, all paths are taken to be fair.
  <p>

  The system is reduced automatically with respect to the set of
  fairness constraints.  If the language is not empty, a proof of this
  fact is generated.  A proof is a fair path starting at an initial
  state.  This is represented by a finite sequence of states starting
  at an initial state (the stem) leading to a fair cycle, i.e., a
  cycle on which there lies a state from each fairness condition.
  <p>

  Command options:
  <p>

  <dl>

  <dt> -b
  <dd> Use backward analysis when performing debugging; the default is to use forward
	   analysis. This should be tried when the debugger spends a large amount of time when
	   creating a path to a fair cycle.
  <p>

  <dt> -d <code> &lt;dbg_level&gt; </code>
  <dd> Specify whether to demonstrate a proof of the language non-emptiness
  <p>
  <dd>
  <code> dbg_level</code> must be one of the following:
  <p>

  <code> 0 </code>: No debugging  performed. This is the default.
  <p>

  <code> 1 </code>: Generate a path to a fair cycle.
  <p>

  <dt> -f &lt;<code>dbg_file</code>&gt;
  <dd> Write the debugger output to <code>dbg_file</code>.
  <p>

  <dt> -h
  <dd> Print the command usage.
  <p>

  <dt> -m
  <dd> Pipe debugger output through the UNIX utility more.
  <p>

  <dt> -i
  <dd> Print input values causing transitions between states during debugging.
  Both primary and pseudo inputs are printed.
  <p>

  <dt> -s
  <dd> Print debug output in the format accepted by the
       <A HREF="simulateCmd.html"><code>simulate</code></A> command.
  <p>

  <dt> -t  <code>&lt;timeOutPeriod&gt;</code>
  <dd> Specify the time out period (in seconds) after which the command
  aborts.  By default this option is set to infinity.
  <p>

  <dt> -v  <code>&lt;verbosity_level&gt;</code>
  <dd> Specify verbosity level. This sets the amount of feedback on CPU
  usage and code status.
  <p>

  <dd>
  <code>verbosity_level</code> must be one of the following:
  <p>

  <code> 0 </code>: No feedback provided. This is the default.<p>

  <code> 1 </code>: Feedback on code location.<p>

  <code> 2 </code>: Feedback on code location and CPU usage.<p>

  <dt> -D <code> &lt;dc_level&gt; </code>
  <dd> Specify extent to which don't cares are used to simplify MDDs.
  Don't cares are minterms on which the value taken by functions does not affect the computation;
  potentially, these minterms can be used to simplify MDDs and reduce time taken to perform
  MDD computations.
  <p>

  <dd>
  <code> dc_level </code> must be one of the following:
  <p>

  <code> 0 </code>: No don't cares are used. <p>
  <code> 1 </code>: Use unreachable states as don't cares.  This is the
  default. <p>

  <dt> -S <code> &lt;schedule&gt; </code>

  <dd> Specify schedule for GSH algorithm, which generalizes the
  Emerson-Lei algorithm and is used to compute greatest fixpoints.
  The choice of schedule affects the sequence in which EX and EU
  operators are applied.  It makes a difference only when fairness
  constraints are specified.

  <br>
  <code> &lt;schedule&gt; </code> must be one of the following:

  <p> <code> EL </code>: EU and EX operators strictly alternate.  This
  is the default.

  <p> <code> EL1 </code>: EX is applied once for every application of all EUs.

  <p> <code> EL2 </code>: EX is applied repeatedly after each application of
  all EUs.

  <p> <code> budget </code>: a hybrid of EL and EL2.

  <p> <code> random </code>: enabled operators are applied in
  (pseudo-)random order.

  <p> <code> off </code>: GSH is disabled, and the old algorithm is
  used instead.  The old algorithm uses the <code> EL </code>, but the
  termination checks are less sophisticated than in GSH.
  <p>

  <dt> -F

  <dd> Use forward analysis in the computation of the greatest fixpoint.
  This option is incompatible with -d 1 or higher and can only be used
  with -D 1.

  <dt> -L <code> &lt;lockstep_mode&gt; </code>

  <dd> Use the lockstep algorithm, which is based on fair SCC enumeration.

  <br>
  <code> &lt;lockstep_mode&gt; </code> must be one of the following:

  <p> <code> off </code>: Lockstep is disabled.  This is the default.
  Language emptiness is checked by computing a hull of the fair SCCs.

  <p> <code> on </code>: Lockstep is enabled.

  <p> <code> all </code>: Lockstep is enabled; all fair SCCs are
  enumerated instead of terminating as soon as one is found.  This can
  be used to study the SCCs of a graph, but it is slower than the
  default option.

  <p> <code> n </code>: (n is a positive integer).  Lockstep is
  enabled and up to <code> n </code> fair SCCs are enumerated.  This
  is less expensive than <code> all </code>, but still less efficient
  than <code> on </code>, even when <code> n = 1 </code>.

  </dl>

  ]

  SideEffects []

  SeeAlso []

******************************************************************************/
static int
CommandLe(
  Hrc_Manager_t **hmgr,
  int argc,
  char **argv)
{
  McOptions_t *options;
  Mc_VerbosityLevel verbosity;
  Mc_DcLevel dcLevel;
  McDbgLevel dbgLevel;
  FILE *dbgFile= NIL(FILE);
  boolean printInputs = FALSE;
  static int timeOutPeriod = 0;
  Mc_GSHScheduleType GSHschedule;
  Mc_FwdBwdAnalysis GSHdirection;
  int lockstep;
  int useMore;
  int simValue;
  int reduceFsm_flag = 1;
  long startRchTime;
  mdd_t *modelCareStates;
  Fsm_Fsm_t *totalFsm;
  Fsm_Fsm_t *modelFsm, *reducedFsm = NIL(Fsm_Fsm_t);
  mdd_t *fairInitStates;
  array_t *modelCareStatesArray;
  long initialTime, finalTime; /* for lang_empty checking */

  Img_ResetNumberOfImageComputation(Img_Both_c);

  /* Read options and set timeout if requested. */
  if (!(options = ParseLeOptions(argc, argv))) {
    return 1;
  }

  totalFsm = Fsm_HrcManagerReadCurrentFsm(*hmgr);
  if (totalFsm == NIL(Fsm_Fsm_t)) {
    (void) fprintf(vis_stdout, "%s\n", error_string());
    error_init();
    McOptionsFree(options);
    return 1;
  }

  initialTime = util_cpu_time();
  
  verbosity     = McOptionsReadVerbosityLevel(options);
  dcLevel       = McOptionsReadDcLevel(options);
  dbgLevel      = McOptionsReadDbgLevel(options);
  printInputs   = McOptionsReadPrintInputs(options);
  timeOutPeriod = McOptionsReadTimeOutPeriod(options);
  GSHschedule   = McOptionsReadSchedule(options);
  GSHdirection  = McOptionsReadTraversalDirection(options);
  lockstep      = McOptionsReadLockstep(options);
  dbgFile       = McOptionsReadDebugFile(options);
  useMore       = McOptionsReadUseMore(options);
  simValue      = McOptionsReadSimValue(options);

  if (dbgLevel != McDbgLevelNone_c && GSHdirection == McFwd_c) {
    (void) fprintf(vis_stderr, "** le error: -d is incompatible with -F\n");
    McOptionsFree(options);
    return 1;
  }
  if (dcLevel !=  McDcLevelRch_c && GSHdirection == McFwd_c) {
    (void) fprintf(vis_stderr, "** le error: -F can only be used with -D1\n");
    McOptionsFree(options);
    return 1;
  }

  if (timeOutPeriod > 0) {
    (void) signal(SIGALRM, (void(*)(int))TimeOutHandle);
    (void) alarm(timeOutPeriod);
    if (setjmp(timeOutEnv) > 0) {
      (void) fprintf(vis_stdout,
		     "# LE: language emptiness - timeout occurred after %d seconds.\n",
		     timeOutPeriod);
      (void) fprintf(vis_stdout, "# LE: data may be corrupted.\n");
      if (verbosity > McVerbosityNone_c) {
	(void) fprintf(vis_stdout,
		       "-- total %d image computations and %d preimage computations\n",
		       Img_GetNumberOfImageComputation(Img_Forward_c),
		       Img_GetNumberOfImageComputation(Img_Backward_c));
      }
      alarm(0);
      return 1;
    }
  }

  /* Reduce FSM to cone of influence of fairness constraints. */
  if (reduceFsm_flag) {
    Ntk_Network_t *network = Fsm_FsmReadNetwork(totalFsm);
    array_t *ctlNormalFormulaArray = array_alloc(Ctlp_Formula_t *, 0);
    reducedFsm = McConstructReducedFsm(network, ctlNormalFormulaArray);
    array_free(ctlNormalFormulaArray);
  }
  
  if (reducedFsm == NIL(Fsm_Fsm_t)) {
    modelFsm = totalFsm;
  }else {
    modelFsm = reducedFsm;
  }

  /* Find care states and put them in an array */
  if (dcLevel == McDcLevelArdc_c) { /* aRDC */
    Fsm_ArdcOptions_t *ardcOptions = Fsm_ArdcAllocOptionsStruct();
    array_t *tmpArray;
    Fsm_ArdcGetDefaultOptions(ardcOptions);

    if (verbosity > 0)
      startRchTime = util_cpu_time();
    else /* to remove uninitialized variable warning */
      startRchTime = 0;
    tmpArray = Fsm_ArdcComputeOverApproximateReachableStates(
      modelFsm, 0, verbosity, 0, 0, 0, 0, 0, 0, ardcOptions);
    modelCareStatesArray = mdd_array_duplicate(tmpArray);

    FREE(ardcOptions);

    if (verbosity > 0 )
      Fsm_ArdcPrintReachabilityResults(modelFsm,
				       util_cpu_time() - startRchTime);

  }else if (dcLevel >= McDcLevelRch_c) { /*RDC*/
    if (verbosity > 0)
      startRchTime = util_cpu_time();
    else /* to remove uninitialized variable warning */
      startRchTime = 0;
    modelCareStates =
      Fsm_FsmComputeReachableStates(modelFsm, 0, verbosity, 0, 0,
				    (lockstep != MC_LOCKSTEP_OFF), 0, 0,
				    Fsm_Rch_Default_c, 0, 0,
				    NIL(array_t), FALSE, NIL(array_t));
    if (verbosity > 0) {
      Fsm_FsmReachabilityPrintResults(modelFsm,
				      util_cpu_time() - startRchTime,
				      Fsm_Rch_Default_c);
    }
    modelCareStatesArray = array_alloc(mdd_t *, 0);
    array_insert_last(mdd_t *, modelCareStatesArray, modelCareStates);

  } else {  /* mdd_one */
    modelCareStates = mdd_one(Fsm_FsmReadMddManager(modelFsm));
    modelCareStatesArray = array_alloc(mdd_t *, 0);
    array_insert_last(mdd_t *, modelCareStatesArray, modelCareStates);
  }

  fairInitStates = Mc_FsmCheckLanguageEmptiness(modelFsm,
						modelCareStatesArray,
						2,
						dcLevel,
						dbgLevel,
						printInputs,
						verbosity,
						GSHschedule,
						GSHdirection,
						lockstep,
						dbgFile,
						useMore,
						simValue,
						"LE");

  if (verbosity > McVerbosityNone_c) {
    finalTime = util_cpu_time();
    fprintf(vis_stdout, "-- total le time = %10g\n",
	    (double)(finalTime - initialTime) / 1000.0);
    fprintf(vis_stdout,
	    "-- total %d image computations and %d preimage computations\n",
	    Img_GetNumberOfImageComputation(Img_Forward_c),
	    Img_GetNumberOfImageComputation(Img_Backward_c));
  }

  /* Clean up. */
  if (fairInitStates) {
    mdd_free(fairInitStates);				
  }
  mdd_array_free(modelCareStatesArray);
  McOptionsFree(options);
  if (reducedFsm) { 
    Fsm_FsmFree(reducedFsm); 
  }

  alarm(0);
  return 0;

} /* CommandLe */


/**Function********************************************************************

  Synopsis    [Parse command line options for lang_empty.]

  SideEffects []

******************************************************************************/
static McOptions_t *
ParseLeOptions(
  int argc,
  char **argv)
{
  unsigned int i;
  int c;
  boolean useSimFormat = FALSE;
  FILE *debugOut=NIL(FILE);
  char *debugOutName=NIL(char);
  Mc_DcLevel dcLevel = McDcLevelRch_c;
  McDbgLevel dbgLevel = McDbgLevelNone_c;
  Mc_VerbosityLevel verbosityLevel = McVerbosityNone_c;
  Mc_FwdBwdAnalysis fwdBwd = McFwd_c;
  McOptions_t *options = McOptionsAlloc();
  boolean printInputs = FALSE;
  boolean useMore = FALSE;
  int timeOutPeriod = 0;
  Mc_GSHScheduleType schedule = McGSH_EL_c;
  Mc_FwdBwdAnalysis direction = McBwd_c;
  int lockstep = MC_LOCKSTEP_OFF;
  Fsm_ArdcOptions_t *ardcOptions;

  /*
   * Parse command line options.
   */

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "bihmt:v:d:D:f:sS:L:F")) != EOF) {
    switch(c) {
    case 'h':
      goto usage;
    case 'v':
      for (i = 0; i < strlen(util_optarg); i++) {
	if (!isdigit((int)util_optarg[i])) {
	  goto usage;
	}
      }
      verbosityLevel = (Mc_VerbosityLevel) atoi(util_optarg);
      break;
    case 'd':
      for (i = 0; i < strlen(util_optarg); i++) {
	if (!isdigit((int)util_optarg[i])) {
	  goto usage;
	}
      }
      dbgLevel = (McDbgLevel) atoi(util_optarg);
      break;
    case 'D':
      for (i = 0; i < strlen(util_optarg); i++) {
	if (!isdigit((int)util_optarg[i])) {
	  goto usage;
	}
      }
      dcLevel = (Mc_DcLevel) atoi(util_optarg);
      break;
    case 'f':
      debugOutName = util_strsav(util_optarg);
      break;
    case 's':
      useSimFormat = TRUE;
      break;
    case 't':
      timeOutPeriod = atoi(util_optarg);
      break;
    case 'i':
      printInputs = TRUE;
      break;
    case 'm':
      useMore = TRUE;
      break;
    case 'b':
      fwdBwd = McBwd_c;
      break;
    case 'S' :
      schedule = McStringConvertToScheduleType(util_optarg);
      break;
    case 'L' :
      lockstep = McStringConvertToLockstepMode(util_optarg);
      break;
    case 'F' :
      direction = McFwd_c;
      break;
    default:
      goto usage;
    }
  }

  if (schedule == McGSH_Unassigned_c) {
    (void) fprintf(vis_stderr, "unknown schedule: %s\n", util_optarg);
    goto usage;
  } else {
    McOptionsSetSchedule(options, schedule);
  }
  McOptionsSetTraversalDirection(options, direction);
  if (lockstep == MC_LOCKSTEP_UNASSIGNED) {
    (void) fprintf(vis_stderr, "invalid lockstep option: %s\n", util_optarg);
    goto usage;
  } else {
    McOptionsSetLockstep(options, lockstep);
  }
  if ((verbosityLevel != McVerbosityNone_c) &&
       (verbosityLevel != McVerbosityLittle_c) &&
       (verbosityLevel != McVerbositySome_c) &&
       (verbosityLevel != McVerbosityMax_c)) {
    goto usage;
  }
  else {
    McOptionsSetVerbosityLevel(options, verbosityLevel);
  }

  if (debugOutName != NIL(char)) {
    debugOut = Cmd_FileOpen(debugOutName, "w", NIL(char *), 0);
    if (debugOut == NIL(FILE)) {
      McOptionsFree(options);
      return NIL(McOptions_t);
    }
  }
  McOptionsSetDebugFile(options, debugOut);

  if ((dbgLevel != McDbgLevelNone_c) &&
       (dbgLevel != McDbgLevelMin_c)) {
    goto usage;
  }
  else {
    McOptionsSetDbgLevel(options, dbgLevel);
  }

  if ((dcLevel != McDcLevelNone_c)  &&
       (dcLevel != McDcLevelRch_c )  &&
       (dcLevel != McDcLevelRchFrontier_c )  &&
       (dcLevel != McDcLevelArdc_c )) {
    goto usage;
  }
  else {
    McOptionsSetDcLevel(options, dcLevel);
  }

  /* set Ardc options for le : C.W. */
  if (dcLevel == McDcLevelArdc_c) {
      ardcOptions = Fsm_ArdcAllocOptionsStruct();
      Fsm_ArdcGetDefaultOptions(ardcOptions);
  } else
      ardcOptions = NIL(Fsm_ArdcOptions_t);
  McOptionsSetArdcOptions(options, ardcOptions);

  McOptionsSetFwdBwd(options, fwdBwd);
  McOptionsSetSimValue(options, useSimFormat);
  McOptionsSetUseMore(options, useMore);
  McOptionsSetPrintInputs(options, printInputs);
  McOptionsSetTimeOutPeriod(options, timeOutPeriod);
  return options;

  usage:
  (void) fprintf(vis_stderr, "usage: lang_empty [-b][-d dbg_level][-f dbg_file][-h][-i][-m][-s][-t period][-v verbosity_level][-D <dc_level>][-S <schedule>][-F][-L <lockstep_mode>]\n");
  (void) fprintf(vis_stderr, "    -b \tUse backward analysis in debugging\n");
  (void) fprintf(vis_stderr, "    -d <dbg_level>\n");
  (void) fprintf(vis_stderr, "        debug_level = 0 => no debugging (Default)\n");
  (void) fprintf(vis_stderr, "        debug_level = 1 => automatic debugging\n");
  (void) fprintf(vis_stderr, "    -f <dbg_out>\n");
  (void) fprintf(vis_stderr, "        write error trace to dbg_out\n");
  (void) fprintf(vis_stderr, "    -h \tPrints this usage message.\n");
  (void) fprintf(vis_stderr, "    -i \tPrint input values in debug traces.\n");
  (void) fprintf(vis_stderr, "    -m \tPipe debug output through more\n");
  (void) fprintf(vis_stderr, "    -s \tWrite error trace in sim format.\n");
  (void) fprintf(vis_stderr, "    -t <period> time out period\n");
  (void) fprintf(vis_stderr, "    -v <verbosity_level>\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 0 => no feedback (Default)\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 1 => feedback\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 2 => feedback and CPU usage profile\n");
  (void) fprintf(vis_stderr, "    -D <dc_level>\n");
  (void) fprintf(vis_stderr, "        dc_level = 0 => no use of don't cares\n");
  (void) fprintf(vis_stderr, "        dc_level = 1 => use unreached states as don't cares (Default)\n");
  (void) fprintf(vis_stderr, "        dc_level = 2 => use unreached states as don't cares\n");

  (void) fprintf(vis_stderr, "                        and aggressive use of DC's to simplify MDD's\n");
  (void) fprintf(vis_stderr, "        dc_level = 3 => use over-approximate unreached states as don't cares\n");
  (void) fprintf(vis_stderr, "                        and aggressive use of DC's to simplify MDD's\n");
  (void) fprintf(vis_stderr, "    -S <schedule>\n");
  (void) fprintf(vis_stderr, "       schedule is one of {EL,EL1,EL2,budget,random,off}\n");
  (void) fprintf(vis_stderr, "    -F \tUse forward analysis in fixpoint computation.\n");
  (void) fprintf(vis_stderr, "    -L <lockstep_mode>\n");
  (void) fprintf(vis_stderr, "       lockstep_mode is one of {off,on,all,n}\n");

  McOptionsFree(options);

  return NIL(McOptions_t);
}


/**Function********************************************************************

  Synopsis [checks that all reachable states in flattened network satisfy invariants]

  CommandName [check_invariant]

  CommandSynopsis [check all states reachable in flattened network satisfy specified invariants]

  CommandArguments [ \[-c\] \[-d &lt;dbg_level&gt;\] \[-f\] \[-h\] \[-i\] \[-m\] \[-r\]
  \[-t &lt;time_out_period&gt;\] \[-v &lt;verbosity_level&gt;\] \[-A &lt;reachability_analysis_type&gt;\] \[-D\] &lt;invar_file&gt;]

  CommandDescription [Performs invariant checking on a flattened  network.
  Before calling this command, the user should have initialized the
  design by calling the command <A HREF="init_verifyCmd.html"> <code> init_verify</code></A>.
  <p>

  An invariant is a set of states.
  Checking the invariant is the process of determining that all states reachable
  from the initial states lie in the invariant.
  <p>


  One way of defining an invariant is through a CTL formula which has no path operators.
  Such formulas should be specified in the file <code>invar_file</code>.
  Note that the support  of any wire referred to in a formula should consist only
  of latches.
  For the precise syntax of CTL formulas,
  see the <A HREF="../ctl_ltl/ctl_ltl/ctl_ltl.html"> VIS CTL and LTL syntax manual</A>.
  <p>

  <code>check_invariant</code> ignores all fairness conditions associated with the FSM.
  <p>

  <code>check_invariant<code> involves reachability analysis where at
   every step of the reachability computation all the specified
   invariants are checked in the reachable states computed thus
   far. If some invariant does not hold, a proof of failure is
   demonstrated. This consists of a path starting from an initial
   state to a state lying outside the invariant. This path is made as
   short as possible. For the -A 0 option or default -A option, it is
   the shortest path leading to a state outside the invariant. If a
   set of invariants is specified, the failed formulas are reported as
   soon as they are detected. The check is continued with the
   remaining invariants.<p>

  Command options:
  <p>

  <dl>

  <dt> -d <code> &lt;dbg_level&gt; </code>
  <dd> Specify the amount of debugging performed when the system fails a formula being checked.
  <p>
  <dd>
  <code> dbg_level</code> must be one of the following:
  <p>

  <code> 0 </code>: No debugging  performed. This is the default.
  <p>

  <code> 1 </code>: Generate a path from an initial state to a state
  lying outside the invariant. This option stores the onion rings just
  as specifying -f would have. Therefore, it may take more time and
  memory if the formula passes. This option is incompatible with -A 2
  option. <p>

  <dt> -f
  <dd> Store the set of new states (onion rings) reached at each step of
  invariant. This option is likely to use more memory but possibly
  faster results for invariants that fail. Therefore, the debug
  information for a failed invariant, if requested, may be provided
  faster. This option is not compatible with -A 2 options.<p>

  <dt> -g &lt;<code>hints_file</code>&gt;</dt>

  <dd> Use guided search.  The file <code>hints_file</code> contains a
  series of hints.  A hint is a formula that does not contain any
  temporal operators, so <code>hints_file</code> has the same syntax
  as a file of invariants used for check_invariant.  The hints are
  used in the order given to change the transition relation.  The
  transition relation is conjoined with the hint to yield an
  underapproximation of the transition relation.  If the hints are
  cleverly chosen, this may speed up the computation considerably,
  because a search with the changed transition relation may be much
  simpler than one with the original transition relation, and results
  obtained can be reused, so that we may never have to do an expensive
  search with the original relation.  See also: Ravi and Somenzi,
  Hints to accelerate symbolic traversal. CHARME'99; Bloem, Ravi, and
  Somenzi, Efficient Decision Procedures for Model Checking of Linear
  Time Logic Properties, CAV'99; Bloem, Ravi, and Somenzi, Symbolic
  Guided Search for CTL Model Checking, DAC'00. This option is not
  compatible with -A 2 option. The description of some options for
  guided search can be found in the help page for
  print_guided_search_options. <p>
  
  <dt> -h
  <dd> Print the command usage.
  <p>

  <dt> -c
  <dd> Use the formula tree so that subformulae are not shared among the CTL
  formulae in the input file. This option is useful in the following
  scenario - formulae A, B and C are being checked in order and there is
  sub-formula sharing between A and C. If the BDDs corresponding to the shared
  sub-formula is huge then computation for B might not be able to finish
  without using this option.
  <p>

  <dt> -i
  <dd> Print input values causing transitions between states during
  debugging. Both primary and pseudo inputs are printed.  <p>

  <dt> -m
  <dd> Pipe debugger output through the UNIX utility  more.
  <p>

  <dt> -r
  <dd> Reduce the FSM derived from the flattened network with respect to each
  of the invariants in the input file. By default, the FSM is reduced with
  respect to the conjunction of the invariants in the input file. If this
  option is used and don't cares are being used for simplification, then
  subformula sharing is disabled (result might be incorrect otherwise).
  <p>

  <dd> The truth of  an invariant  may be independant  of  parts of the network (such as when wires have
  been abstracted; see <A HREF="flatten_hierarchyCmd.html"><code>flatten_hierarchy</code></A>).
  These parts are effectively removed when this option is invoked; this may result in
  more efficient invariant checking.
  <p>

  <dt> -t  <code>&lt;timeOutPeriod&gt;</code>
  <dd> Specify the time out period (in seconds) after which the command aborts. By default this option is set to infinity.
  <p>
  <dt> -v  <code>&lt;verbosity_level&gt;</code>
  <dd> Specify verbosity level. This sets the amount of feedback  on CPU usage and code status.
  <p>

  <dd>
  <code>verbosity_level</code>  must be one of the following:
  <p>

  <code> 0 </code>: No feedback provided. This is the default.<p>

  <code> 1 </code>: Feedback on code location.<p>

  <code> 2 </code>: Feedback on code location and CPU usage.<p>

  <dt> -A &lt;reachability_analysis_type&gt;
  <dd> This option allows specification of the type of reachability computation. <p>

  0:  (default) Breadth First Search. No approximate reachability computation. <p>

  1: High Density Reachability Analysis (HD). Computes reachable
  states in a manner that keeps BDD sizes under control. May be faster
  than BFS in some cases. For larger circuits, this option could
  compute more reachable states  than the -A 0 option for the same
  memory constraints, consequently may prove more invariants
  false. For help on controlling options for HD, look up help on the
  command: print_hd_options <A HREF="print_hd_optionsCmd.html">. Refer
  Ravi & Somenzi, ICCAD95.  The path generated for a failed invariant
  using this method may not be the shortest path. This option is
  available only when compiled with the CUDD package. <p>

  2. Approximate Reachability Don't Cares(ARDC). Computes
  over-approximated reachable states in the reachability analysis
  step. This may be faster than the -A 0 option . The invariants are
  checked in the over-approximation. This may produce false negatives,
  but these are resolved internally using the exact reachable
  states. The final results produced are the same as those for exact
  reachable states. For help on controlling options for ARDC, look up
  help on the command: print_ardc_options. <A
  HREF="print_ardc_optionsCmd.html"> Refer 2 papers in TCAD96 Dec. Cho
  et al, one is for State Space Decomposition and the other is for
  Approximate FSM Traversal. This option is incompatible with -d 1 and
  -g options.<p>

  <dt> -D</dt>
  
  <dd>First compute an overapproximation to the reachable states.  Minimize the
  transition relation using this approximation, and then compute the set of
  reachable states exactly. This may accelerate reachability analysis. Refer to
  the paper by Moon et al, ICCAD98. The BDD minimizing method can be chosen by
  using "set image_minimize_method <method>" <A HREF = "setCmd.txt">.  This
  option is incompatible with -g.  <p>

  <dt> -F &lt;<code>dbg_file</code>&gt;
  <dd> Write the debugger output to <code>dbg_file</code>.
  <p>

  <dt> &lt;invarFile&gt;
  <dd> File containing invariants to be checked.
  <p>
  </dl>

  <dt> Related "set" options: <p>
  <dt> rch_simulate &lt;#&gt;
  <dd> The set option can be used to set this flag rch_simulate to specify the
  number of random vectors to be simulated. Default value for this number is 0.
  <p>

  <dt> ctl_change_bracket &lt;yes/no&gt;
  <dd> Vl2mv automatically converts "\[\]" to "&lt;&gt;" in node names,
  therefore CTL parser does the same thing. However, in some cases a user
  does not want to change node names in CTL parsing. Then, use this set
  option by giving "no". Default is "yes".
  <p>

  <dt> See also command : compute_reach <p>
  ]

  SideEffects []

  SeeAlso [CommandMc]

******************************************************************************/

static int
CommandInv(
  Hrc_Manager_t **hmgr,
  int argc,
  char **argv)
{
  int i, j;
  mdd_t *tautology;
  McOptions_t *options;
  FILE *invarFile;
  array_t *invarArray, *formulas, *sortedFormulaArray;
  static Fsm_Fsm_t *totalFsm, *modelFsm;
  mdd_manager *mddMgr;
  Ntk_Network_t *network;
  Mc_VerbosityLevel verbosity;
  Mc_DcLevel dcLevel;
  array_t *invarNormalFormulaArray, *invarStatesArray;
  static int timeOutPeriod = 0;
  int debugFlag;
  int buildOnionRings;
  Fsm_RchType_t approxFlag;
  int ardc;
  int someLeft;
  Ctlp_Formula_t *invarFormula;
  mdd_t *invarFormulaStates;
  mdd_t *reachableStates;
  array_t *fsmArray;
  int printStep = 0;
  long initTime, finalTime;
  array_t *careSetArray;
  FILE              *guideFile;
  Ctlp_FormulaArray_t *hintsArray    = NIL(Fsm_HintsArray_t);
  array_t             *hintsStatesArray = NIL(array_t); /* array of mdd_t* */
  
  error_init();
  initTime = util_cpu_time();

  /* get command line options */
  if (!(options = ParseInvarOptions(argc, argv))) {
    return 1;
  }
  verbosity = McOptionsReadVerbosityLevel(options);
  dcLevel = McOptionsReadDcLevel(options);
  invarFile = McOptionsReadCtlFile(options);
  timeOutPeriod = McOptionsReadTimeOutPeriod(options);
  approxFlag = McOptionsReadInvarApproxFlag(options);
  buildOnionRings = (int)McOptionsReadInvarOnionRingsFlag(options);

  guideFile =  McOptionsReadGuideFile(options);
 
  if(guideFile != NIL(FILE) ){
    hintsArray = Mc_ReadHints(guideFile);
    fclose(guideFile); guideFile = NIL(FILE);
    if( hintsArray == NIL(array_t) ){
      McOptionsFree(options);
      return 1;
    }
  } /* if guided search */
  
  if(Img_UserSpecifiedMethod() != Img_Iwls95_c &&
     Img_UserSpecifiedMethod() != Img_Monolithic_c &&
     Img_UserSpecifiedMethod() != Img_Mlp_c &&
     guideFile != NIL(FILE)){
    fprintf(vis_stdout, 
 "** inv error: The Tfm and Hybrid image methods are incompatible with -g\n");
    McOptionsFree(options);
    return 1;
  }
  
  if (dcLevel == McDcLevelArdc_c)
    ardc = 1;
  else
    ardc = 0;
  
  /* read the array of invariants */
  invarArray = Ctlp_FileParseFormulaArray(invarFile);
  fclose(invarFile);
  if (invarArray == NIL(array_t)) {
    (void) fprintf(vis_stderr, "** inv error: Error in parsing invariants from file\n");
    McOptionsFree(options);
    return 1;
  }
  if (array_n(invarArray) == 0) {
    (void) fprintf(vis_stderr, "** inv error: No formula in file\n");
    McOptionsFree(options);
    return 1;
  }

  /* read the netwrok */
  network = Ntk_HrcManagerReadCurrentNetwork(*hmgr);
  if (network == NIL(Ntk_Network_t)) {
    fprintf(vis_stderr, "%s\n", error_string());
    error_init();
    McOptionsFree(options);
    return 1;
  }

  /* obtain the fsm and mdd manager */
  totalFsm = Fsm_NetworkReadOrCreateFsm(network);
  if (totalFsm == NIL(Fsm_Fsm_t)) {
    fprintf(vis_stderr, "%s\n", error_string());
    error_init();
    McOptionsFree(options);
    return 1;
  }
  mddMgr = Fsm_FsmReadMddManager(totalFsm);
  tautology = mdd_one(mddMgr);
  
  /* set time out */
  if (timeOutPeriod > 0) {
    (void) signal(SIGALRM, (void(*)(int))TimeOutHandle);
    (void) alarm(timeOutPeriod);
    if (setjmp(timeOutEnv) > 0) {
      (void) fprintf(vis_stdout, "# INV: Checking Invariant: timeout occurred after %d seconds.\n", timeOutPeriod);
      (void) fprintf(vis_stdout, "# INV: data may be corrupted.\n");
      alarm(0);
      return 1;
    }
  }

  /* debugFlag = 1 -> need to store/compute onion shells, else not */
  debugFlag = (McOptionsReadDbgLevel(options) != McDbgLevelNone_c);

  /* use formula tree if reduce option and dont-care level is high */
  if ((McOptionsReadReduceFsm(options) == TRUE) &&
     (dcLevel != McDcLevelNone_c)) {
      McOptionsSetUseFormulaTree(options, TRUE);
  }

  /* derive the normalized array of invariant formulas */
  if (McOptionsReadUseFormulaTree(options) == TRUE) {
    invarNormalFormulaArray =
      Ctlp_FormulaArrayConvertToExistentialFormTree(invarArray);
  } else {
    array_t *temp = Ctlp_FormulaArrayConvertToDAG( invarArray );
    array_free( invarArray );
    invarArray = temp;
    invarNormalFormulaArray = Ctlp_FormulaDAGConvertToExistentialFormDAG(
      invarArray ); 
  }

  if (array_n(invarNormalFormulaArray) == 0) {
    array_free(invarNormalFormulaArray);
    Ctlp_FormulaArrayFree(invarArray);
    mdd_free(tautology);
    return 1;
  }
  fsmArray = array_alloc(Fsm_Fsm_t *, 0);
  sortedFormulaArray = SortFormulasByFsm(totalFsm, invarNormalFormulaArray,
					 fsmArray, options);
  if (sortedFormulaArray == NIL(array_t)) {
    array_free(invarNormalFormulaArray);
    Ctlp_FormulaArrayFree(invarArray);
    mdd_free(tautology);
    return 1;
  }
  assert(array_n(fsmArray) == array_n(sortedFormulaArray));

  careSetArray = array_alloc(mdd_t *, 0);

  /* main loop for array of array of formulas. Each of the latter
     arrays corresponds to one reduced fsm. */
  arrayForEachItem(array_t *, sortedFormulaArray, i, formulas) {
    int *resultArray;
    int fail;
    /* initialize pass, fail array */
    resultArray = ALLOC(int, array_n(formulas));
    for (j = 0; j < array_n(formulas); j++) {
      resultArray[j] = 1;
    }
    /* get reduced fsm for this set of formulas */
    modelFsm = array_fetch(Fsm_Fsm_t *, fsmArray, i);

    /* evaluate hints for this reduced fsm, stop if the hints contain variables
     * outside this model FSM.
     */
    if (hintsArray != NIL(Ctlp_FormulaArray_t)) {
      int k;
      hintsStatesArray = Mc_EvaluateHints(modelFsm, hintsArray);
      if( hintsStatesArray == NIL(array_t)) { /* something wrong, clean up */
	int l;
	fprintf(vis_stdout, "Hints dont match the reduced FSM for this set of invariants.\n");
	fprintf(vis_stdout, "Continuing with the next set of invariants\n");
	for (k = i; k < array_n(sortedFormulaArray); k++) {
	  formulas = array_fetch(array_t *, sortedFormulaArray, k);
	  arrayForEachItem(Ctlp_Formula_t *, formulas, l, invarFormula) {
	    Ctlp_FormulaFree(invarFormula);
	  }
	  array_free(formulas);
	  /* get reduced fsm for this set of formulas */
	  modelFsm = array_fetch(Fsm_Fsm_t *, fsmArray, k);
	  /* free the Fsm if it was reduced here */
	  if (modelFsm != totalFsm) Fsm_FsmFree(modelFsm);
	}
	array_free(careSetArray);
	array_free(fsmArray);
	array_free(sortedFormulaArray);
	array_free(invarNormalFormulaArray);
	(void) fprintf(vis_stdout, "\n");
	
	Ctlp_FormulaArrayFree(invarArray);
	McOptionsFree(options);
	mdd_free(tautology);
	return 1;
      }	
    }/* hints exist */
    

    
    invarStatesArray = array_alloc(mdd_t *, 0);
    array_insert(mdd_t *, careSetArray, 0, tautology);
    arrayForEachItem(Ctlp_Formula_t *, formulas, j, invarFormula) {
      /* compute the set of states represented by the invariant. */
      invarFormulaStates =
	Mc_FsmEvaluateFormula(modelFsm, invarFormula, tautology,
			      NIL(Fsm_Fairness_t), careSetArray,
			      MC_NO_EARLY_TERMINATION,
			      NIL(Fsm_HintsArray_t), Mc_None_c,
			      verbosity, dcLevel, buildOnionRings,
			      McGSH_EL_c);
      
      array_insert_last(mdd_t *, invarStatesArray, invarFormulaStates);
    }

    printStep = (verbosity == McVerbosityMax_c) && (totalFsm == modelFsm);
    /* main loop to check a set of invariants. */
    do {
      boolean compute = FALSE;
      /* check if the computed reachable set in the total FSM already violates an invariant. */
      compute = TestInvariantsInTotalFsm(totalFsm, invarStatesArray, (debugFlag ||
							 buildOnionRings));
      
      /* compute reachable set or until early failure */
      if (compute) 
	reachableStates = Fsm_FsmComputeReachableStates(
	  modelFsm, 0, verbosity , 0, 0, (debugFlag || buildOnionRings), 0, 0,
	  approxFlag, ardc, 0, invarStatesArray, (verbosity > 1),
	  hintsStatesArray); 
      else if (debugFlag || buildOnionRings) {
	(void)Fsm_FsmReachabilityOnionRingsStates(totalFsm, &reachableStates);
      } else {
	reachableStates = mdd_dup(Fsm_FsmReadCurrentReachableStates(totalFsm));
      }
      
      ardc = 0; /* once ardc is applied, we don't need it again. */
      /* updates result array and sets fail if any formula failed.*/
      fail = UpdateResultArray(reachableStates, invarStatesArray, resultArray);
      mdd_free(reachableStates);
      someLeft = 0;
      if (fail) {
	/* some invariant failed */
	if (debugFlag) {
	  assert (approxFlag != Fsm_Rch_Oa_c);
	  InvarPrintDebugInformation(modelFsm, formulas, invarStatesArray,
				     resultArray, options, hintsStatesArray);
	} else if (approxFlag == Fsm_Rch_Oa_c) {
	  assert(!buildOnionRings);
	  /* May be a false negative */
	  /* undo failed results in the result array, report passed formulae */
	  arrayForEachItem(mdd_t *, invarStatesArray, j, invarFormulaStates) {
	    if (invarFormulaStates == NIL(mdd_t)) continue;
	    /* print all formulae that are known to have passed */
	    if (resultArray[j] == 1) {
	      mdd_free(invarFormulaStates);
	      array_insert(mdd_t *, invarStatesArray, j, NIL(mdd_t));
	      (void) fprintf(vis_stdout, "# INV: Early detection - formula passed --- ");
	      invarFormula = array_fetch(Ctlp_Formula_t *, formulas, j);
	      Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(invarFormula));
	      fprintf(vis_stdout, "\n");
	    } else resultArray[j] = 1;
	  }
	  fprintf(vis_stdout, "# INV: Invariant violated by over-approximated reachable states\n");
	  fprintf(vis_stdout, "# INV: Switching to BFS (exact computation) to resolve false negatives\n");
	  /* compute reachable set or until early failure */
	  reachableStates = Fsm_FsmComputeReachableStates(
	    modelFsm, 0, verbosity , 0, 0, 0, 0, 0, Fsm_Rch_Bfs_c, ardc, 0,
	    invarStatesArray, (verbosity > 1), hintsStatesArray);
	  /* either invariant has failed or all reachable states are computed */
	  /* updates result array */
	  fail = UpdateResultArray(reachableStates, invarStatesArray, resultArray);
	  mdd_free(reachableStates);
	}
	/* remove the failed invariants from the invariant list. */
	if (fail) {
	  arrayForEachItem(mdd_t *, invarStatesArray, j, invarFormulaStates) {
	    if (invarFormulaStates == NIL(mdd_t)) continue;
	    /* free the failed invariant mdds */
	    if (resultArray[j] == 0) {
	      mdd_free(invarFormulaStates);
	      array_insert(mdd_t *, invarStatesArray, j, NIL(mdd_t));
	      if (!debugFlag) {
		(void) fprintf(vis_stdout, "# INV: Early detection - formula failed --- ");
		invarFormula = array_fetch(Ctlp_Formula_t *, formulas, j);
		Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(invarFormula));
		fprintf(vis_stdout, "\n");
	      }
	    } else {
	      someLeft = 1;
	    }
	  }
	}
      } /* end of recomputation dur to over approximate computation */
    } while (someLeft);

    arrayForEachItem(mdd_t *, invarStatesArray, j, invarFormulaStates) {
      if (invarFormulaStates == NIL(mdd_t)) continue;
      /* free the passed invariant mdds */
      mdd_free(invarFormulaStates);
    }
    array_free(invarStatesArray);
    if (printStep) {
      finalTime = util_cpu_time();
      Fsm_FsmReachabilityPrintResults(modelFsm,finalTime-initTime ,approxFlag);
    }
    /* free the Fsm if it was reduced here */
    if (modelFsm != totalFsm) Fsm_FsmFree(modelFsm);
    PrintInvPassFail(formulas, resultArray);
    arrayForEachItem(Ctlp_Formula_t *, formulas, j, invarFormula) {
      Ctlp_FormulaFree(invarFormula);
    }
    array_free(formulas);
    FREE(resultArray);
  } /* end of processing the sorted array of array of invariants */
  
  if(hintsStatesArray != NIL(array_t))
    mdd_array_free(hintsStatesArray);
  if(hintsArray)
    Ctlp_FormulaArrayFree(hintsArray);
  array_free(careSetArray);
  array_free(fsmArray);
  array_free(sortedFormulaArray);
  array_free(invarNormalFormulaArray);
  (void) fprintf(vis_stdout, "\n");

  Ctlp_FormulaArrayFree(invarArray);
  McOptionsFree(options);
  mdd_free(tautology);

  alarm(0);
  return 0;
}


/**Function********************************************************************

  Synopsis    [Parse command line options for invar.]

  SideEffects []

******************************************************************************/
static McOptions_t *
ParseInvarOptions(
  int argc,
  char **argv)
{
  unsigned int i;
  int c;
  boolean useMore = FALSE;
  boolean reduceFsm = FALSE;
  boolean printInputs = FALSE;
  boolean useFormulaTree = FALSE;
  FILE *inputFp=NIL(FILE);
  FILE *debugOut=NIL(FILE);
  char *debugOutName=NIL(char);
  McDbgLevel dbgLevel = McDbgLevelNone_c;
  Mc_VerbosityLevel verbosityLevel = McVerbosityNone_c;
  int timeOutPeriod = 0;
  McOptions_t *options = McOptionsAlloc();
  int approxFlag = 0, ardc = 0, shellFlag = 0;
  Fsm_RchType_t rchType;
  FILE *guideFile;
  char *guideFileName = NIL(char);
  
  /*
   * Parse command line options.
   */

  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "ifcrt:g:hmv:d:A:DF:")) != EOF) {
    switch(c) {
      case 'g':
        guideFileName = util_strsav(util_optarg);
        break;
      case 'h':
	goto usage;
      case 't':
	timeOutPeriod = atoi(util_optarg);
	break;
      case 'v':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
	verbosityLevel = (Mc_VerbosityLevel) atoi(util_optarg);
	break;
      case 'd':
	for (i = 0; i < strlen(util_optarg); i++) {
	  if (!isdigit((int)util_optarg[i])) {
	    goto usage;
	  }
	}
	dbgLevel = (McDbgLevel) atoi(util_optarg);
	break;
      case 'f':
	shellFlag = 1;
	break;
      case 'r' :
	reduceFsm = TRUE;
	break;
      case 'm':
	useMore = TRUE;
	break;
      case 'i' :
	printInputs = TRUE;
	break;
      case 'c' :
	useFormulaTree = TRUE;
	break;
      case 'A':
	approxFlag = atoi(util_optarg);
	if ((approxFlag > 2) || (approxFlag < 0)) {
	    goto usage;
	}
	break;
      case 'D':
	ardc = 1;
	break;
      case 'F':
	debugOutName = util_strsav(util_optarg);
	break;
      default:
	goto usage;
    }
  }

  /* translate approx flag */
  switch(approxFlag) {
  case 0: rchType = Fsm_Rch_Bfs_c; break;
  case 1: rchType = Fsm_Rch_Hd_c; break;
  case 2: rchType = Fsm_Rch_Oa_c; break;
  default: rchType = Fsm_Rch_Default_c; break;
  }

  if ((approxFlag == 1) && (bdd_get_package_name() != CUDD)) {
    fprintf(vis_stderr, "** inv error: check_invariant with -A 1 option is currently supported by only CUDD.\n");
    return NIL(McOptions_t);
  }

  if (rchType == Fsm_Rch_Oa_c) {
    if (dbgLevel > 0) {
      fprintf(vis_stdout, "Over approximation and debug level 1 are incompatible\n");
      fprintf(vis_stdout, "Read check_invariant help page\n");
      goto usage;
    } else if (shellFlag == 1) {
      fprintf(vis_stdout, "Over approximation and shell flag are incompatible\n");
      fprintf(vis_stdout, "Read check_invariant help page\n");
      goto usage;
    }
  }

  if (guideFileName != NIL(char)) {
    guideFile = Cmd_FileOpen(guideFileName, "r", NIL(char *), 0);
    FREE(guideFileName);
    if (guideFile == NIL(FILE)) {
      McOptionsFree(options);
      return NIL(McOptions_t);
    }
    McOptionsSetGuideFile(options, guideFile);
  }
  
  /*
   * Open the ctl file.
   */
  if (argc - util_optind == 0) {
    (void) fprintf(vis_stderr, "** inv error: file containing invariant formulas not provided\n");
    goto usage;
  }
  else if (argc - util_optind > 1) {
    (void) fprintf(vis_stderr, "** inv error: too many arguments\n");
    goto usage;
  }

  McOptionsSetUseMore(options, useMore);
  McOptionsSetReduceFsm(options, reduceFsm);
  McOptionsSetPrintInputs(options, printInputs);
  McOptionsSetUseFormulaTree(options, useFormulaTree);
  McOptionsSetTimeOutPeriod(options, timeOutPeriod);
  McOptionsSetInvarApproxFlag(options, rchType);
  if (ardc)
    McOptionsSetDcLevel(options, McDcLevelArdc_c);
  else
    McOptionsSetDcLevel(options, McDcLevelNone_c);
  McOptionsSetInvarOnionRingsFlag(options, shellFlag);

  inputFp = Cmd_FileOpen(argv[util_optind], "r", NIL(char *), 0);
  if (inputFp == NIL(FILE)) {
    McOptionsFree(options);
    return NIL(McOptions_t);
  }
  McOptionsSetCtlFile(options, inputFp);

  if (debugOutName != NIL(char)) {
    debugOut = Cmd_FileOpen(debugOutName, "w", NIL(char *), 0);
    if (debugOut == NIL(FILE)) {
      McOptionsFree(options);
      return NIL(McOptions_t);
    }
  }
  McOptionsSetDebugFile(options, debugOut);

  if ((verbosityLevel != McVerbosityNone_c) &&
       (verbosityLevel != McVerbosityLittle_c) &&
       (verbosityLevel != McVerbositySome_c) &&
       (verbosityLevel != McVerbosityMax_c)) {
    goto usage;
  }
  else {
    McOptionsSetVerbosityLevel(options, verbosityLevel);
  }

  if ((dbgLevel != McDbgLevelNone_c) &&
       (dbgLevel != McDbgLevelMin_c)) {
    goto usage;
  }
  else {
    McOptionsSetDbgLevel(options, dbgLevel);
  }


  return options;

  usage:
  (void) fprintf(vis_stderr, "usage: check_invariant [-c][-d dbg_level][-f][-g <hintfile>][-h][-i][-m][-r][-v verbosity_level][-t time_out][-A <number>][-D dc_level] [-F <dbg_out>] <invar_file>\n");
  (void) fprintf(vis_stderr, "    -c  avoid sub-formula sharing between formulae\n");
  (void) fprintf(vis_stderr, "    -d <dbg_level>\n");
  (void) fprintf(vis_stderr, "        debug_level = 0 => no debugging (Default)\n");
  (void) fprintf(vis_stderr, "        debug_level = 1 => print path to state failing invariant\n");
  (void) fprintf(vis_stderr, "    -f \tStore the onion rings at each step.\n");
  (void) fprintf(vis_stderr, "    -g <hint_file> \tSpecify file for guided search.\n");
  (void) fprintf(vis_stderr, "    -i \tPrint input values in debug traces.\n");
  (void) fprintf(vis_stderr, "    -m  pipe debugger output through UNIX utility more\n");
  (void) fprintf(vis_stderr, "    -r  reduce FSM with respect to invariant being checked\n");
  (void) fprintf(vis_stderr, "    -t <period> Time out period.\n");
  (void) fprintf(vis_stderr, "    -v <verbosity_level>\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 0 => no feedback (Default)\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 1 => code status\n");
  (void) fprintf(vis_stderr, "        verbosity_level = 2 => code status and CPU usage profile\n");
  (void) fprintf(vis_stderr, "    -A  <number> Use Different types of reachability analysis\n");
  (void) fprintf(vis_stderr, "            0 (default) - BFS method.\n");
  (void) fprintf(vis_stderr, "            1 - HD method.\n");
  (void) fprintf(vis_stderr, "            2 - Over-approximation.\n");
  (void) fprintf(vis_stderr, "    -D  minimize transition relation with ARDCs\n");
  (void) fprintf(vis_stderr, "    -F <dbg_out>\n");
  (void) fprintf(vis_stderr, "        write error trace to dbg_out\n");
  (void) fprintf(vis_stderr, "    <invar_file> The input file containing invariants to be checked.\n");
  (void) fprintf(vis_stderr, "    -h \tPrints this usage message.\n");
  McOptionsFree(options);

  return NIL(McOptions_t);
}

/**Function********************************************************************

  Synopsis    [Handle function for timeout.]

  Description [This function is called when the time out occurs.]

  SideEffects []

******************************************************************************/
static void
TimeOutHandle(void)
{
  longjmp(timeOutEnv, 1);
}

/**Function********************************************************************

  Synopsis    [Reachable(p,q) = FwdUntil(p,q) ^ q]

  Description [Reachable(p,q) = FwdUntil(p,q) ^ q. This Reachable operation is
  described in the paper "CTL Model Checking Based on Forward State Traversal"
  by H. Iwashita et al.]

  SideEffects []

******************************************************************************/
static mdd_t *
McForwardReachable(
  Fsm_Fsm_t *modelFsm,
  mdd_t *targetMdd,
  mdd_t *invariantMdd,
  mdd_t *fairStates,
  array_t *careStatesArray,
  array_t *onionRings,
  Mc_VerbosityLevel verbosity,
  Mc_DcLevel dcLevel)
{
  mdd_t *resultMdd, *fuMdd;

  /* When McForwardReachable is called by Mc_FsmEvaluateFwdEHFormula
   * with no fairness constraints, this test saves one image
   * computation. */
  if (mdd_is_tautology(fairStates, 1) &&
      mdd_lequal_mod_care_set_array(invariantMdd, targetMdd,
				    1, 1, careStatesArray)) {
    resultMdd = mdd_dup(invariantMdd);
    if (onionRings) {
      array_insert_last(mdd_t *, onionRings, mdd_dup(resultMdd));
    }
  } else {
    fuMdd = Mc_FsmEvaluateFwdUFormula(modelFsm, targetMdd, invariantMdd,
				      fairStates, careStatesArray, onionRings,
				      verbosity, dcLevel);
    resultMdd = mdd_and(fuMdd, invariantMdd, 1, 1);
    mdd_free(fuMdd);
  }

  /* Updates onionRings not to have warning message in
   * Mc_FsmEvaluateFwdEHFormula since Reachable(p,q) = FwdUntil(p,q) ^ q.
   */
  if (onionRings) {
    int		i;
    mdd_t	*ring, *tmp;

    for (i = 0 ; i < array_n(onionRings); i++) {
      tmp = array_fetch(mdd_t *, onionRings, i);
      ring = mdd_and(tmp, invariantMdd, 1, 1);
      mdd_free(tmp);
      array_insert(mdd_t *, onionRings, i, ring);
    }
  }

  return(resultMdd);
}

/**Function********************************************************************

  Synopsis [Prints out a path from the initial state to a state that
  violates the invariant. ]

  Description [Prints out a path from the initial state to a state
  that violates the invariant. First it checks whether there any onion
  rings built for this fsm. If the onion rings already exist, it
  checks whether there is any violation in these rings. If not,
  additional onion rings are computed.]

  SideEffects []

******************************************************************************/
static void
InvarPrintDebugInformation(Fsm_Fsm_t *modelFsm,
			   array_t *invarFormulaArray,
			   array_t *invarStatesArray,
			   int *resultArray,
			   McOptions_t *options,
			   array_t *hintsStatesArray)
{
  FILE *debugFile = McOptionsReadDebugFile(options);
  FILE *tmpFile = vis_stdout;
  extern FILE *vis_stdpipe;
  mdd_t *outerOnionRing=NIL(mdd_t);
  mdd_t *reachableBadStates;
  mdd_t *aBadReachableState, *reachableStates;
  array_t *aPath;
  boolean computeReach = TRUE;
  int ringsUpToDate;
  array_t *onionRings=NIL(array_t);
  int i, j;
  Fsm_RchType_t approxFlag;
  Ctlp_Formula_t *invarFormula;
  mdd_t *invarFormulaStates;
  approxFlag = McOptionsReadInvarApproxFlag(options);

  if (debugFile)
    vis_stdpipe = debugFile;
  else if (McOptionsReadUseMore(options) == TRUE)
    vis_stdpipe = popen("more", "w");
  else
    vis_stdpipe = vis_stdout;
  vis_stdout = vis_stdpipe;

  arrayForEachItem(mdd_t *, invarStatesArray, i, invarFormulaStates) {
    if ((invarFormulaStates == NIL(mdd_t)) || (resultArray[i] == 1))
      continue;
    onionRings = Fsm_FsmReadReachabilityOnionRings(modelFsm);
    ringsUpToDate = Fsm_FsmTestReachabilityOnionRingsUpToDate(modelFsm);
    /* search for onion ring intersecting invarStates' */
    if (onionRings != NIL(array_t)) {
      for (j = 0; j < array_n(onionRings); j++) {
	outerOnionRing = array_fetch(mdd_t *, onionRings, j);
	if (!mdd_lequal(outerOnionRing, invarFormulaStates, 1, 1)) {
	  computeReach = FALSE;
	  break;
	}
      }
    }

    /* no onion ring found, hence recompute with onion rings */
    if (computeReach == TRUE) {
      Mc_VerbosityLevel verbosity;
      int debugFlag;
      int ardc;
      int dcLevel = McOptionsReadDcLevel(options);
      verbosity = McOptionsReadVerbosityLevel(options);
      debugFlag = (McOptionsReadDbgLevel(options) != McDbgLevelNone_c);
      if (dcLevel == McDcLevelArdc_c)
	ardc = 1;
      else
	ardc = 0;
      assert(!ringsUpToDate);
      reachableStates = Fsm_FsmComputeReachableStates(
	modelFsm, 0, (int)verbosity , 0, 0, debugFlag, 0, 0, approxFlag, ardc,
	0, invarStatesArray, (verbosity > 1), hintsStatesArray);
      mdd_free(reachableStates);
      /* find the onion ring with invariant failure */
      onionRings = Fsm_FsmReadReachabilityOnionRings(modelFsm);
      if (onionRings == NIL(array_t)) {
	fprintf(vis_stdout, "Unable to generate onion rings for counterexample.\n");
	fprintf(vis_stdout, "Try again with standard (without -A) options.\n");
	if (vis_stdout != tmpFile && vis_stdout != debugFile)
	  (void)pclose(vis_stdpipe);
	vis_stdout = tmpFile;
	return;
      }
      ringsUpToDate = Fsm_FsmTestReachabilityOnionRingsUpToDate(modelFsm);
      for (j = 0; j < array_n(onionRings); j++) {
	outerOnionRing = array_fetch(mdd_t *, onionRings, j);
	if (!mdd_lequal(outerOnionRing, invarFormulaStates, 1, 1)) {
	  break;
	}
      }
    }
    invarFormula = array_fetch(Ctlp_Formula_t *, invarFormulaArray, i);
    /* this is the case of  true negative */
    if (outerOnionRing != NIL(mdd_t)) {
      /* compute the set of bad states */
      reachableBadStates = mdd_and(invarFormulaStates, outerOnionRing, 0, 1);
      /* pick one bad state */
      aBadReachableState = Mc_ComputeAState(reachableBadStates, modelFsm);
      mdd_free(reachableBadStates);

      /* build a path from the initial states to the bad state */
      aPath = Mc_BuildPathFromCore(aBadReachableState, onionRings,
				   modelFsm, McGreaterOrEqualZero_c);
      mdd_free(aBadReachableState);
      (void) fprintf(vis_stdout, "# INV: formula %d failed --- ", i+1);
      Ctlp_FormulaPrint(vis_stdout,
			Ctlp_FormulaReadOriginalFormula(invarFormula));
      fprintf(vis_stdout, "\n");

      (void) fprintf(vis_stdout, "# INV: calling debugger\n");
      if (array_n(aPath) > 1) {
	(void) fprintf(vis_stdout,
			"# INV: a sequence of states starting at an initial ");
	(void) fprintf(vis_stdout, "state leading to a bad state\n");
      } else {
	(void) fprintf(vis_stdout,
		    "# INV: invariant fails at the following initial state\n");
      }
      (void) Mc_PrintPath(aPath, modelFsm, McOptionsReadPrintInputs(options));
      (void) fprintf(vis_stdout, "\n");

      mdd_array_free(aPath);
    } else {
      /* Was a false negative */
      assert (approxFlag == Fsm_Rch_Oa_c);
      (void) fprintf(vis_stdout, "# INV: formula passed --- ");
      Ctlp_FormulaPrint(vis_stdout,
			Ctlp_FormulaReadOriginalFormula(invarFormula));
      fprintf(vis_stdout, "\n");
    }
  }

  if (vis_stdout != tmpFile && vis_stdout != debugFile)
    (void)pclose(vis_stdpipe);
  vis_stdout = tmpFile;
} /* end of InvarPrintDebugInformation */

/**Function********************************************************************

  Synopsis    [Sort formulae by fsm]

  Description [Sort formulae by fsm. This creates an array of array of
  formulae. Each element of the outer array corresponds to an fsm
  reduced w.r.t the element (array of formulae). Both the array of
  formulae and the reduced fsms are returned. If there is no
  reduction, the reduced fsm is set to the totalFsm. While freeing the
  reduced fsms in the fsmArray, if the reduced fsm is set to total
  fsm, DO NOT free. ]

  SideEffects [fsmArray is populated with reduced fsms.]

******************************************************************************/
static array_t *
SortFormulasByFsm(Fsm_Fsm_t *totalFsm,
		  array_t *invarFormulaArray,
		  array_t *fsmArray,
		  McOptions_t *options)
{
  array_t *resultArray;
  array_t *formulaArray;
  array_t *tempArray;
  int i, j, found;
  Ctlp_Formula_t *formula;
  Ntk_Network_t *network = Fsm_FsmReadNetwork(totalFsm);
  Fsm_Fsm_t *reducedFsm, *fsm;
  int reducedFsmPos = -1;

  if (invarFormulaArray == NIL(array_t)) return invarFormulaArray;
  if (array_n(invarFormulaArray) == 0) return NIL(array_t);

  tempArray = array_alloc(array_t *, 0);
  /* do a semantic check and a quantifier check */
  arrayForEachItem (Ctlp_Formula_t *, invarFormulaArray, i, formula) {
    if (!Mc_FormulaStaticSemanticCheckOnNetwork(formula, network, FALSE)){
      (void) fprintf(vis_stderr, "** inv error: error in parsing Atomic Formula:\n%s\n", error_string());
      error_init();
      Ctlp_FormulaFree(formula);
      continue;
    } else {
      if (Ctlp_FormulaTestIsQuantifierFree(formula) == FALSE) {
	(void) fprintf(vis_stderr, "** inv error: invariant formula refers to path quantifiers\n");
	Ctlp_FormulaFree(formula);
	continue;
      }
    }
    array_insert_last(Ctlp_Formula_t *, tempArray, formula);
  }
  if (array_n(tempArray) == 0) {
    array_free(tempArray);
    return NIL(array_t);
  }

  /* if no -r option, return the original array */
  resultArray = array_alloc(array_t *, 0);
  if (McOptionsReadReduceFsm(options) == FALSE) {
    array_insert_last(array_t *, resultArray, tempArray);
    reducedFsm = McConstructReducedFsm(network, tempArray);
    if (reducedFsm != NIL(Fsm_Fsm_t)) {
      array_insert_last(Fsm_Fsm_t *, fsmArray, reducedFsm);
    } else {
      array_insert_last(Fsm_Fsm_t *, fsmArray, totalFsm);
    }
    return resultArray;
  }

  /* We want to minimize per formula only when "-r" option is not specified */
  arrayForEachItem (Ctlp_Formula_t *, tempArray, i, formula) {
    /* Create reduced fsm */
    formulaArray = array_alloc(Ctlp_Formula_t *, 0);
    array_insert_last(Ctlp_Formula_t *, formulaArray, formula);

    reducedFsm = McConstructReducedFsm(network, formulaArray);
    if (reducedFsm == NIL(Fsm_Fsm_t)) {
      /* no reduction */
      reducedFsm = totalFsm;
    }

    /* Check if fsm already created */
    found = 0;
    for (j = 0; j < array_n(fsmArray); j++) {
      fsm = array_fetch (Fsm_Fsm_t *, fsmArray, j);
      found = Fsm_FsmCheckSameSubFsmInTotalFsm(totalFsm, reducedFsm, fsm);
      if (found) break;
    }

    /* If found, add to the existing list. Else add a new reduced fsm */
    if (found) {
      if (reducedFsm != totalFsm) Fsm_FsmFree(reducedFsm);
      array_free(formulaArray);
      formulaArray = array_fetch(array_t *, resultArray, j);
      array_insert_last(Ctlp_Formula_t *, formulaArray, formula);
    } else {
      array_insert_last(Fsm_Fsm_t *, fsmArray, reducedFsm);
      array_insert_last(array_t *, resultArray, formulaArray);
      if (reducedFsm == totalFsm) reducedFsmPos = array_n(fsmArray)-1;
    }
  }

  array_free(tempArray);
  /* swap the last reduced fsm with the total fsm. */
  if ((reducedFsmPos != -1) && (reducedFsmPos != array_n(fsmArray)-1)) {
    array_t *totalFsmFormulaArray;
    assert(array_n(fsmArray) == array_n(resultArray));
    fsm = array_fetch(Fsm_Fsm_t *, fsmArray, array_n(fsmArray)-1);
    array_insert(Fsm_Fsm_t *, fsmArray, reducedFsmPos, fsm);
    array_insert(Fsm_Fsm_t *, fsmArray, array_n(fsmArray)-1, totalFsm);
    formulaArray =  array_fetch(array_t *, resultArray, array_n(resultArray)-1);
    totalFsmFormulaArray = array_fetch(array_t *, resultArray, reducedFsmPos);
    array_insert(array_t *, resultArray, reducedFsmPos, formulaArray);
    array_insert(array_t *, resultArray, array_n(resultArray)-1, totalFsmFormulaArray);
  }

  return resultArray;
} /* end of SortFormulasByFsm */

/**Function********************************************************************

  Synopsis [Check whether the reachable states violate the array of
  invariants.]

  Description [Check whether the reachable states violate the array of
  invariants. Return 1 if any formula failed. Update the result
  array. Entry 0 implies fail, 1 implies pass.]

  SideEffects []

******************************************************************************/
static int
UpdateResultArray(mdd_t *reachableStates,
		  array_t *invarStatesArray,
		  int *resultArray)
{
  int i;
  mdd_t *invariant;
  int fail = 0;
  arrayForEachItem(mdd_t *, invarStatesArray, i, invariant) {
    if (invariant == NIL(mdd_t)) continue;
    if (!mdd_lequal(reachableStates, invariant, 1, 1)) {
      fail = 1;
      resultArray[i] = 0;
    }
  }
  return fail;
} /* end of UpdateResultArray */

/**Function********************************************************************

  Synopsis [Prints whether model checking passed or failed]

  Description [Prints whether model checking passed or failed, and if
  requested, a debug trace.]

  SideEffects []

******************************************************************************/
static void
PrintMcPassFail(mdd_manager *mddMgr,
		Fsm_Fsm_t *modelFsm,
		Mc_FwdBwdAnalysis traversalDirection,
		Ctlp_Formula_t *ctlFormula,
		mdd_t *ctlFormulaStates,
		mdd_t *modelInitialStates,
		array_t *modelCareStatesArray,
		McOptions_t *options,
		Mc_VerbosityLevel verbosity)
{
  if (traversalDirection == McFwd_c) {
    if (bdd_is_tautology(ctlFormulaStates, 1))
      (void)fprintf(vis_stdout, "# MC: formula passed --- ");
    else
      (void)fprintf(vis_stdout, "# MC: formula failed --- ");
    Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(ctlFormula));
    (void)fprintf(vis_stdout, "\n");
  } else if (mdd_lequal(modelInitialStates, ctlFormulaStates, 1, 1)) {
    (void)fprintf(vis_stdout, "# MC: formula passed --- ");
    Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(ctlFormula));
    (void)fprintf(vis_stdout, "\n");
  } else {
    (void)fprintf(vis_stdout, "# MC: formula failed --- ");
    Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(ctlFormula));
    (void)fprintf(vis_stdout, "\n");
    if (McOptionsReadDbgLevel(options) != McDbgLevelNone_c)
      McFsmDebugFormula(options, ctlFormula, modelFsm, modelCareStatesArray);
  }
}

/**Function********************************************************************

  Synopsis [Prints whether the set of invariants passed or failed]

  Description [Prints whether the set of invariants passed or failed.]

  SideEffects []

******************************************************************************/
static void
PrintInvPassFail(array_t *invarFormulaArray,
		 int *resultArray)
{
  int i;
  Ctlp_Formula_t *formula;
  fprintf(vis_stdout, "\n# INV: Summary of invariant pass/fail\n");
  arrayForEachItem(Ctlp_Formula_t *, invarFormulaArray, i, formula) {
    if (resultArray[i]) {
      (void) fprintf(vis_stdout, "# INV: formula passed --- ");
      Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(formula));
      fprintf(vis_stdout, "\n");
    } else {
      (void) fprintf(vis_stdout, "# INV: formula failed --- ");
      Ctlp_FormulaPrint(vis_stdout, Ctlp_FormulaReadOriginalFormula(formula));
      fprintf(vis_stdout, "\n");
    }
  }
  return;
} /* end of PrintInvPassFail */


/**Function********************************************************************

  Synopsis [Test if an invariant is violated in the current reachable
  set of the given FSM.]

  Description [Test if an invariant is violated in the current
  reachable set of the given FSM. Returns FALSE if an invariant is
  violated and TRUE is no invariant is violated.]

  SideEffects []

******************************************************************************/
static int
TestInvariantsInTotalFsm(Fsm_Fsm_t *totalFsm,
			 array_t *invarStatesArray,
			 int shellFlag)
{
  int i;
  mdd_t *invariant;
  mdd_t *reachableStates;
  boolean upToDate;
  if (shellFlag) {
    upToDate = Fsm_FsmReachabilityOnionRingsStates(totalFsm, &reachableStates);
  } else {
    reachableStates = Fsm_FsmReadCurrentReachableStates(totalFsm);
  }
  if (reachableStates == NIL(mdd_t)) return TRUE;
  if (mdd_is_tautology(reachableStates, 0)) return TRUE;
  
  arrayForEachItem(mdd_t *, invarStatesArray, i, invariant) {
    if (invariant == NIL(mdd_t)) continue;
    if (!mdd_lequal(reachableStates, invariant, 1, 1)) {
      if (shellFlag) mdd_free(reachableStates);
      return FALSE;
    }
  }
  if (shellFlag) mdd_free(reachableStates);
  if ((shellFlag && !upToDate) ||
      (Fsm_FsmReadReachabilityApproxComputationStatus(totalFsm) ==
       Fsm_Rch_Under_c))
    return TRUE;
  else
    return FALSE;
    
}


/**Function********************************************************************

  Synopsis    [Compute the approximation of the result for
  EvaluateFormulaRecur.]

  Description []

  SideEffects [None]

  SeeAlso     [EvaluateFormulaRecur]

******************************************************************************/
static Ctlp_Approx_t
ComputeResultingAproximation(
  Ctlp_FormulaType formulaType,
  Ctlp_Approx_t leftApproxType,
  Ctlp_Approx_t rightApproxType,
  Ctlp_Approx_t TRMinimization,
  Mc_EarlyTermination_t *earlyTermination,
  boolean fixpoint)
{
  Ctlp_Approx_t thisApproxType;

  switch (formulaType) {
  case Ctlp_ID_c:
  case Ctlp_TRUE_c:
  case Ctlp_FALSE_c:
    thisApproxType = Ctlp_Exact_c;
    break;
  case Ctlp_NOT_c:
    if (leftApproxType == Ctlp_Incomparable_c) {
      thisApproxType = Ctlp_Incomparable_c;
    } else if (leftApproxType == Ctlp_Overapprox_c) {
      thisApproxType = Ctlp_Underapprox_c;
    } else if (leftApproxType == Ctlp_Underapprox_c) {
      thisApproxType = Ctlp_Overapprox_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    break;
  case Ctlp_AND_c:
    if (leftApproxType == Ctlp_Incomparable_c ||
	rightApproxType == Ctlp_Incomparable_c ||
	(leftApproxType == Ctlp_Overapprox_c &&
	 rightApproxType == Ctlp_Underapprox_c) ||
	(leftApproxType == Ctlp_Underapprox_c &&
	 rightApproxType == Ctlp_Overapprox_c)) {
      thisApproxType = Ctlp_Incomparable_c;
    } else if (rightApproxType == Ctlp_Overapprox_c) {
      if ((earlyTermination == MC_NO_EARLY_TERMINATION ||
	   earlyTermination->mode == McCare_c) &&
	  leftApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Exact_c;
      } else {
	thisApproxType = Ctlp_Overapprox_c;
      }
    } else if (leftApproxType == Ctlp_Overapprox_c) {
      thisApproxType = Ctlp_Overapprox_c;
    } else if (rightApproxType == Ctlp_Underapprox_c) {
      if ((earlyTermination == MC_NO_EARLY_TERMINATION ||
	   earlyTermination->mode == McCare_c) &&
	  leftApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Exact_c;
      } else {
	thisApproxType = Ctlp_Underapprox_c;
      }
    } else if (leftApproxType == Ctlp_Underapprox_c) {
      thisApproxType = Ctlp_Underapprox_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    break;
  case Ctlp_EQ_c:
  case Ctlp_XOR_c:
    if (leftApproxType != Ctlp_Exact_c || rightApproxType != Ctlp_Exact_c) {
      thisApproxType = Ctlp_Incomparable_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    break;
  case Ctlp_THEN_c:
    if (leftApproxType == Ctlp_Incomparable_c ||
	rightApproxType == Ctlp_Incomparable_c ||
	(leftApproxType == Ctlp_Overapprox_c &&
	 rightApproxType == Ctlp_Overapprox_c) ||
	(leftApproxType == Ctlp_Underapprox_c &&
	 rightApproxType == Ctlp_Underapprox_c)) {
      thisApproxType = Ctlp_Incomparable_c;
    } else if (rightApproxType == Ctlp_Overapprox_c) {
      if ((earlyTermination == MC_NO_EARLY_TERMINATION ||
	   earlyTermination->mode == McCare_c) &&
	  leftApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Exact_c;
      } else {
	thisApproxType = Ctlp_Overapprox_c;
      }
    } else if (leftApproxType == Ctlp_Underapprox_c) {
      thisApproxType = Ctlp_Overapprox_c;
    } else if (rightApproxType == Ctlp_Underapprox_c) {
      if ((earlyTermination == MC_NO_EARLY_TERMINATION ||
	   earlyTermination->mode == McCare_c) &&
	  leftApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Exact_c;
      } else {
	thisApproxType = Ctlp_Underapprox_c;
      }
    } else if (leftApproxType == Ctlp_Overapprox_c) {
      thisApproxType = Ctlp_Underapprox_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    break;
  case Ctlp_OR_c:
    if (leftApproxType == Ctlp_Incomparable_c ||
	rightApproxType == Ctlp_Incomparable_c ||
	(leftApproxType == Ctlp_Overapprox_c &&
	 rightApproxType == Ctlp_Underapprox_c) ||
	(leftApproxType == Ctlp_Underapprox_c &&
	 rightApproxType == Ctlp_Overapprox_c)) {
      thisApproxType = Ctlp_Incomparable_c;
    } else if (rightApproxType == Ctlp_Overapprox_c) {
      if ((earlyTermination == MC_NO_EARLY_TERMINATION ||
	   earlyTermination->mode == McCare_c) &&
	  leftApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Exact_c;
      } else {
	thisApproxType = Ctlp_Overapprox_c;
      }
    } else if (leftApproxType == Ctlp_Overapprox_c) {
      thisApproxType = Ctlp_Overapprox_c;
    } else if (rightApproxType == Ctlp_Underapprox_c) {
      if ((earlyTermination == MC_NO_EARLY_TERMINATION ||
	   earlyTermination->mode == McCare_c) &&
	  leftApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Exact_c;
      } else {
	thisApproxType = Ctlp_Underapprox_c;
      }
    } else if (leftApproxType == Ctlp_Underapprox_c) {
      thisApproxType = Ctlp_Underapprox_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    break;
  case Ctlp_EX_c:
    if (leftApproxType == Ctlp_Underapprox_c) {
      if (TRMinimization == Ctlp_Overapprox_c) {
	thisApproxType = Ctlp_Incomparable_c;
      } else {
	thisApproxType = Ctlp_Underapprox_c;
      }
    } else if (leftApproxType == Ctlp_Overapprox_c) {
      if (TRMinimization == Ctlp_Underapprox_c) {
	thisApproxType = Ctlp_Incomparable_c;
      } else {
	thisApproxType = Ctlp_Overapprox_c;
      }
    } else if (leftApproxType == Ctlp_Exact_c) {
      thisApproxType = TRMinimization;
    } else {
      thisApproxType = Ctlp_Incomparable_c;
    }
    break;
  case Ctlp_EU_c:
    if (leftApproxType == Ctlp_Incomparable_c ||
	rightApproxType == Ctlp_Incomparable_c ||
	(leftApproxType == Ctlp_Overapprox_c &&
	 rightApproxType == Ctlp_Underapprox_c) ||
	(leftApproxType == Ctlp_Underapprox_c &&
	 rightApproxType == Ctlp_Overapprox_c)) {
      thisApproxType = Ctlp_Incomparable_c;
    } else if (leftApproxType == Ctlp_Overapprox_c ||
	       rightApproxType == Ctlp_Overapprox_c) {
      thisApproxType = Ctlp_Overapprox_c;
    } else if (leftApproxType == Ctlp_Underapprox_c ||
	       rightApproxType == Ctlp_Underapprox_c) {
      thisApproxType = Ctlp_Underapprox_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    if (!fixpoint) {
      if (TRMinimization == Ctlp_Overapprox_c ||
	  thisApproxType == Ctlp_Overapprox_c) {
	thisApproxType = Ctlp_Incomparable_c;
      } else if (thisApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Underapprox_c;
      }
    } else {
      if ((TRMinimization == Ctlp_Overapprox_c &&
	   thisApproxType == Ctlp_Underapprox_c) ||
	  (TRMinimization == Ctlp_Underapprox_c &&
	   thisApproxType == Ctlp_Overapprox_c)) {
	thisApproxType = Ctlp_Incomparable_c;
      } else if (thisApproxType == Ctlp_Exact_c) {
	thisApproxType = TRMinimization;
      }
    }
    break;
  case Ctlp_EG_c:
    if (leftApproxType == Ctlp_Incomparable_c ||
	rightApproxType == Ctlp_Incomparable_c ||
	(leftApproxType == Ctlp_Overapprox_c &&
	 rightApproxType == Ctlp_Underapprox_c) ||
	(leftApproxType == Ctlp_Underapprox_c &&
	 rightApproxType == Ctlp_Overapprox_c)) {
      thisApproxType = Ctlp_Incomparable_c;
    } else if (leftApproxType == Ctlp_Overapprox_c ||
	       rightApproxType == Ctlp_Overapprox_c) {
      thisApproxType = Ctlp_Overapprox_c;
    } else if (leftApproxType == Ctlp_Underapprox_c ||
	       rightApproxType == Ctlp_Underapprox_c) {
      thisApproxType = Ctlp_Underapprox_c;
    } else {
      thisApproxType = Ctlp_Exact_c;
    }
    if (!fixpoint) {
      if (TRMinimization == Ctlp_Underapprox_c ||
	  thisApproxType == Ctlp_Underapprox_c) {
	thisApproxType = Ctlp_Incomparable_c;
      } else if (thisApproxType == Ctlp_Exact_c) {
	thisApproxType = Ctlp_Overapprox_c;
      }
    } else {
      if ((TRMinimization == Ctlp_Overapprox_c &&
	   thisApproxType == Ctlp_Underapprox_c) ||
	  (TRMinimization == Ctlp_Underapprox_c &&
	   thisApproxType == Ctlp_Overapprox_c)) {
	thisApproxType = Ctlp_Incomparable_c;
      } else if (thisApproxType == Ctlp_Exact_c) {
	thisApproxType = TRMinimization;
      }
    }
    break;
  case Ctlp_Cmp_c:
  case Ctlp_Init_c:
  case Ctlp_FwdU_c:
  case Ctlp_FwdG_c:
  case Ctlp_EY_c:
    /* no early termination and no hints in forward model checking */
    thisApproxType = Ctlp_Exact_c;
    break;
  default: fail("Encountered unknown type of CTL formula\n");
  }

  return thisApproxType;
  
} /* ComputeResultingAproximation */
